/**
 * Enemy component
 * @moduleName Enemy
 * @snippet Enemy.snippet
Enemy({
    physics: physics,
    spine: spine,
    onDead: function () {}
})
 */
bento.define('components/enemy', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/components/sprite',
    'components/planetwrap',
    'globals',
    'entities/effect'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Sprite,
    PlanetWrap,
    Globals,
    Effect
) {
    'use strict';
    var animations = null;
    var StarParticle = function (settings) {
        var sprite = new Sprite({
            imageName: 'collectible',
            originRelative: new Vector2(0.5, 0.5)
        });
        var planet = settings.planet || {};
        var velocity = settings.velocity || new Vector2(0, 0);
        var planetWrap = new PlanetWrap({
            planet: planet
        });
        var behavior = {
            start: function () {},
            update: function () {
                // move
                entity.position.x += velocity.x;
                entity.position.y += velocity.y;
                velocity.y += 0.8;

                // rotate
                entity.rotation += 0.1 * Utils.sign(velocity.x);
                // scale down
                // entity.scale.x = Utils.approach(entity.scale.x, 0, 0.05);
                // entity.scale.y = entity.scale.y;

                // fade out
                entity.alpha = Utils.approach(entity.alpha, 0, 1 / 30);

                if (entity.alpha === 0) {
                    entity.removeSelf();
                }
            }
        };
        var entity = new Entity({
            z: Globals.Layers.PARTICLES,
            name: 'starParticle',
            position: settings.position,
            scale: new Vector2(0.75, 0.75),
            components: [
                sprite,
                behavior,
                planetWrap
            ]
        });
        return entity;
    };
    return function (settings) {
        var viewport = Bento.getViewport();
        var physics = settings.physics;
        var spine = settings.spine;
        var onDead = settings.onDead;
        var entity;
        var dead = false;
        var vulnerableOffset = Vector2(0, -2);
        var component = {
            name: 'enemy',
            player: null,
            playerMovable: null,
            getPlayer: function () {
                Bento.objects.get('player', function (player) {
                    component.player = player;
                    player.getComponent('movable', function (movable) {
                        component.playerMovable = movable;
                    });
                });
            },
            start: function (data) {
                component.getPlayer();
            },
            destroy: function (data) {
                EventSystem.fire('enemyDestroyed', entity);
            },
            update: function (data) {
                if (dead) {
                    return;
                }
                if (!component.player) {
                    // no reference to player
                    component.getPlayer();
                    return;
                }
                // check if we collide with the player
                if (component.player.isAlive() && Utils.samePlanet(component.player, entity)) {
                    // check if the player hit our head
                    entity.collidesWith({
                        entity: component.player,
                        offset: vulnerableOffset,
                        onCollide: function () {
                            if (component.playerMovable.velocity.y > 0) {
                                Utils.timeout(0, function () {
                                    // defer the bounce for a bit because simultaneous enemy stomps can cause a death
                                    component.player.bounceUp();
                                });
                                entity.removeFromFamily('deadly');
                                die();
                            } else {
                                EventSystem.fire('killPlayer');
                            }
                        }
                    });
                }

                if (!physics) {
                    return;
                }
                EventSystem.fire('prepareOneWayPlatforms', entity);
                // turn around if we reach a turning point (e.g. at the edge of a platform)
                if (entity.collidesWith({
                        family: 'enemyTurner',
                        offset: new Vector2(Utils.sign(physics.velocity.x, 0) * 10, 0)
                    })) {
                    physics.velocity.x *= -1;
                }

                // look in direction of traveling
                if (!dead && physics.velocity.x !== 0 && Utils.sign(-entity.scale.x) !== Utils.sign(physics.velocity.x)) {
                    entity.scale.x *= -1;
                }
                // walking animation
                if (Math.abs(physics.velocity.x) > 0 && hasAnimation('walk')) {
                    spine.setAnimation('walk');
                } else if (!dead && hasAnimation('idle')) {
                    spine.setAnimation('idle');
                } else {}
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
            }
        };
        var hasAnimation = function (animationName) {
            var cacheAnimations = function () {
                var spineData = spine.getSpineData();
                var animationArray = spineData.skeletonData.data.animations;
                animations = animations || {};
                animations[entity.name] = animations[entity.name] || {};
                Utils.forEach(animationArray, function (data, i, l, breakLoop) {
                    animations[entity.name][data.name] = true;
                });
            };
            if (!animations || !animations[entity.name]) {
                cacheAnimations();
            }
            return animations[entity.name][animationName];
        };
        var die = function () {
            if (dead) {
                return;
            }
            Bento.audio.stopSound('sfx_enemy');
            Bento.audio.playSound('sfx_enemy');
            physics.velocity.x = 0;
            dead = true;
            new Tween({
                from: 1,
                to: 0,
                in: 30,
                ease: 'elastic',
                decay: 5,
                oscillations: 2,
                onUpdate: function (v, t) {
                    entity.scale.x = 1 + v;
                    entity.scale.y = 1 - v;
                }
            });
            if (hasAnimation('death')) {
                spine.setAnimation('death', function () {}, false);
            }
            entity.getComponent('planetWrap', function (planetWrap) {
                Bento.objects.attach(new StarParticle({
                    position: entity.position.add(new Vector2(-40, -64)),
                    velocity: new Vector2(-4, -12),
                    planet: planetWrap.planet
                }));
                Bento.objects.attach(new StarParticle({
                    position: entity.position.add(new Vector2(40, -64)),
                    velocity: new Vector2(4, -12),
                    planet: planetWrap.planet
                }));
            });
            EventSystem.fire('killedEnemy', entity);

            if (onDead) {
                onDead();
            }
            
            // tween out after a while
            new Tween({
                from: entity.scale.x,
                to: 0,
                in: 60,
                ease: 'easeInBack',
                delay: 30,
                onStart: function () {
                    entity.removeByName('movable');
                },
                onUpdate: function (v, t) {
                    entity.scale.x = v;
                    entity.scale.y = Math.abs(v);
                },
                onComplete: function () {
                    entity.getComponent('planetWrap', function (planetWrap) {
                        var effect = new Effect({
                            z: Globals.Layers.ENEMIES,
                            position: entity.position.add(new Vector2(0, -32)),
                            spriteSheet: 'effects/poof'
                        });
                        effect.attach(new PlanetWrap({
                            planet: planetWrap.planet
                        }));
                        effect.scale.x = 1.5;
                        effect.scale.y = 1.5;
                        Bento.objects.attach(effect);
                    });
                    entity.removeSelf();
                }
            });
        };
        Object.defineProperty(component, 'dead', {
            get: function () {
                return dead;
            },
            set: function (value) {
                dead = value;
            }
        });
        return component;
    };
});