/**
 * The movable component moves the parent entity constantly according its speed
 * It will also stop movement if it hits an object with family 'solid'
 */
bento.define('components/movable', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    var EPSILON = Number.EPSILON || Math.pow(2, -52);

    var solids = Bento.objects.getByFamily('solid'),
        shouldIgnore = function (ignoredArray, familyArray) {
            var i, family;

            if (!familyArray) {
                return false;
            }

            for (i = 0; i < familyArray.length; ++i) {
                family = familyArray[i];
                if (ignoredArray.indexOf(family) >= 0) {
                    return true;
                }
            }
            return false;
        },
        Movable = function (settings) {
            this.name = 'movable';
            this.paused = false;
            this.velocity = settings.velocity || new Vector2(0, 0);
            this.maxVelocity = settings.maxVelocity || new Vector2(Infinity, Infinity);
            this.collided = [];
            this.collidedEntity = null;
            this.entity = null;
            this.tileColliders = null;
            this.ignorePolygons = true;
            this.settings = settings || {};
            this.gravity = this.settings.gravity || 0;
            this.settings.friction = this.settings.friction || new Vector2(1, 1);
            this.settings.bounce = this.settings.bounce || new Vector2(0, 0);
            this.settings.ignoreSolids = this.settings.ignoreSolids || false;
            this.settings.onCollide = this.settings.onCollide || function () {};
            this.settings.useHshg = this.settings.useHshg || false;
            this.settings.ignoredFamilies = this.settings.ignoredFamilies || [];
            this.collisionListeners = [];
            this.boxRatio = 1;
            // prevent tunneling, default value set during start
            this.maxStep = this.settings.maxStep || new Vector2(Infinity, Infinity);
        };
    Movable.prototype.update = function (data) {
        var entity = this.entity,
            velocity = this.velocity,
            normalizedVelocity = velocity.clone().normalize(),
            collided = this.collided,
            otherBox,
            settings = this.settings,
            other,
            self = this,
            pushDirection = 'x',
            pushAmount = 0,
            stepX = 0,
            stepY = 0,
            totalStepX = 0,
            totalStepY = 0,
            direction = new Vector2(Utils.sign(velocity.x), Utils.sign(velocity.y)),
            onCollide = function () {
                var i,
                    box,
                    diffBox,
                    pushX = function (dir) {
                        // pushes out entity in x
                        // dir = -1 or 1
                        entity.position.x += Math.ceil(diffBox.width) * dir;
                    },
                    pushY = function (dir) {
                        // pushes out entity in y
                        // dir = -1 or 1
                        entity.position.y += Math.ceil(diffBox.height) * dir;
                    },
                    isColliding = function () {
                        return other.collidesWith(entity);
                    },
                    newVelocity = function (direction) {
                        // calculates new speed: bounce in normal direction, friction in perpendicular
                        // direction should be 'x' or 'y'
                        var perpendicular = (direction === 'x' ? 'y' : 'x');
                        velocity[direction] *= -settings.bounce[direction];
                        velocity[perpendicular] *= settings.friction[perpendicular];
                        pushDirection = direction;
                    };

                // collect difference rectangle 
                for (i = 0; i < collided.length; ++i) {
                    other = collided[i];
                    otherBox = other.getBoundingBox();

                    // polygon?
                    if (otherBox.isPolygon) {
                        // push out in reverse direction until not colliding
                        entity.position.x -= normalizedVelocity.x;
                        entity.position.y -= normalizedVelocity.y;
                        while (isColliding()) {
                            entity.position.x -= normalizedVelocity.x;
                            entity.position.y -= normalizedVelocity.y;
                        }
                        // temp: reverse direction
                        // todo: find out which direction the bounce really is
                        velocity['x'] *= -settings.bounce['x'];
                        velocity['y'] *= -settings.bounce['y'];
                    } else {
                        // normal rectangle collision
                        diffBox = entity.getBoundingBox().intersection(otherBox);

                        if (diffBox.width === 0 && diffBox.height === 0) {
                            // not colliding?
                            // hshg is wrong apparently?
                            continue;
                        }

                        // push out of x or y?
                        // push out based on box size
                        if (diffBox.width / diffBox.height < self.boxRatio) {
                            pushX(-direction.x);
                            // check
                            if (isColliding()) {
                                // failed!
                                // revert x and try y instead
                                pushX(direction.x);
                                pushY(-direction.y);
                                //  
                                newVelocity('y');
                                pushAmount = -direction.y;
                            } else {
                                newVelocity('x');
                                pushAmount = -direction.x;
                            }
                        } else {
                            pushY(-direction.y);
                            if (isColliding()) {
                                pushY(direction.y);
                                pushX(-direction.x);
                                newVelocity('x');
                                pushAmount = -direction.x;
                            } else {
                                newVelocity('y');
                                pushAmount = -direction.y;
                            }
                        }
                    }
                }
                // collision with solid 
                settings.onCollide({
                    entity: self.entity,
                    others: collided,
                    other: other,
                    pushDirection: pushDirection,
                    pushAmount: pushAmount
                });

                for (i = 0; i < self.collisionListeners.length; ++i) {
                    self.collisionListeners[i]({
                        entity: self.entity,
                        others: collided,
                        other: other,
                        pushDirection: pushDirection,
                        pushAmount: pushAmount
                    });
                }
            },
            doCollision = function () {
                // check if collided
                if (self.tileColliders && !settings.ignoreSolids) {
                    self.tileColliders.collidesWith(entity, new Vector2(0, 0), function (other) {
                        self.collidedEntity = other;
                        collided.push(other);
                        if (settings.onCollideHash) {
                            settings.onCollideHash(other);
                        }
                    });
                } else {
                    Bento.objects.get('tileColliders', function (obj) {
                        self.tileColliders = obj;
                    });
                }
                // collision with solids
                if (!self.settings.useHshg && !settings.ignoreSolids) {
                    entity.collidesWith({
                        entities: solids,
                        firstOnly: false,
                        onCollide: function (other) {
                            // fine collision
                            var otherBox = other.getBoundingBox();
                            // is this an ignored object?
                            if (shouldIgnore(self.settings.ignoredFamilies, other.family)) {
                                return;
                            }

                            if (otherBox.isPolygon) {
                                // default behavior: ignore polygons because they cause bugs
                                if (self.ignorePolygons) {
                                    return;
                                }
                                if (!otherBox.intersect(entity.getBoundingBox())) {
                                    return;
                                }
                            }

                            self.collidedEntity = other;
                            self.collided.push(other);
                        }
                    });
                }

                if (collided.length) {
                    // resolve collision
                    onCollide();
                    return true;
                } else {
                    return false;
                }
            };

        if (this.paused) {
            return;
        }

        // gravity
        if (this.gravity) {
            velocity.y += this.gravity * data.speed;
        }

        // cap speed
        if (this.maxVelocity) {
            velocity.x = Utils.clamp(-this.maxVelocity.x, velocity.x, this.maxVelocity.x);
            velocity.y = Utils.clamp(-this.maxVelocity.y, velocity.y, this.maxVelocity.y);
        }

        // direction should not be 0
        direction.x = direction.x || 1;
        direction.y = direction.y || 1;

        // move entity
        // do not overstep max movement to prevent tunneling
        totalStepX = Math.abs(velocity.x * data.speed);
        totalStepY = Math.abs(velocity.y * data.speed);

        while (totalStepX > EPSILON || totalStepY > EPSILON) {
            stepX = Math.min(this.maxStep.x, totalStepX);
            stepY = Math.min(this.maxStep.y, totalStepY);
        
            entity.position.x += direction.x * stepX;
            entity.position.y += direction.y * stepY;

            totalStepX -= stepX;
            totalStepY -= stepY;

            if (doCollision()) {
                // not sure about this yet: if it had to resolve any collision, stop movement now
                break;
            }
        }

        // reset collided
        this.collided.length = 0;
        this.collidedEntity = null;
    };
    Movable.prototype.onParentCollided = function (data) {
        var otherBox,
            other = data.other;
        if (this.settings.ignoreSolids || !this.settings.useHshg) {
            return;
        }
        // is this an ignored object?
        if (shouldIgnore(this.settings.ignoredFamilies, other.family)) {
            return;
        }

        // is of family solid?
        if (data.other.family.indexOf('solid') !== -1) {
            // fine collision
            otherBox = data.other.getBoundingBox();
            if (otherBox.isPolygon) {
                // default behavior: ignore polygons because they cause bugs
                if (this.ignorePolygons) {
                    return;
                }
                if (!otherBox.intersect(data.entity.getBoundingBox())) {
                    return;
                }
            }

            this.collidedEntity = data.other;
            this.collided.push(data.other);
        }
    };
    Movable.prototype.start = function (data) {
        var box = data.entity.boundingBox || data.entity.dimension;
        this.boxRatio = box.width / box.height;
        // default value is half the boundingbox size
        if (!this.settings.maxStep) {
            this.maxStep.x = Math.max(Math.round(box.width / 2), 1);
            this.maxStep.y = Math.max(Math.round(box.height / 2), 1);
        }
    };
    Movable.prototype.attached = function (data) {
        var self = this;
        this.entity = data.entity;
        if (!this.tileColliders) {
            Bento.objects.get('tileColliders', function (obj) {
                self.tileColliders = obj;
            });
        }
    };
    return Movable;
});