/**
 * Causes an entity to follow a path.
 * @moduleName PathFollower
 * @snippet PathFollower.snippet
PathFollower({
    
})
 */
bento.define('components/pathfollower', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    return function (settings) {
        var viewport = Bento.getViewport();
        var pathName = settings.pathName;
        var entity;
        
        /* properties that we get from the path */

        var points;  // array of points e.g. [ {x:10,y:20} , {x:5,y:3} ]
        var timePerPoint;
        var ease;
        var toggleId;

        // 'onOff'    - don't move until button is pressed
        // 'offOn'    - keep moving until button is pressed
        // 'perPoint' - at every point, wait until button is pressed
        // 'perCycle' - every time we reach the starting point, wait until button is pressed
        var toggleMode;

        /* variables for movement logic */
        var startPos, previousFramePos;
        var currentPointIndex = 0;
        var currentPoint;
        var nextPoint;
        var closedLoop = false;  // we move back and forth unless the final point is the same as the first point
        var reversing = false;   // are we moving backwards along the path?
        
        var tween, tweenPaused;

        var component = {
            name: 'pathFollower',
            start: function (data) {
                startPos = entity.position.clone();
                previousFramePos = entity.position.clone();
                EventSystem.on('toggleSwitchPressed', onToggleSwitchPressed);
            },
            destroy: function (data) {
                EventSystem.off('toggleSwitchPressed', onToggleSwitchPressed);
            },
            update: function (data) {
                // TODO: come up with a more graceful way to find the path even if it doesn't exist when we initially spawn?
                if (!points) {
                    obtainPathInfo();
                }
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
            }
        };

        var obtainPathInfo = function () {
            var path = Bento.objects.get(pathName);
            if (!path) {
                return;
            }
            points = path.points;
            timePerPoint = path.timePerPoint || 120;
            ease = path.ease || 'easeInOutQuad';
            toggleId = path.toggle || null;
            toggleMode = path.toggleMode || 'onOff';

            currentPointIndex = 0;
            currentPoint = points[0];
            nextPoint = points[1];

            if (points[0].x === points[points.length-1].x && points[0].y === points[points.length-1].y) {
                closedLoop = true;
            }

            moveToNextPoint();
            if (toggleId && (toggleMode === 'onOff')) {
                tweenPaused = true;
                tween.pause();
            }
        };

        var tweenUpdate = function (t) {
            
            previousFramePos.x = entity.position.x;
            previousFramePos.y = entity.position.y;

            var pathX = currentPoint.x + (nextPoint.x - currentPoint.x) * t;
            var pathY = currentPoint.y + (nextPoint.y - currentPoint.y) * t;

            entity.position.x = startPos.x + pathX;
            entity.position.y = startPos.y + pathY;
        };

        // figure out the next point
        // start the tween to go there
        // when we're done, the current point becomes the next point
        // then we repeat the process.
        var moveToNextPoint = function () {
            var nextPointIndex;
            if (!reversing) {
                nextPointIndex = currentPointIndex + 1;
                if (closedLoop) {
                    if (nextPointIndex >= points.length - 1) {
                        nextPointIndex = 0;
                    }
                } else {
                    if (nextPointIndex >= points.length) {
                        reversing = true;
                        nextPointIndex = currentPointIndex - 1;
                    }
                }
            } else {
                nextPointIndex = currentPointIndex - 1;
                if (nextPointIndex < 0) {
                    nextPointIndex = 1;
                    reversing = false;
                }
            }
            nextPoint = points[nextPointIndex];

            tween = new Tween({
                from: 0, to: 1,
                in: timePerPoint,
                ease: ease,
                onUpdate: tweenUpdate,
                onComplete: function () {
                    currentPointIndex = nextPointIndex;
                    currentPoint = nextPoint;
                    moveToNextPoint();
                }
            });

            // if toggleable, we might have to wait until we're allowed to move again
            if (toggleId) {
                if (toggleMode === 'perPoint' || (toggleMode === 'perCycle' && currentPointIndex === 0)) {
                    tweenPaused = true;
                    tween.pause();
                }
            }
        };

        var onToggleSwitchPressed = function (id) {
            if (toggleId && (id === toggleId)) {
                if (tweenPaused) {
                    tweenPaused = false;
                    tween.resume();
                } else if (toggleMode === 'onOff' || toggleMode === 'offOn') {
                    tweenPaused = true;
                    tween.pause();
                }
            }
        };

        return component;
    };
});