/**
 * Adds arcade style physics (velocity, acceleration, collision) to an entity.
 * @moduleName Physics
 * @snippet Physics.snippet
Physics({
    vel: new Vector2(0, 0),
    maxVel: new Vector2(999, 999),
    accel: new Vector2(0, 0),
    drag: new Vector2(0, 0),
    onCollideX: function (other) {
    
    },
    onCollideY: function (other) {
    
    }
})
 */
bento.define('components/physics', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        
        var subPixel = new Vector2(0, 0);
        var checkCollisions;

        // custom callbacks, called for each solid object we collide with
        // return true to ignore the collision and move through
        var onCollideX = settings.onCollideX || function(){ return false; };
        var onCollideY = settings.onCollideY || function(){ return false; };

        var component = {
            name: 'physics',
            vel: settings.vel || new Vector2(0, 0),
            maxVel: settings.maxVel || new Vector2(999, 999),
            accel: settings.accel || new Vector2(0, 0),
            drag: settings.drag || new Vector2(0, 0),
            start: function (data) {},
            destroy: function (data) {},
            update: function (data) {
                var vel = component.vel;
                var maxVel = component.maxVel;
                var accel = component.accel;
                var drag = component.drag;

                // apply acceleration if any, otherwise apply drag
                if (accel.x) {
                    vel.x += accel.x;
                } else {
                    if (vel.x > 0) vel.x = Math.max(vel.x - drag.x, 0);
                    else if (vel.x < 0) vel.x = Math.min(vel.x + drag.x, 0);
                }
                if (accel.y) {
                    vel.y += accel.y;
                } else {
                    if (vel.y > 0) vel.y = Math.max(vel.y - drag.y, 0);
                    else if (vel.y < 0) vel.y = Math.min(vel.y + drag.y, 0);
                }

                // limit velocity
                vel.x = Utils.clamp(-maxVel.x, vel.x, maxVel.x);
                vel.y = Utils.clamp(-maxVel.y, vel.y, maxVel.y);

                var sign = new Vector2(vel.x > 0 ? 1 : -1, vel.y > 0 ? 1 : -1);
                var signX = new Vector2(sign.x, 0);
                var signY = new Vector2(0, sign.y);
                var offset = vel.clone();
                var position = entity.position;
                var hitSolid, collisions;

                // accumulate fractions of movement, so that the entity can
                // move by whole pixels even if vel is not an integer
                subPixel.addTo(offset);
                offset.x = Math.floor(subPixel.x);
                offset.y = Math.floor(subPixel.y);
                subPixel.subtractFrom(offset);

                // collision resolution on the X axis:
                hitSolid = false;
                entity.collidesWith({
                    family: 'solid',
                    offset: new Vector2(offset.x, 0),
                    firstOnly: false,
                    onCollide: function (other) {
                        // custom handler allows parent entity to ignore the collision if desired
                        if (!onCollideX(other))
                            hitSolid = true;
                    }
                });
                
                // if we are colliding, move pixel-by-pixel until we touch the wall
                if (hitSolid) {
                    while (offset.x !== 0 && !entity.collidesWith({family: 'solid', offset: signX})) {
                        position.x += sign.x;
                        offset.x -= sign.x;
                    }
                } else {
                    position.x += offset.x;
                }

                // do the same for Y axis
                hitSolid = false;
                collisions = entity.collidesWith({
                    family: 'solid',
                    offset: new Vector2(0, offset.y),
                    firstOnly: false,
                    onCollide: function (other) {
                        if (!onCollideY(other))
                            hitSolid = true;
                    }
                });
                if (hitSolid) {
                    while (offset.y !== 0 && !entity.collidesWith({family: 'solid', offset: signY})) {
                        position.y += sign.y;
                        offset.y -= sign.y;
                    }
                } else {
                    position.y += offset.y;
                }
            },
            attached: function (data) {
                entity = data.entity;
            },
        };

        return component;
    };
});