/**
 * Force an entity's rendering to be transformed relative to the perimeter of the planet.
 * Also wraps the entity horizontally once it travels the full perimeter.
 * correctBoundingBox defaults to true, to fix collisions in these polar coordinates.
 * @moduleName PlanetWrap
 * @snippet PlanetWrap.snippet
PlanetWrap({
    planet: planet,             // the planet to wrap around (mandatory)
    correctBoundingBox: true,   // adjust bounding box based on spawn height and movement direction
    managedFamilies: ['solid'], // automatically add/remove entity from these families based on whether our planet is active.
    debug: false                // draw perceived hitbox
})
 */
bento.define('components/planetwrap', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Globals
) {
    'use strict';
    var twoPi = 2 * Math.PI;
    var PlanetWrap = function (settings) {
        
        if (Globals.SHOW_BOUNDING_BOXES) {
            settings.debug = true;
        }
        if (Globals.DISABLE_BOX_CORRECTION) {
            settings.correctBoundingBox = false;
        }

        var planet = settings.planet;
        var center = settings.center|| planet.position || new Vector2(0, 0);
        var managedFamilies = settings.managedFamilies || [];
        var correctBoundingBox = settings.correctBoundingBox !== false;
        var isWall = settings.isWall === true;
        var radius = planet.radius;
        var viewport = Bento.getViewport();
        var entity;
        var x, y, angle;
        var perimeter = planet.mapWidth;

        var onSetActivePlanet = function (newPlanet) {
            var i;
            if (newPlanet === component.planet) {
                for (i = 0; i < managedFamilies.length; i++) {
                    entity.addToFamily(managedFamilies[i]);
                }
            } else {
                for (i = 0; i < managedFamilies.length; i++) {
                    entity.removeFromFamily(managedFamilies[i]);
                }
            }
        };

        // debug rendering
        var colorArr = [0,0,0,1];
        var drawPerceivedBoundingBox = function (renderer) {
            var position = entity.position;
            var box = entity.getBoundingBox();

            if (!Utils.isFiniteRectangle(box)) {
                return;
            }

            var ul = box.getCorner(Rectangle.TOPLEFT);
            var dl = box.getCorner(Rectangle.BOTTOMLEFT);
            var ur = box.getCorner(Rectangle.TOPRIGHT);
            var dr = box.getCorner(Rectangle.BOTTOMRIGHT);

            if (!Globals.FLAT_RENDERING) {
                ul = Utils.wrapPoint(ul, radius, planet.angle).addTo(ul);
                dl = Utils.wrapPoint(dl, radius, planet.angle).addTo(dl);
                ur = Utils.wrapPoint(ur, radius, planet.angle).addTo(ur);
                dr = Utils.wrapPoint(dr, radius, planet.angle).addTo(dr);
            }

            renderer.rotate(-entity.rotation);
            renderer.scale(1 / entity.scale.x, 1 / entity.scale.y);
            renderer.translate(-position.x, -position.y);
            renderer.drawLine(colorArr, ul.x, ul.y, dl.x, dl.y, 3);
            renderer.drawLine(colorArr, dl.x, dl.y, dr.x, dr.y, 3);
            renderer.drawLine(colorArr, dr.x, dr.y, ur.x, ur.y, 3);
            renderer.drawLine(colorArr, ur.x, ur.y, ul.x, ul.y, 3);
            renderer.translate(position.x, position.y);
            renderer.scale(entity.scale.x, entity.scale.y);
            renderer.rotate(entity.rotation);
        };

        // bounding box correction
        var oldGetBoundingBox;
        var customGetBoundingBox = function () {
            var box = oldGetBoundingBox.call(entity);
            var widthToCorrect = isWall ? 128 : box.width;
            var angleDiff = widthToCorrect / radius;
            var entityRadius = radius - box.y;
            var halfWidth = entityRadius * Math.sin(angleDiff*0.5);
            var correctionRatio = widthToCorrect / (halfWidth*2);
            var correctionAmount = widthToCorrect - widthToCorrect*correctionRatio;
            box.width -= correctionAmount;
            box.x += correctionAmount * 0.5;
            return box;
        };

        var component = {
            name: 'planetWrap',
            planet: planet,

            // use these to get information about where the entity is actually rendered in the world
            angle: 0,
            // offset: new Vector2(0, 0),

            start: function (data) {
                EventSystem.on('setActivePlanet', onSetActivePlanet);
                // update once, shadowEntity needs this
                component.update(data);
            },
            destroy: function (data) {
                EventSystem.on('setActivePlanet', onSetActivePlanet);
            },
            update: function (data) {
                var position;
                if (planet.infinite) {
                    // no wrapping
                    return;
                }

                position = entity.position;

                // wrap position once we travel 360 degrees round the planet
                while (position.x > PlanetWrap.wrapOffset + planet.perimeter) {
                    position.x -= Math.floor(planet.perimeter);
                }
                while (position.x < PlanetWrap.wrapOffset) {
                    position.x += Math.floor(planet.perimeter);
                }
            },
            draw: function (data) {

                if (settings.debug) {
                    drawPerceivedBoundingBox(data.renderer);
                    data.renderer.setOpacity(0.5);
                }

                // transform the entity

                // angle = planet.angle + (entity.position.x + component.lean) / radius;
                // var sin = Math.sin(angle);
                // var cos = Math.cos(angle);
                // var entityRadius = radius - entity.position.y;
                // x = sin * entityRadius - entity.position.x + planet.position.x;
                // y = (1-cos) * entityRadius - entity.position.y + planet.position.y - entityRadius;

                // data.renderer.rotate(-entity.rotation);
                // data.renderer.scale(1/entity.scale.x, 1/entity.scale.y);
                // if (Globals.FLAT_RENDERING) {
                //     data.renderer.translate(planet.position.x + planet.angle*planet.perimeter/(2*Math.PI), planet.position.y);
                // } else {
                //     data.renderer.translate(x, y);
                //     data.renderer.rotate(angle);
                // }
                // data.renderer.rotate(entity.rotation);
                // data.renderer.scale(entity.scale.x, entity.scale.y);

                // component.angle = angle;
                // component.offset.x = x;
                // component.offset.y = y;

                /* Alternate way of warping, less calculation */
                // undo original transform
                data.renderer.rotate(-entity.rotation);
                data.renderer.scale(1 / entity.scale.x, 1 / entity.scale.y);

                if (!Globals.FLAT_RENDERING) {
                    // relative to planet center
                    data.renderer.translate(planet.position.x - entity.position.x, planet.position.y - entity.position.y);
                    // get radians
                    angle = (entity.position.x % (twoPi * planet.radius)) / planet.radius + planet.angle;
                    // rotate with planet
                    data.renderer.rotate(angle);
                    // move up to surface
                    data.renderer.translate(0, -planet.radius + entity.position.y);
                }
                // redo transform
                data.renderer.scale(entity.scale.x, entity.scale.y);
                data.renderer.rotate(entity.rotation);

                component.angle = angle;
            },
            attached: function (data) {
                entity = data.entity;

                // transform must happen before any other rendering, so we move ourself to the front
                entity.moveComponentTo(component, 0);

                if (correctBoundingBox) {
                    oldGetBoundingBox = entity.getBoundingBox;
                    entity.getBoundingBox = customGetBoundingBox;
                }
            },
            removed: function (data) {
                if (correctBoundingBox) {
                    entity.getBoundingBox = oldGetBoundingBox;
                }
            },
            manageFamily: function (family) {
                if (managedFamilies.indexOf(family) === -1) {
                    managedFamilies.push(family);
                }
            },
            unmanageFamily: function (family) {
                var n = managedFamilies.indexOf(family);
                if (n !== -1) {
                    managedFamilies.splice(n, 1);
                }
            },

            // bounding box correction works differently for walls
            setWall: function (val) {
                isWall = val;
            },

            clone: function () {
                return new PlanetWrap({
                    planet: planet,
                    correctBoundingBox: correctBoundingBox,
                    managedFamilies: managedFamilies.slice()
                });
            }
        };
        return component;
    };
    PlanetWrap.wrapOffset = 0;
    
    return PlanetWrap;
});