/**
 * Apply scale/translate tweens to entity without modifying its hitbox.
 * Use onComplete callbacks to sequence animations (it's kinda ugly but does the job)
 * If you call a function whose tween is already in-progress, the old tween will be
 *  cancelled and the old onComplete will not fire.
 * @moduleName Squisher
 * @snippet Squisher.snippet
Squisher({})
 */
bento.define('components/squisher', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;

        var origin = settings.origin || Vector2(0, 0);
        var x = 0, y = 0;
        var scaleX = 1, scaleY = 1;
        var xMul = 1;

        var scaleXTween, scaleYTween, offsetXTween, offsetYTween;

        var component = {
            name: 'squisher',
            active: true,

            scale: function (dest, time, ease, onComplete) {
                component.scaleX(dest, time, ease);
                component.scaleY(dest, time, ease, onComplete);
            },
            offset: function (dest, time, ease, onComplete) {
                component.offsetX(dest, time, ease);
                component.offsetY(dest, time, ease, onComplete);
            },

            scaleX: function(dest, time, ease, onComplete) {
                if (scaleXTween) {
                    scaleXTween.stop();
                    Bento.objects.remove(scaleXTween);
                }
                scaleXTween = new Tween({
                    from: scaleX, to: dest, in: time, ease: ease,
                    onUpdate: function (v,t) { scaleX = v; },
                    onComplete: onComplete
                });
            },
            scaleY: function(dest, time, ease, onComplete) {
                if (scaleYTween) {
                    scaleYTween.stop();
                    Bento.objects.remove(scaleYTween);
                }
                scaleYTween = new Tween({
                    from: scaleY, to: dest, in: time, ease: ease,
                    onUpdate: function (v,t) { scaleY = v; },
                    onComplete: onComplete
                });
            },
            offsetX: function(dest, time, ease, onComplete) {
                if (offsetXTween) {
                    offsetXTween.stop();
                    Bento.objects.remove(offsetXTween);
                }
                offsetXTween = new Tween({
                    from: x, to: dest, in: time, ease: ease,
                    onUpdate: function (v,t) { x = v; },
                    onComplete: onComplete
                });
            },
            offsetY: function(dest, time, ease, onComplete) {
                if (offsetYTween) {
                    offsetYTween.stop();
                    Bento.objects.remove(offsetYTween);
                }
                offsetYTween = new Tween({
                    from: y, to: dest, in: time, ease: ease,
                    onUpdate: function (v,t) { y = v; },
                    onComplete: onComplete
                });
            },

            setFlip: function (flip) {
                xMul = flip ? -1 : 1;
            },

            // Plays the bouncing animation used by fish
            playBounceAnimation: function () {
                // leaving the ground
                component.offsetY(-30, 12, 'easeOutQuad', function () {
                    
                    // max height
                    component.offsetY(0, 12, 'easeInQuad', function () {
                        
                        // hitting the ground
                        component.scaleX(1.1, 4, 'easeOutQuad', function () {
                            
                            // max squash
                            component.scaleX(1.0, 4, 'easeInQuad', function () {

                                // begin stretch
                                component.scaleX(0.9, 6, 'easeOutQuad', function () {

                                    // max stretch
                                    component.scaleX(1.0, 6, 'easeInQuad');
                                });
                            });
                        });

                        // hitting the ground (in parallel)
                        component.scaleY(0.9, 4, 'easeOutQuad', function () {

                            // max squash
                            component.scaleY(1.0, 4, 'easeInQuad', function () {

                                // begin stretch
                                component.scaleY(1.1, 6, 'easeOutQuad', function () {

                                    // max stretch
                                    component.scaleY(1.0, 6, 'easeInQuad');
                                });
                            });
                        });
                    });
                });
            },

            start: function (data) {},
            destroy: function (data) {},
            update: function (data) {},
            draw: function (data) {
                if (!component.active) return;
                var renderer = data.renderer;
                renderer.translate(origin.x, origin.y);
                renderer.scale(scaleX*xMul, scaleY);
                renderer.translate(x, y);
                renderer.translate(-origin.x, -origin.y);
            },

            attached: function (data) {
                entity = data.entity;

                // should transform before drawing but after planetwrap
                var wrap = entity.getComponent('planetWrap');

                if (wrap) {
                    entity.moveComponentTo(component, wrap.rootIndex + 1);
                } else {
                    entity.moveComponentTo(component, 0);
                }
            }

        };
        return component;
    };
});