/**
 * Triangle renderer, stolen from spine
 * @moduleName TriangleRenderer
 * @snippet TriangleRenderer.snippet
TriangleRenderer({
    
})
 */
bento.define('components/trianglerenderer', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/packedimage'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    PackedImage
) {
    'use strict';
    var drawTriangle = function (ctx, img, x0, y0, u0, v0, x1, y1, u1, v1, x2, y2, u2, v2) {
        u0 *= img.width;
        v0 *= img.height;
        u1 *= img.width;
        v1 *= img.height;
        u2 *= img.width;
        v2 *= img.height;
        ctx.beginPath();
        ctx.moveTo(x0, y0);
        ctx.lineTo(x1, y1);
        ctx.lineTo(x2, y2);
        ctx.closePath();
        x1 -= x0;
        y1 -= y0;
        x2 -= x0;
        y2 -= y0;
        u1 -= u0;
        v1 -= v0;
        u2 -= u0;
        v2 -= v0;
        var det = 1 / (u1 * v2 - u2 * v1),
            a = (v2 * x1 - v1 * x2) * det,
            b = (v2 * y1 - v1 * y2) * det,
            c = (u1 * x2 - u2 * x1) * det,
            d = (u1 * y2 - u2 * y1) * det,
            e = x0 - a * u0 - c * v0,
            f = y0 - b * u0 - d * v0;
        ctx.save();
        ctx.transform(a, b, c, d, e, f);
        ctx.clip();
        ctx.drawImage(img.image, img.x, img.y, img.width, img.height, 0, 0, img.width, img.height);
        ctx.drawImage(img.image, img.x, img.y, img.width, img.height, 0, 0, img.width, img.height);
        ctx.drawImage(img.image, img.x, img.y, img.width, img.height, 0, 0, img.width, img.height);
        ctx.restore();
    };
    // triangulate a rectangle like this: |/|/|/|
    var triangulate = function (rectangle, verticalSlices) {
        var i = 0;
        var outputArray = [];
        var width = rectangle.width / (verticalSlices + 1);
        var height = rectangle.height;
        for (i = 0; i < verticalSlices + 1; ++i) {
            // upper left
            outputArray.push(new Triangle(
                new Vector2(rectangle.x + i * width, rectangle.y),
                new Vector2(rectangle.x + i * width, rectangle.y + height),
                new Vector2(rectangle.x + (i + 1) * width, rectangle.y)
            ));
            // lower right
            outputArray.push(new Triangle(
                new Vector2(rectangle.x + i * width, rectangle.y + height),
                new Vector2(rectangle.x + (i + 1) * width, rectangle.y + height),
                new Vector2(rectangle.x + (i + 1) * width, rectangle.y)
            ));
        }
        return outputArray;
    };
    var Triangle = function (v1, v2, v3) {
        return {
            v1: v1,
            v2: v2,
            v3: v3,
            forEach: function (callback) {
                callback(v1, 0);
                callback(v2, 1);
                callback(v3, 2);
            },
            clone: function () {
                return new Triangle(v1.clone(), v2.clone(), v3.clone());
            }
        };
    };
    var twoPi = 2 * Math.PI;
    /*
    cacheObject = {
        canvas: HTML Canvas Element
        origin: Vector2
    };
     */
    // note: do we keep the cache forever?
    var canvasCache = {};

    return function (settings) {
        var viewport = Bento.getViewport();
        var tileWidth = settings.tileWidth || 16;
        var tileHeight = settings.tileHeight || 16;
        var objectWidth = settings.objectWidth || tileWidth;
        var objectHeight = settings.objectHeight || tileHeight;
        var gridScale = (settings.imageSize || 128) / tileWidth;
        var entity;
        var destinationTriangles;
        var sourceTriangles;
        var sprite = settings.sprite;
        var planet;
        var planetWrap;
        var slices = 1;

        var imagePattern = settings.imagePattern || 'wall';
        var wallLeft = Bento.assets.getImage(imagePattern + '-left');
        var wallMiddle = Bento.assets.getImage(imagePattern + '-middle');
        var wallRight = Bento.assets.getImage(imagePattern + '-right');

        // create a cached drawing
        var canvas;
        var context;
        var renderer;
        var canvasSize = new Rectangle(0, 0, 1, 1);

        // we identify a canvas by image name and objectWidth
        var canvasId;

        var component = {
            name: 'triangleRenderer',
            calculateArc: function () {
                var posY = entity.position.y;
                var radius = planet.radius;
                var origin;
                var sourceRectangle;
                var destinationRectangle;
                var spriteImage;
                var cacheObject;

                var getCorrectionRatio = function (rr, bb) {
                    var width = tileWidth * gridScale; // halfwidth
                    var height = tileHeight * gridScale;
                    rr = (radius - posY) + (bb ? -height : 0);
                    var a = Math.atan2(width, rr + height) * rr;
                    var arc = (width) / planet.perimeter * twoPi * rr;
                    return tileWidth - tileWidth * a / arc;
                };

                if (!sprite) {
                    sprite = entity.getComponent('sprite');
                }
                spriteImage = sprite.spriteImage;
                origin = sprite.origin.clone();

                // get texturepacker offset
                sourceRectangle = new Rectangle(0, 0, 1, 1);
                // turn into uv (no need to take packed image offset into account yet)
                sourceRectangle.x = 0;
                sourceRectangle.y = 0;
                sourceRectangle.width = 1;
                sourceRectangle.height = 1;

                // triangulate source
                sourceTriangles = triangulate(sourceRectangle, slices);

                // triangulate destination
                // this is different from source triangulation: we want to achieve constant width triangles
                // first: what is the constant width --> half of the image
                var sliceWidth = sprite.spriteImage.width / 2;

                // how many squares/triangles?
                // the number of squares must be an integer and a multiple of 2
                var arcLength = (objectWidth * gridScale) / planet.perimeter * twoPi * (radius - posY);
                var squares = Math.round(arcLength / sliceWidth);
                var destWidth = (objectWidth * gridScale);
                if (squares % 2) {
                    squares -= 1;
                }
                if (objectWidth === tileWidth) {
                    // force 16 width blocks to be the exact image
                    squares = 2;
                }

                destinationRectangle = new Rectangle(-origin.x - destWidth / 2, -origin.y, destWidth, sprite.frameHeight);
                destinationTriangles = triangulate(destinationRectangle, squares - 1);

                var correctionAmountTop = getCorrectionRatio(radius, false);
                var correctionRatioTop = (objectWidth - correctionAmountTop) / objectWidth;
                var correctionAmountBottom = getCorrectionRatio(radius, objectWidth === tileWidth ? true : false) + (objectWidth === tileWidth ? 0 : -1.2);
                var correctionRatioBottom = (objectWidth - correctionAmountBottom) / objectWidth;
                Utils.forEach(destinationTriangles, function (triangle, i) {
                    triangle.forEach(function (vec2, j) {
                        if (i % 2 === 0) { // upper
                            if (j % 2 === 0) {
                                vec2.x *= correctionRatioTop;
                            } else {
                                vec2.x *= correctionRatioBottom;
                            }
                        } else { // lower
                            if (j < 2) {
                                vec2.x *= correctionRatioBottom;
                            } else {
                                vec2.x *= correctionRatioTop;
                            }
                        }

                    });
                });

                // match the amount of source triangles with destination triangles by cloning the first ones, creating a pattern
                while (sourceTriangles.length < destinationTriangles.length) {
                    sourceTriangles.push(sourceTriangles[sourceTriangles.length % 4].clone());
                }

                /**
                 * Arc correction for destination triangles
                 */
                // move triangles up radius
                Utils.forEach(destinationTriangles, function (triangle, i, l, breakLoop) {
                    triangle.v1.y -= radius;
                    triangle.v2.y -= radius;
                    triangle.v3.y -= radius;

                    if (objectWidth !== tileWidth) {
                        // triangle.v1.x += correctionAmountTop * 3;
                        // triangle.v2.x += correctionAmountTop * 3;
                        // triangle.v3.x += correctionAmountTop * 3;
                    }
                });
                // note: we don't need to go over every triangle
                // convert to polar
                var maxA = null;
                Utils.forEach(destinationTriangles, function (triangle, i, l, breakLoop) {
                    triangle.forEach(function (vec2, j) {
                        var r, a;
                        var corr = 1;
                        var corrAmount = 0;
                        if (i % 2 === 0) { // upper
                            if (j % 2 === 0) {
                                corr = correctionRatioTop;
                                corrAmount = correctionAmountTop;
                            } else {
                                corr = correctionRatioBottom;
                                corrAmount = correctionAmountBottom;
                            }
                        } else { // lower
                            if (j < 2) {
                                corr = correctionRatioBottom;
                                corrAmount = correctionAmountBottom;
                            } else {
                                corr = correctionRatioTop;
                                corrAmount = correctionAmountTop;
                            }
                        }

                        // turn to polar coordinates
                        r = vec2.y + posY;
                        a = (vec2.x) / planet.perimeter * twoPi;
                        a += Math.PI / 2;

                        // correction
                        a += (destWidth / 2) / planet.perimeter * twoPi * corr;
                        if (objectWidth !== tileWidth) {
                            a += (corrAmount * 3) / planet.perimeter * twoPi;
                        }
                        vec2.x = r * Math.cos(a);
                        vec2.y = r * Math.sin(a) - posY;
                    });
                });
                // move triangles down radius
                var minX = Infinity,
                    minY = Infinity,
                    maxX = -Infinity,
                    maxY = -Infinity;
                Utils.forEach(destinationTriangles, function (triangle, i, l, breakLoop) {
                    triangle.v1.y += radius;
                    triangle.v2.y += radius;
                    triangle.v3.y += radius;

                    // since we'er looping here already, find out the size of the triangle
                    triangle.forEach(function (vec2) {
                        if (vec2.x < minX) {
                            minX = vec2.x;
                        }
                        if (vec2.x > maxX) {
                            maxX = vec2.x;
                        }
                        if (vec2.y < minY) {
                            minY = vec2.y;
                        }
                        if (vec2.y > maxY) {
                            maxY = vec2.y;
                        }
                    });
                });
                canvasSize.x = minX;
                canvasSize.y = minY;
                canvasSize.width = maxX - minX;
                canvasSize.height = maxY - minY;
                var overshootX = canvasSize.x + origin.x;
                var overshootY = canvasSize.y + origin.y;

                // draw onto canvas
                canvas = document.createElement('canvas');
                canvas.width = Math.ceil(canvasSize.width);
                canvas.height = Math.ceil(canvasSize.height);
                context = canvas.getContext('2d');

                context.translate(origin.x - overshootX, origin.y - overshootY);
                Utils.forEach(destinationTriangles, function (dest, i, l, breakLoop) {
                    var source = sourceTriangles[i];
                    var image = sprite.spriteImage;

                    // temp: wall overwrite
                    if (i < 2) {
                        image = wallLeft;
                    } else if (i >= destinationTriangles.length - 2) {
                        image = wallRight;
                    }

                    drawTriangle(
                        context,
                        image,
                        dest.v1.x,
                        dest.v1.y,
                        source.v1.x,
                        source.v1.y,
                        dest.v2.x,
                        dest.v2.y,
                        source.v2.x,
                        source.v2.y,
                        dest.v3.x,
                        dest.v3.y,
                        source.v3.x,
                        source.v3.y
                    );
                });

                // swap out image in sprite
                cacheObject = {
                    image: new PackedImage(canvas),
                    origin: origin.add(new Vector2(-overshootX, -overshootY))
                };
                component.swapSprite(cacheObject);

                // add to canvas cache
                canvasCache[canvasId] = cacheObject;
            },
            swapSprite: function (cacheObject) {
                // swap the sprite's image with the canvas cache
                sprite.setup({
                    image: cacheObject.image,
                    origin: cacheObject.origin.clone()
                });
            },
            start: function (data) {
                var cacheObject;
                var height = 0;
                // cache planet
                planetWrap = entity.getComponent('planetWrap');
                planet = planetWrap.planet;
                if (!sprite) {
                    sprite = entity.getComponent('sprite');
                    if (!sprite) {
                        Utils.log("Entity doesn't have sprite.");
                        return;
                    }
                }

                if (objectWidth === tileWidth) {
                    // don't care about height
                    height = 0;
                } else {
                    height = entity.position.y;
                }
                canvasId = imagePattern + '|' + objectWidth + '|' + height;
                cacheObject = canvasCache[canvasId];
                if (!cacheObject) {
                    component.calculateArc();
                } else {
                    component.swapSprite(cacheObject);
                }
            },
            attached: function (data) {
                entity = data.entity;
            }
        };
        return component;
    };
});