/**
 * A poppable balloon with confetti :D
 * @moduleName Balloon
 */
bento.define('entities/balloon', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'globals',
    'bento/components/eventlistener',
    'components/movable',
    'components/planetwrap',
    'bento/components/spine',
    'entities/effect'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Globals,
    EventListener,
    Movable,
    PlanetWrap,
    Spine,
    Effect
) {
    'use strict';

    var colors = ['blue','green','orange','yellow','purple','red'];

    // generate table of static anims, named by color
    // to be used by all balloons
    var i, animations = {};
    for (i=0; i < colors.length; i++) {
        animations[colors[i]] = {
            speed: 0,
            frames: [i]
        };
    }

    var BalloonPop = function (settings) {
        var sprite = new Sprite({
            imageName: 'balloon-pop',
            originRelative: new Vector2(0.5, 0.5)
        });
        var planet = settings.planet || {};
        var velocity = settings.velocity || new Vector2(0, 0);
        var planetWrap = new PlanetWrap({
            planet: planet
        });
        var behavior = {
            start: function () {
                new Tween({
                    from: 0.8, to: 1.2,
                    in: 6,
                    onUpdate: function (v, t) {
                        entity.scale.x = entity.scale.y = v;
                    },
                });
            },
            update: function (data) {
                entity.alpha -= data.speed * 0.13;
                if (entity.alpha < 0) {
                    entity.alpha = 0;
                    entity.removeSelf();
                }
            }
        };
        var entity = new Entity({
            z: Globals.Layers.PARTICLES,
            name: 'balloonPop',
            position: settings.position,
            scale: new Vector2(0,0),
            components: [
                sprite,
                behavior,
                planetWrap
            ]
        });
        return entity;
    };

    var Confetti = function (settings) {
        var sprite = new Sprite({
            imageName: 'confettis',
            originRelative: new Vector2(0.5, 0.5),
            frameCountX: 3,
            animations: {
                default: { frames: [Utils.getRandom(3)] }
            }
        });
        var planet = settings.planet || {};
        var planetWrap = new PlanetWrap({
            planet: planet
        });
        var initialVelocity = new Vector2(
            Utils.getRandomRangeFloat(-10, 10),
            Utils.getRandomRangeFloat(0, -20)
        );
        var velocity = new Vector2(0, Utils.getRandomRangeFloat(6, 8));
        var velocityLerp = 0.0;                                 // we start off moving in some initial direction then gradually switch to swaying.
        var swaySpeed = Utils.getRandomRangeFloat(0.1, 0.3);    // rate at which we sway
        var sway = Utils.getRandomFloat(Math.PI*2);             // directly corrseponds to velocity
        var bias = Utils.getRandomRangeFloat(-2, 2);            // but we might drift slightly to one side over time

        var behavior = {
            start: function () {},
            update: function (data) {
                sway += swaySpeed;
                velocity.x = Math.sin(sway) * 2 + bias;
                velocityLerp = Utils.approach(velocityLerp, 1.0, data.speed * 0.03);
                
                entity.position.addTo(initialVelocity.scalarMultiply(1.0 - velocityLerp));
                entity.position.addTo(velocity.scalarMultiply(velocityLerp));

                entity.alpha -= data.speed * 0.01;
                if (entity.alpha < 0) {
                    entity.removeSelf();
                }
            }
        };
        var entity = new Entity({
            z: Globals.Layers.PARTICLES,
            name: 'balloonConfetti',
            position: settings.position,
            scale: new Vector2(0.2, 0.2),
            rotation: Utils.getRandomFloat(Math.PI*2),
            components: [
                sprite,
                behavior,
                planetWrap
            ]
        });
        return entity;
    };

    return function (settings) {
        var viewport = Bento.getViewport();
        
        var color = Utils.getObjectProperty(settings, 'color', colors[Utils.getRandom(colors.length)]);

        var sprite = new Sprite({
            imageName: 'balloon',
            originRelative: new Vector2(0.5, 1.0),
            frameCountX: 3,
            frameCountY: 2,
            animations: animations
        });
        sprite.setAnimation(color);

        var size = Utils.getTileProperty(settings, 'size', 128);
        var vulnerableOffset = Vector2(0, -2);
        var flightTimer = 0;
        var flightUpwardsVel = 0;

        var direction = 'left';

        var physics = new Movable({
            velocity: new Vector2(0, 0),
            gravity: 0,
            bounce: new Vector2(1, 0),
            maxVel: new Vector2(63, 63)
        });
        var behavior = {
            name: 'behaviorComponent',
            player: null,
            getPlayer: function () {
                Bento.objects.get('player', function (player) {
                    behavior.player = player;
                });
            },
            start: function (data) {
                entity.position.y -= size * 0.25;
                behavior.getPlayer();
            },
            destroy: function (data) {},
            update: function (data) {
                if (!behavior.player) {
                    // no reference to player
                    behavior.getPlayer();
                    return;
                }

                // check if we collide with the player
                if (behavior.player.isAlive() && Utils.samePlanet(behavior.player, entity)) {
                    // check if the player hit our head
                    entity.collidesWith({
                        entity: behavior.player,
                        offset: vulnerableOffset,
                        onCollide: function () {
                            behavior.player.bounceUp();
                            die();
                        }
                    });
                }

                // bobbing up and down
                flightTimer += Bento.getGameSpeed();
                flightUpwardsVel = Math.sin(flightTimer / 30);
                physics.velocity.y = flightUpwardsVel;
            },
            draw: function (data) {}
        };
        
        var entity = new Entity({
            z: Globals.Layers.ITEMS,
            name: 'balloon',
            scale: Vector2(size / 128, size / 128),
            boundingBox: new Rectangle(-50, -160, 100, 100),
            components: [
                sprite,
                behavior,
                physics
            ]
        });

        var die = function () {
            Bento.audio.stopSound('sfx_balloon-pop');
            Bento.audio.playSound('sfx_balloon-pop');
            
            entity.getComponent('planetWrap', function (planetWrap) {
                Bento.objects.attach(new BalloonPop({
                    position: entity.position.add(new Vector2(0, -170)),
                    planet: planetWrap.planet
                }));
                Utils.repeat(10, function () {
                    Bento.objects.attach(new Confetti({
                        position: entity.position.add(new Vector2(
                            Utils.getRandomRange(-20, 20),
                            Utils.getRandomRange(-100, -180)
                        )),
                        planet: planetWrap.planet
                    }));
                });
            });

            entity.removeSelf();
        };

        return entity;
    };
});