/**
 * CanvasBackground is canvas containing all the backgrounds drawn on it. The canvas will
 * be 2048 x 2048, with the planet in the middle. Tiled objects with tileset property "background" set to true
 * will be added to this canvas and no longer draw on the screen. Each planet should have its own CanvasBackground
 * @moduleName CanvasBackground
 * @snippet CanvasBackground.snippet
CanvasBackground({})
 */
bento.define('entities/canvasbackground', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'globals',
    'bento/renderers/canvas2d',
    'bento/packedimage'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Globals,
    Canvas2DRenderer,
    PackedImage
) {
    'use strict';
    var CanvasBackground = function (settings) {
        var viewport = Bento.getViewport();
        var position = settings.position;
        var sprite = new Sprite({});
        var canvas;
        var context;
        var renderer;
        var backgrounds = [];
        var size = CanvasBackground.size;
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {},
            destroy: function (data) {
                if (canvas && canvas.dispose) {
                    canvas.dispose();
                    canvas = null;
                    renderer = null;
                    context = null;
                }
            },
            update: function (data) {},
            draw: function (data) {}
        };
        var initializeCanvas = function () {
            if (canvas) {
                return renderer;
            }
            canvas = document.createElement('canvas');
            canvas.width = size;
            canvas.height = size;
            context = canvas.getContext('2d');
            renderer = new Canvas2DRenderer(canvas, {
                pixelSize: settings.pixelSize || 1
            });
            // context.fillRect(0, 0, size, size);
            return renderer;
        };
        /**
         * The main drawing of all backgrounds will happen in here
         */
        var drawToCanvas = function (planet) {
            var data = {
                viewport: new Rectangle(0, 0, size, size),
                entity: null,
                renderer: initializeCanvas()
            };
            renderer.begin();
            Utils.forEach(backgrounds, function (background, i, l, breakLoop) {
                data.entity = background;
                background.draw(data);
                background.visible = false;
            });
            renderer.flush();

            sprite.setup({
                image: new PackedImage(canvas),
                originRelative: new Vector2(0.5, 0.5)
            });
        };
        var entity = new Entity({
            z: Globals.Layers.BACKGROUND,
            name: 'canvasBackground',
            components: [
                sprite,
                behavior
            ]
        }).extend({
            backgrounds: backgrounds,
            drawToCanvas: drawToCanvas
        });

        return entity;
    };
    CanvasBackground.size = 2048;
    return CanvasBackground;
});