/**
 * Fire events to the rhythm
 * @moduleName MusicManager
 * @snippet MusicManager.snippet
MusicManager({
    position: new Vector2(0, 0)
})
 */
bento.define('entities/musicmanager', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'globals',
    'bento/components/eventlistener'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Globals,
    EventListener
) {
    'use strict';
    return function (settings) {

        var visible = settings.visible || false;

        var song;
        var currentTime = 0;
        var waitingForSongLoop = false;
        var lastKnownTime = 0;
        var bpm = 127;
        var timePerBeat = 60 * 1/bpm;
        var nextBeatTime = 0;
        var latency = Utils.isCocoonJs() ? 0.38 : 0.1;

        var useMicroseconds = false;

        var getCurrentTime = function () {
            return useMicroseconds ? song.currentTime/1000000 : song.currentTime;
        };

        var sprite = new Sprite({
            imageName: 'collectible',
            originRelative: Vector2(0.5, 0.5)
        });

        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                song = Bento.assets.getAudio('bgm_music-long');
                Bento.audio.playMusic('bgm_music-long');
            },
            destroy: function (data) {
                song.stop();
            },
            update: function (data) {
                // song stopped? pretend the music is still going
                if (song.paused) {
                    currentTime = ((Date.now() / 1000) % 45.72) + latency;
                } else {
                    currentTime = getCurrentTime() + latency;
                }

                // if nextBeatTime is at the beginning of the song after looping
                // we should wait for currentTime to move backwards
                // before we can check whether we have exceeded nextBeatTime
                if (waitingForSongLoop) {
                    if (currentTime >= lastKnownTime) {
                        return;
                    } else {
                        lastKnownTime = 0;
                        waitingForSongLoop = false;
                    }
                } else {
                    // correct some really weird behaviour with certain MP3s on Cordova
                    if (!useMicroseconds && currentTime - lastKnownTime > 100) {
                        useMicroseconds = true;
                    }
                }

                // if we exceed nextBeatTime, fire a beat event
                if (currentTime >= nextBeatTime) {
                    EventSystem.fire('songBeat');
                    if (visible) {
                        new Tween({
                            from: 1,
                            to: 0,
                            in: 60,
                            onUpdate: function(value, time) {
                                entity.scale.x = entity.scale.y = value*2;
                                entity.alpha = value;
                            }
                        });
                    }

                    nextBeatTime += timePerBeat;
                    if (nextBeatTime >= song.duration) {
                        nextBeatTime = 0;
                        waitingForSongLoop = true;
                    }
                }

                lastKnownTime = currentTime;
            }
        };
        var eventListener = new EventListener({
            name: 'eventListener',
            eventName: 'onMute',
            ignorePause: false,
            onEvent: function (data) {
                // reset next beat, just let it catch up
                nextBeatTime = 0;
            }
        });
        var entity = new Entity({
            z: Globals.Layers.DEBUG,
            name: 'musicManager',
            position: Vector2(200, 200),
            updateWhenPaused: 1,
            float: true,
            global: true,
            visible: visible,
            components: [
                sprite,
                behavior,
                eventListener
            ]
        });

        entity.visible = visible;

        return entity;
    };
});