/**
 * One-way platforms.
 * note: Any entity which cares about whether or not these are solid, should
 *       fire the 'prepareOneWayPlatforms' event before checking for any collisions.
 * @moduleName OneWayPlatform
 */
bento.define('entities/onewayplatform', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'components/debug',
    'components/planetwrap',
    'globals',
    'components/trianglerenderer',
    'bento/components/eventlistener'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Debug,
    PlanetWrap,
    Globals,
    TriangleRenderer,
    EventListener
) {
    'use strict';
    return function (settings) {
        var viewport = Bento.getViewport();
        var size = Utils.getTileProperty(settings, 'size', 128);

        // all these must be true for this platform to become solid.
        var planetActive = true;
        var active = true;

        var sprite = new Sprite({
            imageName: 'oneway-middle',
            originRelative: new Vector2(0.5, 0),
        });

        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                entity.getComponent('planetWrap', function (planetWrap) {
                    planetWrap.setWall(true);
                });

                // make boundingBox match Tiled size
                var w = Utils.getTiledObject(settings).width * 128/16;
                var h = 128;
                entity.boundingBox = new Rectangle(-128*0.5, 0, w, h);
                entity.position.y -= 128;

                EventSystem.on('prepareOneWayPlatforms', onPrepareOneWayPlatforms);
                EventSystem.on('setActivePlanet', onSetActivePlanet);
            },
            destroy: function (data) {
                EventSystem.off('prepareOneWayPlatforms', onPrepareOneWayPlatforms);
                EventSystem.off('setActivePlanet', onSetActivePlanet);
            },
            update: function (data) {},
            draw: function (data) {}
        };

        var onSetActivePlanet = function (newPlanet) {
            entity.getComponent('planetWrap', function (wrap) {
                if (newPlanet === wrap.planet) {
                    planetActive = true;
                } else {
                    planetActive = false;
                }
            });
        };

        var onPrepareOneWayPlatforms = function (other) {
            // we are only active if the entity's feet are above us, and the entity is moving down
            var movable = other.getComponent('movable');
            active = movable
                  && movable.velocity.y >= 0
                  && other.position.y <= entity.position.y;

            // we should only count as solid if we are active and our planet is also active
            if (active && planetActive) {
                if (!entity.isFamily('solid')) {
                    entity.addToFamily('solid');
                }

                // testing sticky platforms
                // if (other.collidesWith({ entity:entity, offset: Vector2(0, 1) })) {
                //     movable.velocity.x *= 0.85;
                // }

            } else {
                if (entity.isFamily('solid')) {
                    entity.removeFromFamily('solid');
                }
            }
        };

        var triangleRenderer = new TriangleRenderer({
            sprite: sprite,
            imagePattern: 'oneway',
            imageSize: Utils.getTileProperty(settings, 'size', 128),
            objectWidth: Utils.getTiledObject(settings).width,
            objectHeight: Utils.getTiledObject(settings).height,
            tileWidth: ((settings.tiled || {}).tileSet || {}).tilewidth || 16,
            tileHeight: ((settings.tiled || {}).tileSet || {}).tileheight || 16,
        });

        var entity = new Entity({
            z: Globals.Layers.ONE_WAY_PLATFORMS,
            name: 'oneWayPlatform',
            family: [],
            updateWhenPaused: 0,
            float: false,
            scale: Vector2(size / 128, size / 128),
            components: [
                behavior,
                triangleRenderer,
                sprite
            ]
        });

        return entity;
    };
});