/**
 * It's the player!
 * @moduleName Player
 * @snippet Player.snippet
Player({
    position: new Vector2(0, 0)
})
 */
bento.define('entities/player', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'utils',
    'bento/tween',
    'components/physics',
    'components/planetwrap',
    'components/squisher',
    'globals',
    'components/modal',
    'components/movable',
    'bento/components/spine',
    'entities/particle',
    'bento/components/clickable',
    'entities/effect',
    'entities/glowparticle',
    'entities/expbonustext',
    'gui/gametext',
    'managers/save',
    'modules/tapticengine'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Physics,
    PlanetWrap,
    Squisher,
    Globals,
    Modal,
    Movable,
    Spine,
    Particle,
    Clickable,
    Effect,
    GlowParticle,
    ExpBonusText,
    GameText,
    SaveManager,
    TapticEngine
) {
    'use strict';
    return function (settings) {
        var viewport = Bento.getViewport();
        var position = settings.position;
        var chunkLoader;
        var isBossFight = false;
        var planet = settings.planet;
        var size = settings.size || 128;
        var onFloor = true;
        var timeSinceOnFloor = 0; // for coyote time
        var timeSinceLastPress = 0; // for input buffering
        var landTimer = 100;
        var highestVel = -10;
        var trickTimer = 0;
        var trickTimerTarget = 20;
        var trickJumpCombo = -1;
        var trickJumpComboTotal = 0;
        var rollApplauseTween;
        var started = false;
        var dead = false;
        var container = null; // portal, rocket, or other entity that may hold the player and prevent movement
        var score = 0;
        var flip = planet && planet.flip;
        var variableJump = true;
        var canDoubleJump = false;
        var forcedJump = false; // while this is set, variable jump will be ignored until the next landing
        var trailTimer = 10;
        var acceleration = new Vector2(0, 0);
        var drag = new Vector2(0, 0);

        var spineOffset = new Vector2(-20, 0);
        var jumpScale = new Vector2(1, 1);
        var fishRotation = 0;
        var jumpScaleTween;

        // cached values when dying
        var endBossFight;
        var cachedGameSpeed = 1;

        var sprite = new Spine({
            spine: 'redfish',
            skin: SaveManager.load('skin'),
            animation: 'idle',
            scale: 1.1,
            triangleRendering: false
        });

        var squisher = new Squisher({
            origin: new Vector2(0, 50)
        });

        var invincibleTimer = 0;

        var onSongBeat = function () {
            if (container || !onFloor) {
                return;
            }
            // squisher.playBounceAnimation();
            // TODO: sync Spine animation to the beat.
        };
        var onPress = function (data) {
            // reserve the top screen for the pause button
            if (data && data.position.y < 256) {
                return;
            }
            timeSinceLastPress = 0;
            isHolding = true;
        };
        var onRelease = function () {
            if (physics.velocity.y < 0 && variableJump && !forcedJump) {
                physics.velocity.y *= 0.3;
            }
            isHolding = false;
        };
        var clickable = new Clickable({
            pointerDown: onPress,
            pointerUp: onRelease
        });
        var walkDustTimer = 0;
        var spawnJumpParticle = function () {
            var velocity = new Vector2(0, 1);
            var particle = new Particle({
                position: entity.position.clone(),
                planet: planet,
                spriteSheet: 'effects/poof',
                velocity: velocity,
                shouldScale: false,
                scale: 1 - 0.75 * trailTimer / 10 + Utils.getRandomFloat(0.2)
            });
            particle.rotation = Utils.getRandomFloat(Math.PI);
            velocity.rotateRadian(Utils.getRandomRangeFloat(-35, 35));
            Bento.objects.attach(particle);
        };
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                EventSystem.on('songBeat', onSongBeat);
                if (planet) {
                    EventSystem.on('startLevel', start);
                    EventSystem.on('playerInPortal', onPortal);
                    EventSystem.on('playerInRocket', onRocket);
                    EventSystem.on('collectStar', onCollectStar);
                    EventSystem.on('killPlayer', die);
                    EventSystem.on('setDirection', entity.setDirection);
                    if (Utils.isDev()) {
                        EventSystem.on('buttonDown-up', onPress);
                        EventSystem.on('buttonUp-up', onRelease);
                        EventSystem.on('buttonDown-x', onPress);
                        EventSystem.on('buttonUp-x', onRelease);
                    }
                    entity.setSize(planet.playerSize, planet.flip);
                }
                chunkLoader = Bento.objects.get('chunkLoader');
            },
            destroy: function (data) {
                EventSystem.off('songBeat', onSongBeat);
                EventSystem.off('startLevel', start);
                EventSystem.off('playerInPortal', onPortal);
                EventSystem.off('playerInRocket', onRocket);
                EventSystem.off('collectStar', onCollectStar);
                EventSystem.off('killPlayer', die);
                EventSystem.off('setDirection', entity.setDirection);
                if (Utils.isDev()) {
                    EventSystem.off('buttonDown-up', onPress);
                    EventSystem.off('buttonUp-up', onRelease);
                    EventSystem.off('buttonDown-x', onPress);
                    EventSystem.off('buttonUp-x', onRelease);
                }
            },
            update: function (data) {
                if (chunkLoader) {
                    isBossFight = chunkLoader.isBossFight;
                } else {
                    chunkLoader = Bento.objects.get('chunkLoader');
                }
                if (dead) {
                    // fishRotation += 0.1 * Math.PI;
                    // updateVelocity();
                    // return;
                }
                if (invincibleTimer > 0) {
                    invincibleTimer -= 1;
                    entity.alpha = invincibleTimer % 2;
                } else if (!dead && !entity.alpha) {
                    entity.alpha = 1;
                }
                entity.collidesWith({
                    family: 'deadly',
                    onCollide: function (other) {
                        if (Utils.samePlanet(entity, other)) {
                            die();
                        }
                    }
                });

                if (planet) {
                    onFloor = false;

                    EventSystem.fire('prepareOneWayPlatforms', entity);

                    entity.collidesWith({
                        family: 'solid',
                        offset: new Vector2(0, 1),
                        onCollide: function (other) {
                            if (Utils.samePlanet(entity, other)) {
                                timeSinceOnFloor = 0;
                                onFloor = true;
                            }
                        }
                    });

                    // var offset = new Vector2(flip ? -1 : 1, 0);
                    var offset = new Vector2(physics.velocity.x, 0);
                    entity.collidesWith({
                        family: 'solid',
                        offset: offset,
                        onCollide: function (other) {
                            if (Utils.samePlanet(entity, other)) {
                                var forgiveness = 40;
                                do {
                                    offset.y -= 2;
                                    if (offset.y <= -forgiveness) {
                                        die();
                                        return;
                                    }
                                } while (entity.collidesWith({
                                        family: 'solid',
                                        offset: offset
                                    }));
                                physics.velocity.y = Math.min(physics.velocity.y, 0);
                                physics.velocity.x = physics.maxVelocity.x;
                                entity.position.y += offset.y;
                            }
                        }
                    });
                    if (timeSinceLastPress <= Globals.JUMP_BUFFER && physics.velocity.y >= 0) {
                        jump();
                    }
                }

                if (!onFloor) {
                    if (isHolding && physics.velocity.y < 0) {
                        trickTimer += data.speed;
                    }
                    highestVel = physics.velocity.y;
                    if (physics.velocity.y < 0) {
                        sprite.setAnimation('rise', function () {}, true);
                        if (jumpScaleTween) {
                            jumpScaleTween.removeSelf();
                            jumpScaleTween = null;
                        }
                        jumpScaleTween = new Tween({
                            from: 0.25,
                            to: 0,
                            in: 10,
                            ease: 'elastic',
                            decay: 2,
                            oscillations: 1,
                            onUpdate: function (v, t) {
                                jumpScale.x = 1 - v;
                                jumpScale.y = 1 + v;
                            },
                            onComplete: function () {

                            }
                        });
                    } else if (physics.velocity.y > 0) {
                        if (sprite.getAnimationName() !== 'fall' && sprite.getAnimationName() !== 'apex') {
                            sprite.setAnimation('apex', function () {
                                sprite.setAnimation('fall', function () {});
                            });
                        }

                        // stretch out
                        if (jumpScaleTween) {
                            jumpScaleTween.removeSelf();
                            jumpScaleTween = null;
                        }
                        jumpScale.y = Utils.approach(jumpScale.y, 1, 0.25 / 30);
                        jumpScale.x = Utils.approach(jumpScale.x, 1, 0.25 / 30);
                    }
                    trailTimer += 1;

                    // spawn particles DEPRECATED
                    if (trailTimer < 10 && trailTimer % 4 === 0) {
                        // spawnJumpParticle();
                    }

                    if (landTimer <= 10) {
                        // player dropped while still able to trickjump
                        landTimer = 0;
                        if (rollApplauseTween) {
                            rollApplauseTween.removeSelf();
                            rollApplauseTween = null;
                        }
                    }
                } else {
                    if (settings.sad) {
                        sprite.setAnimation('sad', function () {}, true);
                    } else {
                        sprite.setAnimation('idle', function () {}, true);
                    }

                    // spawn walking dust
                    if (Math.abs(physics.velocity.x) > Utils.EPSILON && walkDustTimer > 20) {
                        var walkDust = new Effect({
                            z: entity.z - 1,
                            position: entity.position.add(new Vector2(-32, 0)),
                            spriteSheet: 'effects/run',
                            components: [
                                new PlanetWrap({
                                    planet: planet
                                })
                            ]
                        });
                        walkDustTimer = 0;
                        Bento.objects.attach(walkDust);
                    }
                    walkDustTimer += data.speed;

                    // cancel any combo of trickjumps
                    landTimer += data.speed;
                    if (landTimer > 10) {
                        if (trickJumpCombo !== 0) {
                            trickJumpCombo = 0;
                            trickJumpComboTotal = 0;
                            EventSystem.fire('comboEnded');
                        }
                    }

                    // bugfix: make sure fishrotation goes to zero on ground
                    if (!dead) {
                        fishRotation = 0;
                    }

                    forcedJump = false;
                }

                updateVelocity();
                timeSinceOnFloor += data.speed;
                timeSinceLastPress += data.speed;
            },
            draw: function (data) {
                var renderer = data.renderer;

                // offset spine graphic
                renderer.translate(spineOffset.x, spineOffset.y);

                // additional scales
                renderer.scale(jumpScale.x, jumpScale.y);

                // additional rotation
                if (fishRotation) {
                    renderer.translate(0, -64);
                    renderer.rotate(fishRotation);
                    renderer.translate(0, 64);
                }
            }
        };

        var physics = new Movable({
            velocity: new Vector2(0, 0),
            gravity: Globals.GRAVITY,
            bounce: new Vector2(-1, 0), // prevent speed from stopping
            maxVelocity: new Vector2(Globals.SPEED, 63),
            onCollide: function (data) {
                var pushDirection = data.pushDirection;
                var pushAmount = data.pushAmount;
                if (pushDirection === 'y' && pushAmount < 0 && !dead) {
                    if (!jumpScaleTween) {
                        if (trickTimer > trickTimerTarget) {
                            if (jumpScaleTween) {
                                jumpScaleTween.removeSelf();
                                jumpScaleTween = null;
                            }
                            jumpScaleTween = new Tween({
                                from: 0.5,
                                to: 0,
                                in: 40,
                                ease: 'elastic',
                                decay: 2,
                                oscillations: 1,
                                onStart: function () {
                                    Bento.audio.playSound('sfx_land');

                                    // add landing effect
                                    var effect = new Effect({
                                        z: entity.z - 1,
                                        position: entity.position.clone(),
                                        spriteSheet: 'effects/land',
                                        components: [
                                            new PlanetWrap({
                                                planet: planet
                                            })
                                        ]
                                    });
                                    Bento.objects.attach(effect);
                                },
                                onUpdate: function (v, t) {
                                    jumpScale.x = 1 + v;
                                    jumpScale.y = 1 - v;
                                },
                                onComplete: function () {

                                }
                            });
                            var particle = new GlowParticle({
                                position: entity.position.add(new Vector2(-16, -88)),
                                planet: planet
                            });
                            particle.attach({
                                name: 'followPlayer',
                                update: function (dataParticle) {
                                    dataParticle.entity.position.x = entity.position.x - 16;
                                    dataParticle.entity.position.y = entity.position.y - 88;
                                }
                            });
                            Bento.objects.attach(particle);

                            TapticEngine.selection(); // selection is softer than light impact
                        }
                        landTimer = 0;

                        // schedule applause after landing with 2 trickJumpCombo
                        // will be cancelled if player makes a new roll
                        if (trickJumpComboTotal >= 3) {
                            if (rollApplauseTween) {
                                rollApplauseTween.stop();
                                rollApplauseTween.removeSelf();
                                rollApplauseTween = null;
                            }
                            var currentCombo = trickJumpComboTotal; // cache combo amount
                            rollApplauseTween = new Tween({
                                delay: 10,
                                onComplete: function () {
                                    Bento.audio.stopSound('sfx_applause-short');
                                    Bento.audio.playSound('sfx_applause-short');

                                    // create combo text
                                    var comboText = new GameText({
                                        z: Globals.Layers.HUD,
                                        position: new Vector2(viewport.width * 0.5, 480),
                                        text: currentCombo + 'x combo!',
                                        fontSize: 100,
                                        float: true
                                    });
                                    new Tween({
                                        from: comboText.position.x + viewport.width,
                                        to: comboText.position.x,
                                        in: 30,
                                        ease: 'elastic',
                                        decay: 6,
                                        oscillations: 1,
                                        onStart: function () {
                                            Bento.objects.attach(comboText);
                                        },
                                        onUpdate: function (v, t) {
                                            comboText.position.x = v;
                                        },
                                        onComplete: function () {
                                            new Tween({
                                                from: 1,
                                                to: 0,
                                                in: 10,
                                                delay: 30,
                                                ease: 'linear',
                                                onUpdate: function (v, t) {
                                                    comboText.alpha = v;
                                                },
                                                onComplete: function () {
                                                    comboText.removeSelf();
                                                }
                                            });
                                        }
                                    });
                                }
                            });

                        }
                    }
                    onFloor = true;

                    if (boostDelay && boostDelay.isAdded) {
                        // cancel any enemy boost
                        boostDelay.removeSelf();
                        boostDelay = null;
                    }
                }
            }
        });

        // apply acceleration or drag
        var updateVelocity = function () {
            if (acceleration.x === 0) {
                physics.velocity.x = Utils.approach(physics.velocity.x, 0, drag.x);
            } else {
                physics.velocity.x = Utils.approach(physics.velocity.x, flip ? -Globals.SPEED : Globals.SPEED, acceleration.x);
            }
        };

        var entity = new Entity({
            z: Globals.Layers.PLAYER,
            name: 'player',
            position: position,
            components: [
                squisher,
                behavior,
                clickable,
                sprite,
            ]
        }).extend({
            setSkin: function (skinName) {
                // swap out spine component
                entity.removeByName('spine');
                entity.attach(new Spine({
                    spine: 'redfish',
                    skin: skinName,
                    animation: 'idle',
                    scale: 1.1
                }));
            }
        });

        if (settings.planet === null) {
            // this player just dances but can't be controlled
            return entity;
        }

        var planetWrap = new PlanetWrap({
            planet: planet
        });
        var hasDoubleJumped = false;
        var isHolding = false;
        var spawnParticles = function () {
            /*var spriteEffect = new Sprite({
                imageName: 'effects/jump',
                originRelative: new Vector2(0.5, 1),
                frameCountX: 2,
                frameCountY: 4,
                animations: {
                    effect: {
                        speed: 0.2,
                        frames: [0, 1, 2, 3, 4, 5, 6, 7]
                    }
                }
            });
            var effect = new Entity({
                z: Globals.Layers.PARTICLES,
                name: 'effect',
                position: entity.position.clone(),
                components: [
                    new PlanetWrap({
                        planet: planetWrap.planet
                    }),
                    spriteEffect
                ]
            });
            spriteEffect.setAnimation('effect', function () {
                effect.removeSelf();
            });
            Bento.objects.attach(effect);*/
        };

        entity.attach(planetWrap);
        entity.attach(physics);
        var doTrickJump = function () {
            // rotation (should be done with spine animation?)
            new Tween({
                from: fishRotation,
                to: fishRotation + (Math.PI * 2) * (flip ? -1 : 1),
                in: 40,
                ease: 'exponential',
                decay: 5,
                oscillations: 1,
                onUpdate: function (v, t) {
                    fishRotation = v;
                    fishRotation %= Math.PI * 2;
                }
            });
            // play higher pitches of jump sound
            Bento.audio.stopSound('sfx_jump' + ((trickJumpCombo % 3) + 1));
            Bento.audio.playSound('sfx_jump' + ((trickJumpCombo % 3) + 1));

            trickJumpCombo += 1;
            trickJumpComboTotal += 1;
            var trickJumpScore = Math.round(trickJumpCombo * Globals.TRICK_JUMP_BASE_SCORE * Globals.extraTrickJumpMultiplier);

            // show some kind of feedback?
            if (!isBossFight) {
                Bento.objects.attach(new ExpBonusText({
                    planet: planetWrap.planet,
                    posInPlanet: entity.position.add(Vector2(0, -128 * 3)),
                    value: trickJumpScore
                }));
                EventSystem.fire('trickJump', {
                    combo: trickJumpComboTotal,
                    value: trickJumpScore
                });
            }

            if (trickJumpCombo === 3) {
                /* applause no longer played on 3rd jump, instead it's scheduled when the player ends their combo */
                // Utils.timeout(5, function () {
                //     Bento.audio.stopSound('sfx_applause-short');
                //     Bento.audio.playSound('sfx_applause-short');
                // });
                trickJumpCombo = 2;
            }

            // cancel any scheduled applause
            if (rollApplauseTween) {
                rollApplauseTween.stop();
                rollApplauseTween.removeSelf();
                rollApplauseTween = null;
            }
        };
        var jump = function () {
            if (dead || container || !started) {
                return;
            }
            if (entity.canJump()) {
                onFloor = false;
                hasDoubleJumped = false;
                spawnParticles();

                if (landTimer < 10 && trickJumpCombo !== -1 && trickTimer > trickTimerTarget) {
                    doTrickJump();
                } else {
                    if (trickJumpComboTotal > 0) {
                        EventSystem.fire('comboEnded');
                    }
                    trickJumpCombo = -1;
                    trickJumpComboTotal = 0;
                    Bento.audio.playSound('sfx_jump');
                }

                trickTimer = 0;
                if (trickJumpCombo === -1) {
                    trickJumpCombo = 0;
                    trickJumpComboTotal = 0;
                }
                // remove any glow
                Bento.objects.getByFamily('glow', function (array) {
                    // removeSelf modifies array, so we need to loop over a copy instead.
                    var arrayCopy = array.slice();
                    Utils.forEach(arrayCopy, function (item, i, l, breakLoop) {
                        item.removeSelf();
                    });
                });

                physics.velocity.y = -Globals.JUMP_POWER - Math.min(3, trickJumpComboTotal * 1.4);

                // prevent accidental double jump due to input buffering
                timeSinceLastPress += Globals.JUMP_BUFFER;

            } else if (!hasDoubleJumped && canDoubleJump) {
                hasDoubleJumped = true;
                physics.velocity.y = -Globals.JUMP_POWER;
                Bento.audio.playSound('sfx_jump');
                spawnParticles();
                new Tween({
                    from: fishRotation,
                    to: fishRotation + Math.PI * 2 * (flip ? -1 : 1),
                    in: 20,
                    ease: 'exponential',
                    decay: 5,
                    oscillations: 1,
                    onUpdate: function (v, t) {
                        fishRotation = v;
                        fishRotation %= Math.PI * 2;
                    }
                });
            }
            squisher.offset(0, 0, 'linear');

            trailTimer = 0;
        };
        var die = function () {
            if (Globals.INVINCIBLE || invincibleTimer || dead || container) {
                return;
            }
            dead = true;
            physics.velocity.x = 0;
            acceleration.x = 0;
            // physics.velocity.y = -20;
            // drag.x = 0.1;
            // acceleration.x = 0;
            EventSystem.fire('playerDead');
            Bento.audio.playSound('sfx_hit');

            TapticEngine.impact({
                type: "medium"
            });

            var shake = {
                start: function () {
                    if (jumpScaleTween) {
                        jumpScaleTween.removeSelf();
                        jumpScaleTween = null;
                    }
                    jumpScale.x = 1;
                    jumpScale.y = 1;
                    new Tween({
                        from: 20,
                        to: 0,
                        in: 20,
                        ease: 'linear',
                        onStart: function () {
                            sprite.setAnimation('death');
                            // entity.remove(squisher);
                        },
                        onUpdate: function (v, t) {
                            // sprite is paused, force update
                            sprite.update({
                                speed: 1,
                                deltaT: 0.16
                            });
                            entity.transform.x = Utils.getRandomRangeFloat(-v, v);
                            entity.transform.y = Utils.getRandomRangeFloat(-v, v);
                        },
                        onComplete: function () {
                            entity.transform.x = 0;
                            entity.transform.y = 0;
                            freeze.removeSelf();
                            Bento.setGameSpeed(1);
                        }
                    });
                },
                destroy: function () {
                    entity.visible = false;
                    // throw dead fish
                    throwDeadFish();
                }
            };
            var freeze = new Entity({
                components: [
                    new Modal({}),
                    shake
                ]
            });
            Bento.objects.attach(freeze);

            // remove boss event listener asap
            endBossFight = Bento.objects.get('endBossFight');
            Bento.objects.remove(endBossFight);

            // cache gamespeed
            cachedGameSpeed = Bento.getGameSpeed();
        };
        var throwDeadFish = function () {
            var deadBehavior = {
                name: 'dead',
                update: function (data) {
                    deadFish.rotation += 0.1 * Math.PI * data.speed;
                    deadMovable.velocity.x = Utils.approach(deadMovable.velocity.x, 0, 0.1);

                    if (deadFish.position.y > viewport.height * 2) {
                        deadFish.removeSelf();
                    }
                },
                draw: function (data) {
                    data.renderer.translate(0, 64);
                }
            };
            var deadFishSprite = new Spine({
                spine: 'redfish',
                skin: SaveManager.load('skin'),
                animation: 'death',
                scale: 1.1,
                triangleRendering: false
            });
            var deadMovable = new Movable({
                ignoreSolids: true,
                gravity: Globals.GRAVITY,
                velocity: new Vector2(Globals.SPEED, -20)
            });
            var deadFish = new Entity({
                z: entity.z,
                name: 'deadFish',
                position: entity.position.add(new Vector2(0, -64)),
                components: [
                    deadBehavior,
                    deadFishSprite,
                    deadMovable,
                    new PlanetWrap({
                        planet: planet
                    })
                ]
            });
            Bento.objects.attach(deadFish);
        };
        var onPortal = function (portal) {
            if (!dead && container !== portal) {
                container = portal;
                drag.x = 0.2;
                acceleration.x = 0;
                physics.velocity.y = 0;
                squisher.offset(0, 0, 'linear');
                squisher.scale(0, 30, 'linear');
                new Tween({
                    from: 0,
                    to: 1,
                    in: 60,
                    onUpdate: function (v, t) {
                        fishRotation += v;
                    },
                    onComplete: function () {
                        Bento.screens.show('screens/winscreen', {
                            score: score
                        });
                    }
                });
                new Tween({
                    from: position.x,
                    to: portal.position.x,
                    in: 40,
                    onUpdate: function (v, t) {
                        position.x = v;
                    }
                });
                new Tween({
                    from: position.y,
                    to: portal.position.y,
                    in: 40,
                    onUpdate: function (v, t) {
                        position.y = v;
                    }
                });
            }
        };
        var onRocket = function (rocket) {
            if (!dead && !container) {
                container = rocket;
                squisher.offsetY(-100, 30, 'easeOutCirc', function () {
                    squisher.offsetY(0, 30, 'easeInCirc');
                });
                squisher.scale(0, 30, 'linear');
                physics.velocity.x = 0;
                acceleration.x = 0;
                Tween({
                    from: position.x,
                    to: rocket.position.x,
                    in: 30,
                    ease: 'easeOutCirc',
                    onUpdate: function (v, t) {
                        position.x = v;
                    }
                });
                Tween({
                    from: position.y,
                    to: rocket.position.y - 140,
                    in: 30,
                    ease: 'easeOutCirc',
                    onUpdate: function (v, t) {
                        position.y = v;
                    }
                });
                //drag.x = 0.2;
            }
        };
        var onCollectStar = function () {
            score++;
        };

        var start = function () {
            // physics.velocity.x = flip ? -Globals.SPEED : Globals.SPEED;
            acceleration.x = flip ? -Globals.ACCEL : Globals.ACCEL;
            physics.velocity.y = 0;
            drag.x = 0;
            if (dead) {
                dead = false;
            }
            fishRotation = 0;
            container = null;
            started = true;
        };

        entity.setSize = function (value, flipVal) {
            size = value;
            entity.scale.x = entity.scale.y = size / 128;
            if (flipVal !== undefined) {
                flip = flipVal;
                squisher.setFlip(flip);
            }
            entity.boundingBox = new Rectangle(-60, -80, 100, 80);
        };
        entity.isAlive = function () {
            return !dead;
        };
        entity.canJump = function () {
            return !dead && !container && started && timeSinceOnFloor <= Globals.JUMP_LEEWAY;
        };
        entity.onFloor = function () {
            return onFloor;
        };
        entity.getDirection = function () {
            return flip ? 'left' : 'right';
        };
        entity.setDirection = function (data) {
            var dir = data.direction;
            var animate = data.animate !== false;
            var oldFlip = flip;

            if (dead) {
                return;
            }
            if (!Utils.isDefined(dir)) {
                flip = !flip;
            } else {
                flip = (dir === 'left');
            }
            squisher.setFlip(flip);

            acceleration.x = flip ? -Globals.ACCEL : Globals.ACCEL;
        };
        entity.getPlanet = function () {
            return planet;
        };
        entity.setPlanet = function (newPlanet) {
            planet = newPlanet;
            entity.remove(planetWrap);
            planetWrap = new PlanetWrap({
                planet: planet
            });
            entity.attach(planetWrap);
            entity.setSize(planet.playerSize);
            fishRotation = 0;
            squisher.scale(1, 1, 'linear');
            start();
        };

        entity.hasStarted = function () {
            return started;
        };
        var boostDelay;
        entity.bounceUp = function (minPower, maxPower) {
            var jumpHigher = function () {
                physics.velocity.y = -Globals.JUMP_POWER * maxPower;
                new Tween({
                    from: fishRotation,
                    to: fishRotation + Math.PI * 2 * (flip ? -1 : 1),
                    in: 20,
                    ease: 'exponential',
                    decay: 5,
                    oscillations: 1,
                    onUpdate: function (v, t) {
                        fishRotation = v;
                        fishRotation %= Math.PI * 2;
                    }
                });
            };
            var giveDelay = function () {
                if (boostDelay && boostDelay.isAdded) {
                    // cancel any previous tween
                    boostDelay.removeSelf();
                    boostDelay = null;
                }
                boostDelay = new Tween({
                    // give an extra 5 frame window
                    in: 5,
                    onUpdate: function (v, t) {
                        if (isHolding && !onFloor) {
                            jumpHigher();
                            // cancel tween
                            if (boostDelay && boostDelay.isAdded) {
                                boostDelay.removeSelf();
                                boostDelay = null;
                            }
                        }
                    },
                    onComplete: function () {
                        boostDelay = null;
                    }
                });
            };
            minPower = minPower || 0.75;
            maxPower = maxPower || 1.25;
            onFloor = false;
            hasDoubleJumped = false; // allow another double jump
            if (isHolding) {
                jumpHigher();
            } else {
                physics.velocity.y = -Globals.JUMP_POWER * minPower;
                giveDelay();
            }

            TapticEngine.impact({
                type: "medium"
            });
        };
        entity.revive = function () {
            dead = false;
            entity.visible = true;
            invincibleTimer = 120;
            // reset physics values
            start();

            // restore boss event listener
            if (endBossFight) {
                Bento.objects.attach(endBossFight);
            }
            // revert gamespeed
            new Tween({
                from: Bento.getGameSpeed(),
                to: cachedGameSpeed,
                in: 30,
                ignoreGameSpeed: true,
                onUpdate: function (v, t) {
                    Bento.setGameSpeed(v);
                }
            });

            EventSystem.fire('onPlayerRevive');
        };
        entity.bounceUpForced = function (power) {
            forcedJump = true;
            entity.bounceUp(power, power);
        };

        return entity;
    };
});