/**
 * Sending the player from a point on one planet to a point on another planet in the same level
 * Travels from one point to another, spits out the player when it's done
 * @moduleName PlayerCarrier
 * @snippet PlayerCarrier.snippet
PlayerCarrier({
    position: new Vector2(0, 0)
})
 */
bento.define('entities/playercarrier', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'globals',
    'entities/particle'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Globals,
    Particle
) {
    'use strict';
    return function (settings) {
        var viewport = Bento.getViewport();

        var sourcePlanet = settings.sourcePlanet;
        var destPlanet = settings.destPlanet;
        var posInSourcePlanet = settings.posInSourcePlanet;
        var posInDestPlanet = settings.posInDestPlanet;
        
        var sourcePos = Utils.wrapPoint(posInSourcePlanet, sourcePlanet.radius, sourcePlanet.angle)
                             .addTo(sourcePlanet.position)
                             .addTo(posInSourcePlanet);
        var destPos = Utils.wrapPoint(posInDestPlanet, destPlanet.radius, destPlanet.angle)
                           .addTo(destPlanet.position)
                           .addTo(posInDestPlanet);
        
        var destAngle = -(posInDestPlanet.x / destPlanet.radius);
        
        while (destAngle - destPlanet.angle > Math.PI) destAngle -= Math.PI*2;
        while (destAngle - destPlanet.angle < -Math.PI) destAngle += Math.PI*2;


        var sprite = new Sprite({
            imageName: 'rocket',
            originRelative: new Vector2(0.5, 1.0)
        });

        var position = sourcePos.clone();

        var entity = new Entity({
            z: Globals.Layers.ROCKETS,
            name: 'playerCarrier',
            position: position,
            components: [
                sprite,
            ]
        });

        var curveAmountX = destPos.x > sourcePos.x ? -500 : 500;
        var curveAmountY = destPos.y < sourcePos.y ? -1200 : 1200;

        var getFlightPositionInfo = function (v, oldPos) {
            var curvature = Math.sin(v * Math.PI);
            var x = sourcePos.x + (destPos.x - sourcePos.x) * v + curvature*curveAmountX;
            var y = sourcePos.y + (destPos.y - sourcePos.y) * v + curvature*curveAmountY;
            var dx = x - oldPos.x;
            var dy = y - oldPos.y;
            return {
                x:x, y:y, dx:dx, dy:dy,
                angle: Math.atan2(dy, dx) + Math.PI*0.5
            };
        };

        var startInfo = getFlightPositionInfo(0.000001, sourcePos);

        // rotate to starting angle:
        new Tween({
            from: 0, to: 1,
            in: Globals.ROCKET_FLIGHT_DELAY,
            ease: 'easeInOutQuad',
            onUpdate: function (v) {
                entity.rotation = v * startInfo.angle;
                position.x = startInfo.x + v * Utils.getRandomRangeFloat(-8, 8);
                position.y = startInfo.y + v * Utils.getRandomRangeFloat(-8, 8);
            },
            onComplete: function () {
                position.x = sourcePos.x;
                position.y = sourcePos.y;
            }
        });

        // liftoff!  move to destination planet 
        new Tween({
            from: 0.00001, to: 1,
            in: Globals.ROCKET_FLIGHT_DURATION,
            delay: Globals.ROCKET_FLIGHT_DELAY+1,
            ease: 'easeInSine',
            onUpdate: function (v, t) {
                destPos = Utils.wrapPoint(posInDestPlanet, destPlanet.radius, destPlanet.angle).addTo(destPlanet.position).addTo(posInDestPlanet);
                
                var info = getFlightPositionInfo(v, position);
                position.x = info.x;
                position.y = info.y;
                entity.rotation = info.angle;

                Bento.objects.attach(new Particle({
                    position: entity.position.subtract(new Vector2(info.dx, info.dy)),
                    fadeSpeed: Utils.getRandomRangeFloat(0.02, 0.04),
                    velocity: new Vector2(
                        -info.dx*0.05 + Utils.getRandomRangeFloat(-3, 3),
                        -info.dy*0.05 + Utils.getRandomRangeFloat(-3, 3))
                }));
            },
            onComplete: function () {
                if (settings.onComplete) {
                    settings.onComplete();
                }
                entity.removeSelf();
            }
        });

        // rotate destination planet so that we always land at the top
        new Tween({
            from: destPlanet.angle, to: destAngle,
            in: Globals.ROCKET_FLIGHT_DURATION,
            delay: Globals.ROCKET_FLIGHT_DELAY,
            ease: 'easeInOutQuad',
            onUpdate: function (v, t) {
                destPlanet.angle = v;
            }
        });

        return entity;
    };
});