/**
 * Single-instance entity in charge of increasing & retrieving score, exp points, etc.
 * @moduleName ScoreManager
 * @snippet ScoreManager.snippet
ScoreManager({})
 */
bento.define('entities/scoremanager', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'gui/gametext',
    'globals',
    'entities/expbonustext',
    'managers/save',
    'modules/expitems'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Utils,
    Tween,
    GameText,
    Globals,
    ExpBonusText,
    SaveManager,
    ExpItems 
) {
    'use strict';

    var calculateTargetScoreToBoss = function (n) {
        switch (n) {
            // case 1:  return 500;
            // default: return (n-1) * 1000;
            default: return n * 1000;
        }
    };

    var calculateExpToLevel = function (n) {
        return 500 + (n-1) * 500;
    };

    var ScoreManager = function (settings) {
        var viewport = Bento.getViewport();
        
        var distancePoints = Globals.STARTING_DISTANCE;  // same as score but without the bonuses (used for boss meter)
        var internalScore = Globals.STARTING_SCORE; // floating point
        var scoreRate = 0;     // quickly tweens up from 0 for effect.
        var comboPoints = 0;   // additional points from the ongoing combo

        var gainedExp = 0;  // additional exp amount gained in the current run

        // various stats from the current run
        var coinsCollected = 0;
        var spikesCleared = 0;
        var enemiesKilled = 0;
        var bossesCleared = Globals.STARTING_BOSSES_CLEARED;
        var trickJumps = 0;
        var maxTrickCombo = 0;

        var targetScoreToBoss = calculateTargetScoreToBoss(1);
        var bossInProgress = false;
        var postBoss = 0;

        var behavior = {
            start: function (data) {
                EventSystem.on('startLevel', onStartLevel);
                EventSystem.on('crossedSpike', onCrossedSpike);
                EventSystem.on('killedEnemy', onKilledEnemy);
                EventSystem.on('updateCoins', onCollectCoin);
                EventSystem.on('bossCleared', onBossCleared);
                EventSystem.on('trickJump', onTrickJump);
                EventSystem.on('comboEnded', onComboEnded);
            },
            destroy: function (data) {
                EventSystem.off('startLevel', onStartLevel);
                EventSystem.off('crossedSpike', onCrossedSpike);
                EventSystem.off('killedEnemy', onKilledEnemy);
                EventSystem.off('updateCoins', onCollectCoin);
                EventSystem.off('bossCleared', onBossCleared);
                EventSystem.off('trickJump', onTrickJump);
                EventSystem.off('comboEnded', onComboEnded);
            },
            update: function (data) {
                var player = Bento.objects.get('player');
                if (!bossInProgress && player && player.hasStarted() && player.isAlive()) {
                    if (distancePoints > targetScoreToBoss) {
                        distancePoints = targetScoreToBoss;
                        EventSystem.fire('startBoss', {
                            bossNumber: bossesCleared
                        });
                        bossInProgress = true;
                    } else {
                        if (postBoss > 0) {
                            postBoss -= data.speed;
                        } else {
                            internalScore += data.speed * scoreRate;
                            distancePoints += data.speed * scoreRate;
                        }
                    }
                }
            },
            draw: function (data) {}
        };

        var onStartLevel = function () {
            new Tween({
                from: 0.1,
                to: 0.5,
                in: 60,
                ease: 'linear',
                onUpdate: function (v, t) {
                    scoreRate = v;
                }
            });
        };

        var onCrossedSpike = function (spike) {
            // spike.getComponent('planetWrap', function (planetWrap) {
            //     Bento.objects.attach(new ExpBonusText({
            //         planet: planetWrap.planet,
            //         posInPlanet: spike.position.add(Vector2(-30, -80)),
            //         value: 20
            //     }));
            // });
            // spikesCleared++;
            // gainedExp += 20;
        };
        var onKilledEnemy = function (enemy) {
            enemy.getComponent('planetWrap', function (planetWrap) {
                Bento.objects.attach(new ExpBonusText({
                    planet: planetWrap.planet,
                    posInPlanet: enemy.position.add(Vector2(-30, -80)),
                    value: 80
                }));
            });
            enemiesKilled++;
            gainedExp += 80;
        };
        var onCollectCoin = function (data) {
            // coin.getComponent('planetWrap', function (planetWrap) {
            //     Bento.objects.attach(new ExpBonusText({
            //         planet: planetWrap.planet,
            //         posInPlanet: coin.position.add(Vector2(-30, -80)),
            //         value: 50
            //     }));
            // });
            // gainedExp += 50;
        };
        var onTrickJump = function (data) {
            trickJumps++;
            if (data.combo > maxTrickCombo) {
                maxTrickCombo = data.combo;
            }
            // gainedExp += data.value;
            // internalScore += data.value;
            comboPoints += data.value;
        };

        var onComboEnded = function (data) {
            internalScore += comboPoints;
            comboPoints = 0;
        };

        var onBossCleared = function () {
            var targetScoreToOldBoss = targetScoreToBoss;
            bossesCleared++;
            bossInProgress = false;
            // gap after boss
            // TODO: this number could change depending on the 
            postBoss = 100;
            targetScoreToBoss = calculateTargetScoreToBoss(bossesCleared+1);
            EventSystem.fire('targetScoreUpdated', {
                current: targetScoreToBoss,
                previous: targetScoreToOldBoss 
            });
        };

        var entity = new Entity({
            z: Globals.Layers.HUD,
            name: 'scoreManager',
            updateWhenPaused: 0,
            float: true,
            components: [
                behavior
            ]
        });

        // only one instance is supposed to be active at a time:

        Bento.objects.get('scoreManager', function (scoreManager) {
            Utils.log('WARNING: Created a ScoreManager while the previous one is still attached.');
        });


        // update the globally accessible functions for the new instance:


        /**
         * Retrieve the current travelled distance.
         * @snippet ScoreManager.getDistancePoints|snippet
            ScoreManager.getDistancePoints()
         */
        ScoreManager.getDistancePoints = function () {
            return distancePoints;
        };

        /**
         * Retrieve the additional score to be added from the current combo.
         * (returns 0 if there is no ongoing combo)
         * @snippet ScoreManager.getComboPoints|snippet
            ScoreManager.getComboPoints()
         */
        ScoreManager.getComboPoints = function () {
            return comboPoints;
        };

        /**
         * Retrieve the current score before combo is applied (for display purposes)
         * @snippet ScoreManager.getVisualScore|snippet
            ScoreManager.getVisualScore()
         */
        ScoreManager.getVisualScore = function () {
            return Math.floor(internalScore);
        };

        /**
         * Retrieve the current score plus in-progress combo points.
         * This score will always be accurate for saving/analytics, even if the player is in the middle of a combo.
         * @snippet ScoreManager.getScore|snippet
            ScoreManager.getScore()
         */
        ScoreManager.getScore = function () {
            return Math.floor(internalScore) + comboPoints;
        };

        /**
         * Retrieve the highscore for the current world.
         * @snippet ScoreManager.getHighScore|snippet
            ScoreManager.getHighScore()
         */
        ScoreManager.getHighScore = function () {
            return SaveManager.load('highScore-' + Globals.currentWorld);
        };

        /**
         * Overwrites the previous highscore for the current world, if the currently stored score is better.
         * @snippet ScoreManager.saveScore|snippet
            ScoreManager.saveScore()
         */
        ScoreManager.saveScore = function () {
            var score = ScoreManager.getScore();
            var highScore = ScoreManager.getHighScore();
            if (score > highScore) {
                SaveManager.save('highScore-' + Globals.currentWorld, score);
            }
        };

        /**
         * Retrieve the amount of exp earned during this run.
         * @snippet ScoreManager.getGainedExp|snippet
            ScoreManager.getGainedExp()
         */
        ScoreManager.getGainedExp = function () {
            return gainedExp;
        };

        /**
         * Add the experience gained in the last run to the total, and store it.
         * @snippet ScoreManager.storeExp|snippet
            ScoreManager.storeExp()
         */
        ScoreManager.storeExp = function () {
            if (gainedExp > 0) {
                EventSystem.fire('GameAnalytics-custom', {
                    id: 'xp:gain',
                    value: SaveManager.load('expLevel')
                });
                SaveManager.save('storedExp', ScoreManager.getStoredExp() + gainedExp);
                gainedExp = 0;
            }
        };

        /**
         * Get the total experience, except for that which has not yet been stored.
         * @snippet ScoreManager.getStoredExp|snippet
            ScoreManager.getStoredExp()
         */
        ScoreManager.getStoredExp = function () {
            return SaveManager.load('storedExp');
        };

        /**
         * Get the amount of exp required to reach the next experience level.
         * @snippet ScoreManager.getTargetExp|snippet
            ScoreManager.getTargetExp()
         */
        ScoreManager.getTargetExp = function() {
            return calculateExpToLevel(ScoreManager.getExpLevel() + 1);
        };

        /**
         * Get the current experience level.
         * @snippet ScoreManager.getExpLevel|snippet
            ScoreManager.getExpLevel()
         */
        ScoreManager.getExpLevel = function () {
            return SaveManager.load('expLevel');
        };

        /**
         * Increase the experience level and subtract the target points from our stored exp.
         * @snippet ScoreManager.expLevelUp|snippet
            ScoreManager.expLevelUp()
         */
        ScoreManager.expLevelUp = function () {
            var diff = ScoreManager.getStoredExp() - ScoreManager.getTargetExp();
            SaveManager.save('storedExp', Math.max(diff, 0));
            SaveManager.save('expLevel', ScoreManager.getExpLevel() + 1);

            // unlock new exp items
            var item = ExpItems.getNextUnlock();
            if (item) {
                item.unlock();
            }

            EventSystem.fire('GameAnalytics-custom', {
                id: 'xp:levelUp',
                value: SaveManager.load('expLevel')
            });
        };

        /**
         * Returns the number of points until a boss battle.
         * @snippet ScoreManager.getTargetScoreToBoss|snippet
            ScoreManager.getTargetScoreToBoss()
         */
        ScoreManager.getTargetScoreToBoss = function () {
            return targetScoreToBoss;
        };

        /**
         * Returns the number of bosses defeated.
         * @snippet ScoreManager.getBossesCleared|snippet
            ScoreManager.getBossesCleared()
         */
        ScoreManager.getBossesCleared = function () {
            return bossesCleared;
        };

        /**
         * Retrieve the current number of coins collected
         * @snippet ScoreManager.getCoins|snippet
            ScoreManager.getCoins()
         */
        ScoreManager.getCoins = function () {
            return SaveManager.load('coins');
        };

        return entity;
    };

    // create a default instance (so we can use ScoreManager functions in the menus)
    new ScoreManager();

    return ScoreManager;
});
