/**
 * Block that explodes when triggered
 * @moduleName TntBlock
 */
bento.define('entities/tntblock', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/components/eventlistener',
    'globals',
    'components/planetwrap'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Utils,
    Tween,
    EventListener,
    Globals,
    PlanetWrap
) {
    'use strict';

    var Explosion = function (settings) {
        var sprite = new Sprite({
            imageName: 'effects/fx-explosion',
            originRelative: new Vector2(0.5, 0.5),
            frameCountX: 4,
            frameCountY: 3,
            animations: {
                default: {
                    speed: 0.3,
                    frames: [0,1,2,4,5,6,8,9,3,10,7,11]
                }
            }
        });
        sprite.setAnimation('default', function () {
            entity.removeSelf();
        });

        var scale = Utils.getRandomRangeFloat(0.7, 0.9);

        var entity = new Entity({
            z: Globals.Layers.PARTICLES,
            name: 'tntExplosion',
            position: settings.position,
            scale: new Vector2((Math.random() < 0.5 ? scale: -scale), scale),
            // rotation: Utils.getRandomRangeFloat(-0.2, 0.2),
            components: [
                sprite,
                new PlanetWrap({planet: settings.planet, correctBoundingBox: false})
            ]
        });
        Bento.audio.playSound('sfx_jump');
        return entity;
    };

    return function (settings) {
        var viewport = Bento.getViewport();
        var size = Utils.getTileProperty(settings, 'size', 128);
        var triggerId = Utils.getObjectProperty(settings, 'trigger', null); // null represents 'any' toggle
        var explodeTimer = -1;

        var sprite = new Sprite({
            imageName: 'tnt',
            originRelative: new Vector2(0.5, 1.0)
        });
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {},
            update: function (data) {
                if (explodeTimer !== -1) {
                    explodeTimer -= data.speed;
                    if (explodeTimer <= 0) {
                        explode();
                    }
                }

            },
        };
        var toggleListener = new EventListener({
            eventName: 'toggleSwitchPressed',
            onEvent: function (id) {
                if (id === triggerId || id === null || triggerId === null) {
                    if (explodeTimer === -1) {
                        explodeTimer = 25 * getPlayerSeparationAngle();
                        new Tween({
                            from: 0, to: 1, in: 30,
                            delay: Math.max(explodeTimer - 30, 0),
                            ease: 'easeInQuad',
                            onUpdate: function (v, t) {
                                entity.scale.x = entity.scale.y = 1 + v*0.2;
                                entity.rotation = Math.sin(v * 30) * v * 0.2;
                            }
                        });
                    }
                }
            }
        });

        // we use this to explode later when the player is further away
        var getPlayerSeparationAngle = function () {
            var player = Bento.objects.get('player');
            var playerPlanetWrap = player && player.getComponent('planetWrap');
            var planetWrap = entity.getComponent('planetWrap');
            if (!planetWrap || !playerPlanetWrap) {
                return 0;
            }
            var angle = Utils.deltaAngle(playerPlanetWrap.angle, planetWrap.angle);
            angle += 0.2;
            while (angle < 0) {
                angle += Math.PI * 2;
            }
            return angle;
        };

        var explode = function () {
            entity.getComponent('planetWrap', function (planetWrap) {
                Bento.objects.attach(new Explosion({
                    position: entity.position.add(new Vector2(0, -70)),
                    planet: planetWrap.planet
                }));
            });
            entity.collidesWith({
                family: 'toggleSwitch',
                offset: Vector2(0, -1),
                onCollide: function (other) {
                    other.removeSelf();
                }
            });
            entity.removeSelf();
            Bento.audio.stopSound('sfx_explosion');
            Bento.audio.playSound('sfx_explosion');
        };

        var entity = new Entity({
            z: Globals.Layers.TNT_BLOCKS,
            name: 'tntBlock',
            family: ['solid'],
            updateWhenPaused: 0,
            float: false,
            scale: Vector2(size / 128, size / 128),
            components: [
                sprite,
                behavior,
                toggleListener,
            ]
        });

        return entity;
    };
});