/**
 * Block that can be turned on/off by buttons
 * @moduleName ToggleBlock
 * @snippet ToggleBlock.snippet
ToggleBlock({
    toggle: 'myObjectName',
    active: true,
    position: new Vector2(0, 0)
})
 */
bento.define('entities/toggleblock', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/components/eventlistener',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Utils,
    Tween,
    EventListener,
    Globals
) {
    'use strict';

    return function (settings) {
        var viewport = Bento.getViewport();
        var position = settings.position;
        var size = Utils.getTileProperty(settings, 'size', 128);
        var toggleId = Utils.getObjectProperty(settings, 'toggle', null); // null represents 'any' toggle
        var active = Utils.getTileProperty(settings, 'active', false);
        var planetActive = true;

        var sprite = new Sprite({
            imageName: 'toggle-block',
            frameCountX: 2,
            frameCountY: 1,
            originRelative: new Vector2(0.5, 1.0),
            animations: {
                default: {
                    frames: [0]
                },
                on: {
                    frames: [1]
                },
            }
        });
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                if (active) {
                    activate();
                } else {
                    deactivate();
                }
            },
            update: function (data) {
                // we should only count as solid if we are active and our planet is also active
                if (active && planetActive) {
                    if (!entity.isFamily('solid')) {
                        entity.addToFamily('solid');
                    }
                } else {
                    if (entity.isFamily('solid')) {
                        entity.removeFromFamily('solid');
                    }
                }
            },
            draw: function (data) {}
        };
        // if we belong to a planet, disable collisions unless our planet is currently active
        // we can't use wrap.manageFamily for this, because collisions should also be disabled if the block itself is switched off.
        var activePlanetListener = new EventListener({
            name: 'eventListener',
            eventName: 'setActivePlanet',
            ignorePause: false,
            onEvent: function (newPlanet) {
                entity.getComponent('planetWrap', function (wrap) {
                    if (newPlanet === wrap.planet) {
                        planetActive = true;
                    } else {
                        planetActive = false;
                    }
                });
            }
        });
        var toggleListener = new EventListener({
            name: 'eventListener',
            eventName: 'toggleSwitchPressed',
            ignorePause: false,
            onEvent: function (id) {
                if (id === toggleId || id === null || toggleId === null) {
                    if (!active) {
                        activate();
                    } else {
                        deactivate();
                    }
                }
            }
        });
        var activate = function () {
            active = true;
            sprite.setAnimation('on');
        };
        var deactivate = function () {
            active = false;
            sprite.setAnimation('default');
        };
        var toggle = function (source) {
            if (!source || source.name === settings.toggle) {
                if (active) {
                    deactivate();
                } else {
                    activate();
                }
            }
        };
        var entity = new Entity({
            z: Globals.Layers.BLOCKS,
            name: 'block',
            family: ['solid'],
            position: position,
            updateWhenPaused: 0,
            float: false,
            scale: Vector2(size / 128, size / 128),
            components: [
                sprite,
                behavior,
                toggleListener,
                activePlanetListener
            ]
        });

        return entity;
    };
});