/**
 * General entity for all kinds of buttons/levers/switches
 * @moduleName ToggleSwitch
 */
bento.define('entities/toggleswitch', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'components/physics',
    'globals',
    'components/squisher',
    'bento/components/eventlistener',
    'components/planetwrap'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Physics,
    Globals,
    Squisher,
    EventListener,
    PlanetWrap
) {
    'use strict';

    return function (settings) {
        var viewport = Bento.getViewport();
        var position = settings.position;
        
        // toggleId is the name in tiled, otherwise null represents 'any' toggle
        var toggleId = Utils.getTiledObject(settings).name || null;

        // note: some image names result in slightly different behaviour
        // (e.g. tnt-trigger can only be pressed once)
        var imageName = Utils.getTileProperty(settings, 'imageName', 'toggle-switch');

        var sprite = new Sprite({
            imageName: imageName,
            frameCountX: 2,
            originRelative: new Vector2(0.5, 0.9),
            animations: {
                default: { frames: [1] },
                pressed: { frames: [0] },
            }
        });

        var size = Utils.getTileProperty(settings, 'size', 128);
        var pressOffset = Vector2(0, -2);
        var pressed = false;

        var squisher = new Squisher({});

        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                entity.getComponent('planetWrap', function (planetWrap) {
                    planetWrap.manageFamily('solid');
                });
            },
            destroy: function (data) {},
            update: function (data) {
                if (pressed) {
                    return;
                }
                // check if we collide with the player
                var player = Bento.objects.get('player');
                if (player && player.isAlive() && Utils.samePlanet(player, entity)) {

                    // check if the player hit us from above
                    entity.collidesWith({
                        entity: player,
                        offset: pressOffset,
                        onCollide: press
                    });
                }
            },
            draw: function (data) {}
        };

        var entity = new Entity({
            z: Globals.Layers.SWITCHES,
            name: 'toggleSwitch',
            family: ['solid', 'toggleSwitch'],
            position: position,
            scale: Vector2(size / 128, size / 128),
            boundingBox: new Rectangle(-50, -30, 100, 40),
            components: [
                squisher,
                sprite,
                behavior
            ]
        });

        var press = function () {
            if (pressed) {
                return;
            }
            pressed = true;
            if (imageName !== 'tnt-trigger') {
                Utils.timeout(60, function () {
                    pressed = false;
                });
            }
            Utils.timeout(6, function () {
                EventSystem.fire('toggleSwitchPressed', toggleId);
                sprite.setAnimation('pressed');
            });
            squisher.scaleY(0.3, 6, 'linear', function () {
                squisher.scaleY(1, 6, 'linear');
            });
            squisher.scaleX(1.8, 6, 'linear', function () {
                squisher.scaleX(1, 6, 'linear');
            });
            
            Bento.audio.playSound('sfx_switch');
        };

        var onSetActivePlanet = function (newPlanet) {
            entity.getComponent('planetWrap', function (wrap) {
                if (newPlanet === wrap.planet) {
                    entity.addToFamily('solid');
                } else {
                    entity.removeFromFamily('solid');
                }
            });
            squisher.offsetY(128*0.7, 6, 'linear');
        };

        return entity;
    };
});