/**
 * Scrolling tiled background
 * @moduleName BackScroller
 * @snippet BackScroller|snippet
BackScroller({
    imageName: '$1',
    position: new Vector2(${2:0}, ${3:0}),
    spacing: new Vector2(${4:0}, ${5:0}),
    velocity: new Vector2(${6:0.5}, ${7:0.5}),
    wrapX: ${8:true},
    wrapY: ${9:true},
    planet: null,
    parallax: new Vector2(0, 0)
});
 */
bento.define('gui/backscroller', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Globals
) {
    'use strict';
    return function (settings) {
        var viewport = Bento.getViewport();
        var wrapX = Utils.getDefault(settings.wrapX, true);
        var wrapY = Utils.getDefault(settings.wrapY, true);
        var image = Bento.assets.getImage(settings.imageName);
        // note: wrapping does not take origin into account yet
        var origin = settings.origin || new Vector2(0, 0);
        var originRelative = settings.originRelative || new Vector2(0, 0);
        
        // parallax based on planet rotation
        var parallax = settings.parallax || 0;

        var onPlanetAngleChanged = function (data) {
            behavior.offset.x += (data.delta / Math.PI*2) * (image.width + behavior.spacing.x) * parallax;
        };
        
        var behavior = {
            name: 'scrollComponent',
            offset: new Vector2(0, 0),
            fillSize: settings.fillSize || viewport.clone(),
            velocity: settings.velocity || new Vector2(0.5, -0.5),
            spacing: settings.spacing || new Vector2(0, 0),
            countX: 0,
            countY: 0,
            start: function (data) {                
                // count x and y
                behavior.countX = wrapX ? this.fillSize.width / (image.width + this.spacing.x) : 0;
                behavior.countY = wrapY ? this.fillSize.height / (image.height + this.spacing.y) : 0;
                EventSystem.on('planetAngleChanged', onPlanetAngleChanged);
            },
            destroy: function (data) {
                EventSystem.off('planetAngleChanged', onPlanetAngleChanged);
            },
            update: function (data) {
                behavior.offset.x += behavior.velocity.x;
                behavior.offset.y += behavior.velocity.y;
            },
            draw: function (data) {
                var i, j;
                var countX = behavior.countX;
                var countY = behavior.countY;
                var effWidth = (image.width + this.spacing.x);
                var effHeight = (image.height + this.spacing.y);
                var renderer = data.renderer;
                // clamp offset
                var offset = new Vector2(0, 0);
                offset.x = Utils.modulo(behavior.offset.x, effWidth);
                offset.y = Utils.modulo(behavior.offset.y, effHeight);

                for (i = -1; i < countX; ++i) {
                    for (j = -1; j < countY; ++j) {
                        renderer.drawImage(
                            image,
                            0, 
                            0, 
                            image.width, 
                            image.height,
                            offset.x + i * effWidth - origin.x, 
                            offset.y + j * effHeight - origin.y, 
                            image.width, 
                            image.height
                        );
                    }
                }
            },
            attached: function () {
                // setup origin relative
                if (originRelative) {
                    origin.x = image.width * originRelative.x;
                    origin.y = image.height * originRelative.y;
                }
            },
        };
        // container entity
        var entitySettings = Utils.extend({
            z: 0,
            name: 'backScroller',
            position: settings.position,
            originRelative: new Vector2(0, 0),
            updateWhenPaused: 0,
            float: true,
            components: [
                behavior
            ]
        }, settings);
        var entity = new Entity(entitySettings);

        return entity;
    };
});