/**
 * Shows how many enemies are remaining in the current fight
 * Assumptions:
 * - No enemies are added during the middle of the fight.
 * - No enemies are killed while the health icons are spawning.
 * @moduleName BossHealthBar
 * @snippet BossHealthBar.snippet
BossHealthBar({
    position: new Vector2(0, 0)
})
 */
bento.define('gui/bosshealthbar', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Globals
) {
    'use strict';
    
    var BossHealthIcon = function (settings) {
        var scale = new Vector2(0, 0);
        return new Entity({
            scale: scale,
            components: [
                {
                    start: function () {
                        new Tween({
                            from: 0,
                            to: 1,
                            in: 20,
                            delay: 0,
                            applyOnDelay: 0,
                            ease: 'elastic',
                            decay: 4,
                            oscillations: 2,
                            onUpdate: function (v, t) {
                                scale.x = scale.y = v;
                            }
                        });
                    }
                },
                new Sprite({
                    imageName: 'boss-ui/boss-health',
                    originRelative: new Vector2(0.5, 0.5),
                })
            ]
        });
    };

    var iconSpacing = 140;

    return function (settings) {
        var viewport = Bento.getViewport();
        var position = settings.position;
        
        var totalEnemies = 0;
        var icons = [];
        var chunkLoader;

        var getChunkLoader = function () {
            Bento.objects.get('chunkLoader', function (c) {
                chunkLoader = c;
            });
        };

        // add a new icon and increase the enemy count
        var addIcon = function () {
            var icon = new BossHealthIcon({});
            if (icons.length > 0) {
                icon.position.x = icons[icons.length-1].position.x + iconSpacing;
                Utils.timeout(icons.length * 16, function () {
                    entity.attach(icon);
                    new Tween({
                        from: position.x,
                        to: position.x - iconSpacing/2,
                        in: 16,
                        ease: 'linear',
                        onUpdate: function (v, t) {
                            position.x = v;
                        }
                    });
                });
            } else {
                entity.attach(icon);
            }
            icons.push(icon);
        };

        // change the graphic for the last killed icon to 'damaged'
        var killIcon = function (n) {
            var icon = icons[n];
            if (!icon) {
                Utils.log('invalid boss health icon '+n);
                return;
            }
            icon.getComponent('sprite', function (sprite) {
                sprite.setup({
                    imageName: 'boss-ui/boss-damage',
                    originRelative: new Vector2(0.5, 0.5),
                });
            });
            new Tween({
                from: 0.4,
                to: 0,
                in: 50,
                ease: 'elastic',
                decay: 2,
                oscillations: 5,
                onStart: function () {},
                onUpdate: function (v, t) {
                    icon.rotation = v;
                }
            });
        };

        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                getChunkLoader();
                EventSystem.on('killedBossEnemy', onKilledBossEnemy);
                EventSystem.on('bossCleared', onBossCleared);
            },
            destroy: function (data) {
                EventSystem.off('killedBossEnemy', onKilledBossEnemy);
                EventSystem.off('bossCleared', onBossCleared);
            },
            update: function (data) {
                if (chunkLoader) {
                    if (chunkLoader.bossEnemies > totalEnemies) {
                        totalEnemies = chunkLoader.bossEnemies;
                    }
                    if (icons.length < totalEnemies) {
                        addIcon();
                    }
                } else {
                    getChunkLoader();
                }
            },
        };

        var onKilledBossEnemy = function (data) {
            killIcon(data.enemiesKilled - 1);
        };
        var onBossCleared = function () {
            var iconsRemaining = icons.length;
            icons.forEach(function (icon, i) {
                new Tween({
                    from: 1, to: 0, in: 30,
                    delay: i * 14,
                    ease: 'easeInQuad',
                    onUpdate: function (v, t) {
                        icon.scale.x = icon.scale.y = v;
                        icon.alpha = v;
                        icon.rotation = -(1-v);
                    },
                    onComplete: function() {
                        // if all the icons are done tweening, remove the container
                        if (--iconsRemaining === 0) {
                            entity.removeSelf();
                        }
                    }
                });
            });
        };
        
        var entity = new Entity({
            z: Globals.Layers.GUI,
            name: 'bossHealthBar',
            position: position,
            float: true,
            components: [
                behavior,
            ]
        });
        return entity;
    };
});