/**
 * Overlay that shows before / after a boss fight
 * @moduleName BossOverlay
 * @snippet BossOverlay.snippet
BossOverlay({})
 */
bento.define('gui/bossoverlay', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/components/fill',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'globals',
    'gui/gametext',
    'entities/scoremanager',
    'gui/confetti',
    'gui/bossintrooutrotext',
    'bento/canvas',
    'gui/experiencebar'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Fill,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Globals,
    GameText,
    ScoreManager,
    Confetti,
    BossIntroOutroText,
    Canvas,
    ExperienceBar
) {
    'use strict';

    return function (settings) {
        var viewport = Bento.getViewport();
        var position = settings.position;
        var mode = settings.mode;
        var completionCallback = settings.onComplete || null;
        var extraDuration = 0;

        var behavior = {
            start: function (data) {
                Bento.objects.get('pauseButton', function (pauseButton) {
                    pauseButton.active = false;
                });

                fadeIn();

                // different behaviour based on overlay type
                switch (mode) {

                case 'boss-intro':
                    extraDuration = 10;

                    new Tween({
                        in: 33,
                        ignoreGameSpeed: true,
                        onComplete: function () {
                            Bento.audio.playSound('sfx_boss_intro');
                        }
                    });

                    Bento.objects.attach(new BossIntroOutroText({
                        text: 'BOSS\nFIGHT',
                        position: new Vector2(viewport.width * 0.5, viewport.height * 0.4),
                        animation: 'drop-in',
                        extraDuration: extraDuration
                    }));

                    // the skull
                    var i;
                    var bossIcon = new Entity({
                        position: new Vector2(viewport.width * 0.5, viewport.height * 0.4 + 370),
                        scale: new Vector2(0, 0),
                        components: [
                            new Sprite({
                                imageName: 'boss-ui/boss-icon',
                                originRelative: new Vector2(0.5, 0.5),
                            })
                        ]
                    });
                    entity.attach(bossIcon);

                    // scale in skull
                    new Tween({
                        ignoreGameSpeed: true,
                        from: 0,
                        to: 1,
                        in: 80,
                        delay: 40,
                        ease: 'easeOutElastic',
                        onUpdate: function (v, t) {
                            bossIcon.scale.x = bossIcon.scale.y = v;
                            bossIcon.alpha = v;
                        }
                    });

                    // wiggle animation
                    for (i = 0; i < 4; i++) {
                        new Tween({
                            ignoreGameSpeed: true,
                            from: bossIcon.rotation,
                            to: 0.1,
                            in: 10,
                            delay: 50 + i * 20,
                            ease: 'linear',
                            onUpdate: function (v, t) {
                                bossIcon.rotation = v;
                            }
                        });
                        new Tween({
                            ignoreGameSpeed: true,
                            from: bossIcon.rotation,
                            to: -0.1,
                            in: 10,
                            delay: 60 + i * 20,
                            ease: 'linear',
                            onUpdate: function (v, t) {
                                bossIcon.rotation = v;
                            }
                        });
                    }

                    // scale out skull
                    new Tween({
                        ignoreGameSpeed: true,
                        from: 1,
                        to: 0,
                        in: 40,
                        delay: 100 + extraDuration,
                        ease: 'easeInQuad',
                        onUpdate: function (v, t) {
                            bossIcon.scale.x = bossIcon.scale.y = v;
                            bossIcon.alpha = v;
                        }
                    });
                    break;

                case 'boss-cleared':
                    extraDuration = 120;
                    var bossCleared = new BossIntroOutroText({
                        text: 'BOSS\nCLEARED',
                        position: new Vector2(viewport.width * 0.5, viewport.height * 0.4),
                        animation: 'slide-from-right',
                        extraDuration: extraDuration
                    });
                    Bento.objects.attach(bossCleared);
                    experienceBar = new ExperienceBar({
                        isGrey: false,
                        position: new Vector2(viewport.width * 0.5, viewport.height * 0.4 + 280)
                    });
                    // scale in exp bar
                    new Tween({
                        ignoreGameSpeed: true,
                        from: 0,
                        to: 1,
                        in: 80,
                        applyOnDelay: 40,
                        ease: 'easeOutElastic',
                        onUpdate: function (v, t) {
                            experienceBar.scale.x = experienceBar.scale.y = v;
                            experienceBar.alpha = v;
                        },
                    });
                    // scale out exp bar
                    new Tween({
                        ignoreGameSpeed: true,
                        from: 1,
                        to: 0,
                        in: 30,
                        delay: 90 + extraDuration,
                        ease: 'easeInQuad',
                        onUpdate: function (v, t) {
                            experienceBar.scale.x = experienceBar.scale.y = v;
                            experienceBar.alpha = v;
                        }
                    });
                    Bento.objects.attach(experienceBar);
                    break;
                case 'boss-failed':
                    experienceBar = new ExperienceBar({
                        isGrey: true,
                        position: new Vector2(viewport.width * 0.5, viewport.height * 0.62)
                    });
                    Bento.objects.attach(experienceBar);
                    break;
                default:
                    Utils.log('no such boss overlay mode "' + mode + '"');
                }
            },
            destroy: function (data) {
                Bento.objects.remove(experienceBar);

                Bento.objects.get('pauseButton', function (pauseButton) {
                    pauseButton.active = true;
                });
            },
            update: function (data) {},
            draw: function (data) {}
        };

        var fillColor = [0, 0, 0, 0];
        var fill = new Fill({
            color: fillColor
        });

        var experienceBar;
        
        var fxTada = new Entity({
            position: new Vector2(viewport.width*0.5, viewport.height*0.5),
            alpha: 0,
            components: [
                {
                    update: function () {
                        fxTada.rotation += 0.03;
                    }
                },
                new Sprite({
                    imageName: 'boss-ui/fx-tada',
                    originRelative: new Vector2(0.5, 0.5),
                })
            ]
        });
        new Tween({
            ignoreGameSpeed: true,
            from: 0,
            to: 0.6,
            in: 60,
            ease: 'linear',
            onUpdate: function (v, t) {
                fxTada.alpha = v;
            },
            onComplete: function () {
                new Tween({
                    ignoreGameSpeed: true,
                    from: fxTada.alpha,
                    to: 0,
                    in: 60,
                    ease: 'linear',
                    onUpdate: function (v, t) {
                        fxTada.alpha = v;
                    }
                });
            }
        });

        var fadeIn = function () {
            new Tween({
                ignoreGameSpeed: true,
                from: 0,
                to: 1,
                in: 50,
                ease: 'easeInOutQuad',
                onUpdate: function (v, t) {
                    fillColor[3] = v * 0.7; // set overlay background alpha
                },
                onComplete: function () {
                    new Tween({
                        ignoreGameSpeed: true,
                        in: 80,
                        onComplete: fadeOut
                    });
                }
            });
        };
        var fadeOut = function () {
            new Tween({
                ignoreGameSpeed: true,
                from: 1,
                to: 0,
                in: 50,
                ease: 'easeInOutQuad',
                onUpdate: function (v, t) {
                    fillColor[3] = v * 0.7; // set overlay background alpha
                },
                onComplete: function () {
                    entity.removeSelf();
                    if (completionCallback) {
                        completionCallback();
                    }
                }
            });
        };

        var entity = new Entity({
            z: Globals.Layers.OVERLAY,
            name: 'bossOverlay',
            updateWhenPaused: 0,
            float: true,
            components: [
                behavior,
                fill,
                fxTada
            ]
        });

        return entity;
    };
});