/**
 * The gift button keeps track of the time left until user can open
 * @moduleName GiftButton
 * @snippet GiftButton.snippet
GiftButton({
    position: new Vector2(0, 0)
})
 */
bento.define('gui/buttons/gift', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'gui/gametext',
    'managers/time',
    'managers/save',
    'gui/overlays/coins',
    'globals',
    'gui/overlays/dialog',
    'modules/localization',
    'modules/notifications',
    'components/bouncewithmusic',
    'gui/badge'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    GameText,
    TimeManager,
    SaveManager,
    CoinsOverlay,
    Globals,
    Dialog,
    Localization,
    Notifications,
    BounceWithMusic,
    Badge
) {
    'use strict';
    return function (settings) {
        var viewport = Bento.getViewport();
        var position = settings.position;
        var time = new GameText({
            position: new Vector2(0, 64),
            text: '',
            fontSize: 48,
            textBaseline: 'middle',
            linebreaks: false,
            maxWidth: 192
        });
        var canClaim = false;
        var bounceWithMusic = new BounceWithMusic({
            additionalScale: 0.1
        });
        var behavior = {
            name: 'behaviorComponent',
            update: function (data) {
                // update text every second
                if (!canClaim && entity.timer % 60 === 0) {
                    time.setText(getTimeString());
                }

                // wiggle
                if (canClaim) {
                    // entity.rotation = Math.PI / 20 * Math.sin(entity.timer / 10);
                    bounceWithMusic.active = true;
                    badge.visible = true;

                    entity.rotation = Math.PI / 20 * Math.sin(entity.timer / 10);
                } else {
                    entity.rotation = 0;
                    bounceWithMusic.active = false;
                    badge.visible = false;
                }
            },
        };
        var badge = new Badge({});
        /**
         * Returns the remaining time in milliseconds
         */
        var getRemaining = function () {
            var now = TimeManager.getTime();
            var scheduled = SaveManager.load('scheduledGiftTime');
            var remaining = scheduled - now;
            if (remaining > 0) {
                canClaim = false;
            } else {
                canClaim = true;
            }
            return remaining;
        };
        /**
         * Returns time remaining in "00:00:00" format
         * If the time has passed, it returns "Claim now!" as string
         */
        var getTimeString = function () {
            // look up remaining time
            var remaining = getRemaining();

            if (remaining > 0) {
                return Utils.getTimeString(remaining);
            } else {
                return Localization.getText('claimNow');
            }
        };
        var askNotifications = function () {
            var dialog = new Dialog({
                name: 'askNotifications',
                text: Localization.getText('notifications').ask,
                type: 'yesno',
                onYes: function () {
                    // initialize notifications
                    Notifications.init(function (initialized) {
                        var allowed = initialized ? true : false;
                        SaveManager.save('notificationsAllowed', allowed);
                        if (allowed) {
                            scheduleNotification();
                        } else {
                            // user pressed no at the system dialog...
                        }
                    });
                },
                onNo: function () {
                    // try again another time...
                }
            });
            Bento.objects.attach(dialog);

            SaveManager.add('notificationsAsked', 1);
        };
        var scheduleNotification = function () {
            // fire event to schedule notification
            EventSystem.fire('scheduleNotification', {
                id: 'freeGift',
                title: Localization.getText('notifications')['freeGiftTitle'],
                body: Localization.getText('notifications')['freeGiftBody'],
                time: getRemaining()
            });
        };
        var lastGiftAmount = 0;
        var askSecondGift = function () {
            var dialog = new Dialog({
                name: 'doubleGift',
                text: "Get 100 extra coins!",
                // text: Localization.getText('doubleGift').replace('{NUMBER}', 100),
                canClose: true,
                type: 'yes',
                onYes: function () {
                    bento.require(['gui/buttons/coinsforad'], function (CoinsForAdButton) {
                        CoinsForAdButton.watchRewarded({
                            amount: 100,
                            onWatched: function () {}
                        });
                    });
                },
                onNo: function () {},
                onClose: function () {},
                onComplete: function () {}
            });
            dialog.getComponent('buttonOk', function (buttonOk) {
                buttonOk.attach(new BounceWithMusic({
                    additionalScale: 0.1
                }));
                buttonOk.getComponent('sprite', function (sprite) {
                    sprite.setup({
                        imageName: 'ui/button-ad-green',
                        originRelative: new Vector2(0.5, 0.5)
                    });
                });
            });
            Bento.objects.attach(dialog);
        };
        var entity = new ClickButton({
            z: settings.z,
            name: 'giftButton',
            sfx: 'sfx_button',
            imageName: 'title/button-freegift',
            frameCountX: 1,
            frameCountY: 1,
            scale: settings.scale,
            position: settings.position,
            visible: settings.visible,
            onButtonDown: function () {
                if (canClaim) {
                    this.transform.y = 8;
                }
            },
            onButtonUp: function () {
                this.transform.y = 0;
            },
            onClick: function () {
                var giftAmount = Globals.getGiftAmount();
                // check time
                var remaining = getRemaining();

                lastGiftAmount = giftAmount;
                if (remaining > 0) {
                    // cannot claim yet, shake button
                    new Tween({
                        from: 32,
                        to: 0,
                        in: 30,
                        ease: 'elastic',
                        decay: 5,
                        oscillations: 5,
                        onUpdate: function (v, t) {
                            entity.transform.x = v;
                        }
                    });
                } else {
                    // claim gift
                    // schedule x minutes into future
                    SaveManager.add('receivedGifts', 1);
                    SaveManager.save(
                        'scheduledGiftTime',
                        TimeManager.getTime() + Globals.getGiftTime(SaveManager.load('receivedGifts'))
                    );

                    EventSystem.fire('GameAnalytics-source', {
                        currency: 'coins',
                        amount: giftAmount,
                        sourceType: 'reward',
                        sourceId: 'freeGift'
                    });

                    EventSystem.fire('GameAnalytics-custom', {
                        id: Utils.currentScreenName() + ':gui:freeGift',
                        value: giftAmount
                    });

                    // update button
                    canClaim = false;
                    bounceWithMusic.active = false;
                    badge.visible = false;
                    time.setText(getTimeString());

                    SaveManager.add('coins', giftAmount);
                    entity.rotation = 0;
                    Bento.objects.attach(new CoinsOverlay({
                        amount: giftAmount,
                        onComplete: function () {
                            EventSystem.fire('updateCoins', {
                                coins: giftAmount
                            });

                            // after having asked about notifications, ask about doubling up
                            if (SaveManager.load('notificationsAsked') > 0) {
                                askSecondGift();
                            }

                            if (SaveManager.load('notificationsAllowed')) {
                                // schedule notification
                                scheduleNotification();
                            } else if (SaveManager.load('notificationsAsked') === 0) {
                                // ask for notifications after the first gift
                                askNotifications();
                            }
                        }
                    }));
                }

                if (SaveManager.load('totalDeaths-' + Globals.currentWorld) === 1) {
                    EventSystem.fire('GameAnalytics-custom', {
                        id: Globals.currentWorld + ':firstDeathOverlay:buttonGift'
                    });
                }
            },
            components: [
                behavior,
                time,
                badge,
                bounceWithMusic
            ]
        });
        return entity;
    };
});