/**
 * Bar that shows gained experience.
 * @moduleName ExperienceBar
 * @snippet ExperienceBar.snippet
ExperienceBar({
    position: new Vector2(0, 0)
})
 */
bento.define('gui/experiencebar', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'bento/canvas',
    'globals',
    'gui/gametext',
    'entities/scoremanager',
    'modules/localization'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Counter,
    Text,
    Utils,
    Tween,
    Canvas,
    Globals,
    GameText,
    ScoreManager,
    Localization
) {
    'use strict';
    return function (settings) {
        var viewport = Bento.getViewport();
        var position = settings.position;
        var showLevel = settings.showLevel;
        var isGrey = settings.isGrey;

        var extraFullness = 0; // for the yellow (exp gained) bar
        var fullness = 0; // for the blue (total exp) bar

        var oldExp = 0;
        var newExp = 0;
        var targetExp;

        var unlocked = false;

        var barOuter = new Entity({
            position: new Vector2(0, 0),
            components: [
                new Sprite({
                    imageName: 'boss-ui/xp-bar',
                })
            ]
        });
        var barInnerSprite = new Sprite({
            imageName: isGrey ? 'boss-ui/xp-bar-fail' : 'boss-ui/xp-bar-full',
        });
        barInnerSprite.draw = function (data) {
            if (!this.currentAnimation || !this.visible) {
                return;
            }
            this.updateFrame();

            var width = Math.round(fullness * barInner.dimension.width);
            if (!width) {
                return;
            }
            data.renderer.drawImage(
                this.spriteImage,
                this.sourceX,
                this.sourceY,
                width,
                this.frameHeight,
                (-this.origin.x) | 0,
                (-this.origin.y) | 0,
                width,
                this.frameHeight
            );
        };
        var barInner = new Entity({
            position: new Vector2(149, 13),
            scale: new Vector2(1, 1.1),
            components: [
                barInnerSprite
            ]
        });
        var barExtraSprite = new Sprite({
            imageName: 'boss-ui/xp-bar-progress',
        });

        barExtraSprite.draw = function (data) {
            if (!this.currentAnimation || !this.visible) {
                return;
            }
            this.updateFrame();

            var width = Math.round(extraFullness * barExtra.dimension.width);
            if (!width) {
                return;
            }
            data.renderer.drawImage(
                this.spriteImage,
                this.sourceX,
                this.sourceY,
                width,
                this.frameHeight,
                (-this.origin.x) | 0,
                (-this.origin.y) | 0,
                width,
                this.frameHeight
            );
        };
        var barExtra = new Entity({
            position: new Vector2(148, 13),
            scale: new Vector2(1, 1.1),
            components: [
                barExtraSprite
            ]
        });

        // Canvas container for the inner fill of the bar.
        // Acts as a rectangular mask if we try to draw outside it.
        // var canvas = new Canvas({
        //     width: barInner.dimension.width,
        //     height: barInner.dimension.height,
        //     preventAutoClear: false, // prevent canvas from clearing every tick
        //     pixelSize: 1, // multiplies internal canvas size
        //     drawOnce: false, // draw canvas only once
        //     position: new Vector2(148, 12),
        //     components: [
        //         barExtra,
        //         barInner
        //     ]
        // });

        var xpText = new GameText({
            position: new Vector2(80, 8),
            text: 'XP',
            fontSize: 70,
            fontColor: '#000'
        });

        var unlockText;
        var levelText, defeatBossesText;
        var expLevel;

        var behavior = {
            start: function (data) {
                if (showLevel) {
                    expLevel = ScoreManager.getExpLevel();
                    fullness = ScoreManager.getStoredExp() / ScoreManager.getTargetExp();
                    extraFullness = 0.0;

                    levelText = new GameText({
                        position: new Vector2(70, 15),
                        text: Localization.getText('levelN').replace('{NUMBER}', expLevel).toUpperCase(),
                        fontSize: 50,
                    });
                    defeatBossesText = new GameText({
                        position: new Vector2(0, -66),
                        text: Localization.getText('defeatBosses').toUpperCase(),
                        fontSize: 38,
                    });
                    entity.attach(levelText);
                    entity.attach(defeatBossesText);
                } else {
                    oldExp = ScoreManager.getStoredExp();
                    ScoreManager.storeExp();
                    newExp = ScoreManager.getStoredExp();
                    targetExp = ScoreManager.getTargetExp();

                    fullness = oldExp / targetExp;
                    extraFullness = newExp / targetExp;

                    new Tween({
                        from: fullness,
                        to: newExp / targetExp,
                        in: 50,
                        delay: 70,
                        ease: 'easeInQuad',
                        onUpdate: function (v, t) {
                            fullness = v;
                        }
                    });

                    if (newExp >= targetExp) {
                        ScoreManager.expLevelUp();
                    }
                }
            },
            update: function (data) {
                fullness = Utils.clamp(0, fullness, 1);
                extraFullness = Utils.clamp(0, extraFullness, 1);
                // barExtra.position.x = (1 - extraFullness) * -barExtra.dimension.width - 2;
                // barInner.position.x = (1 - fullness) * -barInner.dimension.width;

                if (fullness >= 1.0 && !unlocked && !showLevel) {
                    unlocked = true;
                    Utils.timeout(4, function () {
                        barExtra.removeSelf();
                        barInner.removeSelf();
                        xpText.removeSelf();
                        barOuter.getComponent('sprite', function (sprite) {
                            sprite.setup({
                                imageName: 'boss-ui/xp-bar-all-white',
                            });
                        });
                    });

                    Utils.timeout(10, function () {
                        new Tween({
                            from: 1,
                            to: 0,
                            in: 40,
                            delay: 30,
                            ease: 'easeOutQuad',
                            onUpdate: function (v, t) {
                                container.alpha = v;
                            }
                        });

                        unlockText = new GameText({
                            position: new Vector2(300, 8).addTo(container.position),
                            text: Localization.getText('levelUp').toUpperCase(),
                            fontSize: 60,
                        });
                        entity.attach(unlockText);
                    });
                }
            },
            draw: function (data) {

            }
        };

        var container = new Entity({
            position: new Vector2(-barOuter.dimension.width / 2, 0),
            components: [
                barExtra,
                barInner,
                barOuter,
                xpText,
            ]
        });

        var entity = new Entity({
            z: Globals.Layers.HUD,
            name: 'experienceBar',
            position: position,
            float: true,
            components: [
                behavior,
                container,
            ]
        });
        return entity;
    };
});