/**
 * An overlay when the user gets a bunch of coins
 * @moduleName CoinsOverlay
 * @snippet CoinsOverlay.snippet
CoinsOverlay({
    onComplete: function () {}
})
 */
bento.define('gui/overlays/coins', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'components/modal',
    'bento/components/fill',
    'globals',
    'entities/coin',
    'gui/gametext',
    'modules/tapticengine'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Modal,
    Fill,
    Globals,
    Coin,
    GameText,
    TapticEngine
) {
    'use strict';
    // coin constructor for the overlay
    var makeCoin = function (settings) {
        var coinSprite = new Sprite({
            imageName: 'coin',
            originRelative: new Vector2(0.5, 0.5),
        });
        var lifeTime = settings.lifeTime || 60;
        var behavior = {
            name: 'behavior',
            start: function () {},
            update: function () {
                if (coin.timer > lifeTime) {
                    coin.removeSelf();
                }
            },
            destroy: function () {
                Coin.createSparkleEffect(coin);
            }
        };
        var coin = new Entity({
            name: 'coin',
            position: settings.position,
            components: [
                coinSprite,
                behavior
            ]
        });
        return coin;
    };
    return function (settings) {
        var viewport = Bento.getViewport();
        var onComplete = settings.onComplete;
        var amount = settings.amount || 100;
        var fill = new Entity({
            name: 'fill',
            components: [
                new Fill({
                    color: [0, 0, 0, 1]
                })
            ]
        });
        var modal = new Modal({});
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                // fade in
                new Tween({
                    from: 0,
                    to: 0.75,
                    in: 10,
                    ease: 'linear',
                    onUpdate: function (v, t) {
                        fill.alpha = v;
                    },
                    onComplete: function () {
                        Bento.audio.stopSound('sfx_coinshower');
                        Bento.audio.playSound('sfx_coinshower');
                        Utils.repeat(spawnCoin, 30);
                        showAmount();

                    }
                });
            }
        };
        var text = new GameText({
            position: new Vector2(viewport.width / 2, viewport.height / 2),
            text: '+' + amount,
            fontSize: 100,
            textBaseline: 'middle'
        });
        var entity = new Entity({
            z: settings.z || Globals.Layers.COINSOVERLAY,
            name: 'coinsOverlay',
            float: true,
            components: [
                modal,
                fill,
                behavior
            ]
        });
        /**
         * Show text amount
         */
        var showAmount = function () {
            new Tween({
                from: 0,
                to: 2,
                in: 40,
                delay: 60,
                ease: 'elastic',
                decay: 5,
                oscillations: 2,
                onStart: function () {
                    entity.attach(text);
                },
                onUpdate: function (v, t) {
                    text.scale.x = v;
                    text.scale.y = v;
                },
                onComplete: function () {
                    // fade out
                    fadeOut();
                }
            });

        };
        var fadeOut = function () {
            new Tween({
                from: 1,
                to: 0,
                in: 30,
                delay: 60,
                applyOnDelay: 0,
                ease: 'linear',
                decay: 1,
                oscillations: 1,
                onStart: function () {},
                onUpdate: function (v, t) {
                    fill.alpha = 0.75 * v;
                    text.alpha = v;
                },
                onComplete: function () {
                    entity.removeSelf();
                    if (onComplete) {
                        onComplete();
                    }
                }
            });
        };
        /**
         * Spawn a coin randomly from the bottom
         */
        var spawnCoin = function (i) {
            var coin = makeCoin({
                position: new Vector2(Utils.getRandomRange(128, viewport.width - 128), viewport.height),
                lifeTime: 60 + i
            });
            var targetPosition = new Vector2(
                Utils.getRandomRange(128, viewport.width - 128),
                Utils.getRandomRange(128, viewport.height - 128)
            );
            new Tween({
                from: coin.position,
                to: targetPosition,
                in: 60,
                ease: 'exponential',
                decay: 5,
                onStart: function () {},
                onUpdate: function (v, t) {
                    coin.position = v;
                },
                onComplete: function () {}
            });
            entity.attach(coin);
            if (i % 5 === 0) {
                coin.attach({
                    destroy: function () {
                        // Bento.audio.stopSound('sfx_coin');
                        // Bento.audio.playSound('sfx_coin');

                        TapticEngine.selection();
                    }
                });
            }
        };
        return entity;
    };
});