/**
 * Overlay that waits for IAP callback
 * Contains general code flow for making a purchase
 * Advised to include this module asap, as it contains global event listeners also used for restoring purchases 
 * @moduleName IapOverlay
 */
bento.define('gui/overlays/iap', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/store',
    'modules/localization',
    'bento/components/fill'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    ClickButton,
    Counter,
    Text,
    EventSystem,
    Utils,
    Tween,
    Store,
    Localization,
    Fill
) {
    'use strict';
    var IapOverlay = function (settings) {
        /*settings = {
            // describe your settings object parameters
            timeOutTime: Number // frames until timeout, defaults to 600 (10 seconds)
            onTimeOut: Function // what should happen if the IAP timed out
        }*/
        var viewport = Bento.getViewport();
        // timeout
        var timeOutTime = settings.timeOutTime || 600;
        var onTimeOut = settings.onTimeOut;
        // dark overlay
        var overlay = new Entity({
            name: 'overlay',
            components: [new Fill({
                color: [0, 0, 0, 0.75]
            })]
        });
        var fontSettings = Utils.extend({
            position: new Vector2(viewport.width / 2, viewport.height / 3),
            text: '...',
            font: 'font',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            maxWidth: viewport.width - 16,
            textBaseline: 'middle'
        }, settings.fontSettings);
        var text = new Text(fontSettings);
        // pauselevel is 1 higher than current pauselevel
        var pauseLevel = 0;
        var pauseBehavior = {
            oldPauseLevel: 0,
            start: function () {
                this.oldPauseLevel = Bento.objects.isPaused();
                pauseLevel = this.oldPauseLevel + 1;
                entity.updateWhenPaused = pauseLevel;
                Bento.objects.pause(pauseLevel);
            },
            destroy: function () {
                Bento.objects.pause(this.oldPauseLevel);
            }
        };
        var timeOutBehavior = {
            hasTimedOut: false,
            update: function () {
                if (this.hasTimedOut === false && entity.timer > timeOutTime) {
                    this.hasTimedOut = true;
                    if (onTimeOut) {
                        onTimeOut();
                    }
                }
            }
        };
        var entity = new Entity({
            z: settings.z,
            name: 'iapOverlay',
            float: true,
            components: [
                pauseBehavior,
                overlay,
                text,
                timeOutBehavior
            ]
        });

        // public
        entity.extend({
            /**
             * Close overlay (with a delay)
             */
            exit: function (delay, callback) {
                Utils.timeout(delay || 0, function () {
                    Bento.objects.remove(entity);
                    if (callback) {
                        callback();
                    }
                }, pauseLevel);
            },
            /**
             * Set overlay text
             */
            setText: function (str) {
                text.setText(str);
            }
        });

        return entity;
    };
    /*
     * Global purchase events (game specific)
     */
    // global event listener (place non-consumable purchases here!!)
    EventSystem.on('purchase_success', function (purchaseInfo) {
        IapOverlay.onNonConsumable(purchaseInfo);
    });
    // overwrite this function if needed
    IapOverlay.onNonConsumable = function (purchaseInfo) {
        var productId = purchaseInfo.productId;
        // save flag into savestate
        Bento.saveState.save('purchased-' + productId, true);
    };

    // send business event to GameAnalytics

    EventSystem.on('purchase_success', function (purchaseInfo) {
        
        var productId = purchaseInfo.productId;
        var transactionId = purchaseInfo.transactionId;

        // extended properties
        var product = purchaseInfo.product;
        var lastReceipt = purchaseInfo.lastReceipt;

        if (!lastReceipt || !product) {
            // don't send analytics event
            return;
        }

        var businessEvent = {
            currency: product.currency,
            amount: Math.round(product.price * 100),  // GameAnalytics wants amount in cents (this works even for currencies that aren't usually divided)
            cartType: 'shop',        // Will we ever need different types of cart?
            itemType: 'iap',         // How do we pass in a variety of different product types? (all of them need to be known by GameAnalytics.)
            itemId: productId+'_'+product.title,
        };

        // Purchase validation fields (platform specific)
        if (Utils.isNativeIos()) {
            businessEvent.receipt = lastReceipt;                 // app store receipt (base64)
        } else if (Utils.isNativeAndroid()) {
            businessEvent.receipt = lastReceipt.purchaseData;    // android purchase data (json)
            businessEvent.signature = lastReceipt.dataSignature; // purchase data signature (base64)
        }

        // if (!isSandbox(lastReceipt)) {
            EventSystem.fire('GameAnalytics-business', businessEvent);
        // } else {
            // don't send real business events in sandbox mode
            // console.log('purchased', businessEvent);
        // }
    });

    // analytics + tenjin
    /*EventSystem.on('purchase_success', function (productInfo) {
        var productId = productInfo.productId;
        var product = productInfo.product;
        var transactionId = productInfo.transactionId;
        var trackPurchase = true;
        var lastReceipt = productInfo.lastReceipt;

        if (!lastReceipt) {
            // validation failed, don't send anything to tenjin
            return;
        }

        if (product && trackPurchase) {
            // Tenjin
            if (Utils.isNativeIos()) {
                var priceString = product.price.toString();
                Tenjin.iosTransaction(
                    productInfo.productId, // sku 
                    product.currency || '', // currencycode
                    1, // quantity
                    priceString, // price
                    productInfo.transactionId, // transaction id
                    productInfo.lastReceipt // receipt in base64
                );
            } else if (Utils.isNativeAndroid()) {
                Tenjin.androidTransaction(
                    productInfo.productId, // sku 
                    product.currency || '', // currencycode
                    1, // quantity
                    product.price, // price
                    productInfo.lastReceipt.purchaseData, // purchase data
                    productInfo.lastReceipt.dataSignature // data signature
                );
            }

            Tenjin.sendEvent("validatedPayment");

            // check receipt on iOS for sandbox value
            if (!isSandbox(lastReceipt)) {
                // google analytics
                Analytics.trackPurchase(
                    productInfo.transactionId, // id
                    product.title, // title
                    productInfo.productId, // sku
                    'IAP', // category
                    product.price, // price
                    1, // quantity
                    product.currency || '' // currencycode
                );
            } else {
                Analytics.event('IAP', 'sandbox', product.price, productInfo.productId);
            }
        }

    });*/
    var isSandbox = function (receipt) {
        // this check is only for ios
        if (!Utils.isNativeIos()) {
            return false;
        }
        try {
            var receiptStr;
            if (window.decodeB64) {
                receiptStr = window.decodeB64(receipt);
            } else {
                // atob is not so reliable in cocoon
                receiptStr = window.atob(receipt);
            }
            // check if the receipt contains the string '"environment" = "Sandbox"'
            return receiptStr.indexOf('"environment" = "Sandbox"') >= 0;
        } catch (e) {
            Utils.log('ERROR: malformed receipt');
            return false;
        }
    };
    /*
     * Purchase flow
     * @snippet IapOverlay.Purchase|snippet
IapOverlay.Purchase({
    z: 0,
    id: 'productId',
    fontSettings: {},
    onPurchaseComplete: function (success, id, purchaseInfo) {}
});
     */
    IapOverlay.Purchase = function (settings) {
        var overlayZ = settings.z;
        var id = settings.id; // product id
        var onPurchaseComplete = settings.onPurchaseComplete;
        var fontSettings = settings.fontSettings || {};
        var iapOverlay = new IapOverlay({
            z: overlayZ,
            fontSettings: fontSettings
        });
        var isRestoring = settings.restore || false;
        var purchase = function () {
            var success;

            // show overlay
            Bento.objects.attach(iapOverlay);
            iapOverlay.setText(Localization.getText('purchasing'));

            // wait for callback
            EventSystem.on('purchase_success', onPurchaseSuccess);
            EventSystem.on('purchase_error', onPurchaseError);

            // purchase!
            success = Store.purchase(id);

            if (!success) {
                if (!Utils.isCocoonJS()) {
                    onPurchaseError('Only available in the App version!');
                } else {
                    onPurchaseError(Localization.getText('errorstore'));
                }
            }
        };
        var restore = function () {
            var success;

            // show overlay
            Bento.objects.attach(iapOverlay);
            iapOverlay.setText(Localization.getText('restoring'));

            // wait for callback
            EventSystem.on('restore_success', onPurchaseSuccess);
            EventSystem.on('restore_error', onPurchaseError);
            EventSystem.on('purchase_error', onPurchaseError);

            success = Store.restore(id);

            if (!success) {
                if (!Utils.isCocoonJS()) {
                    onPurchaseError('Only available in the App version!');
                } else {
                    onPurchaseError(Localization.getText('errorstore'));
                }
            }
        };
        var onPurchaseSuccess = function (purchaseInfo) {
            var sendAnalytics = function () {
                var productId = purchaseInfo.productId;
                var iapAmount = Bento.saveState.load('iapAmount', 0);
                var highSpender = Bento.saveState.load('highSpender', false);
                var productAmount = Bento.saveState.load('productAmount-' + productId, 0);

                iapAmount++;
                Bento.saveState.save('iapAmount', iapAmount);
                productAmount++;
                Bento.saveState.save('productAmount-' + productId, productAmount);

                // which IAP was bought
                EventSystem.fire('Analytics', {
                    category: 'IAP',
                    action: 'bought_' + productId
                });
                // amount of IAP's purchased along with the name of the IAP being purchased
                EventSystem.fire('Analytics', {
                    category: 'IAP',
                    action: 'iapAmount_' + iapAmount,
                    label: productId
                });
                EventSystem.fire('Analytics', {
                    category: 'UI',
                    action: 'bought_' + productId
                });

                // FireBase.setUserProperty('iapAmount', iapAmount);
                // FireBase.setUserProperty(productId, productAmount);
            };
            if (isRestoring) {
                iapOverlay.setText(Localization.getText('restored'));
            } else {
                iapOverlay.setText(Localization.getText('thanks'));
            }
            iapOverlay.exit(60, function () {
                if (onPurchaseComplete) {
                    onPurchaseComplete(true, id, purchaseInfo);
                }
            });
            // clean up event listeners
            cleanUp();

            // increment the amount of IAP's purchased
            if (!isRestoring) {
                sendAnalytics();
            }
        };
        var onPurchaseError = function (err) {
            console.log('purchase failed!', JSON.stringify(err));
            EventSystem.fire('Analytics', {
                category: 'UI',
                action: 'purchaseError',
                label: JSON.stringify(err)
            });
            iapOverlay.setText(Localization.getText('error') + '\n\n' + JSON.stringify(err));
            iapOverlay.exit(120, function () {
                if (onPurchaseComplete) {
                    onPurchaseComplete(false, id, null);
                }
            });
            cleanUp();
        };
        var cleanUp = function () {
            EventSystem.off('restore_success', onPurchaseSuccess);
            EventSystem.off('restore_error', onPurchaseError);
            EventSystem.off('purchase_success', onPurchaseSuccess);
            EventSystem.off('purchase_error', onPurchaseError);
        };

        if (isRestoring) {
            restore();
        } else {
            purchase();
        }
        return iapOverlay;
    };
    /*
     * Restore flow
     * Note that Cocoon doesn't actually have a restore flow: Restoring fires the purchase callbacks again
     * @snippet IapOverlay.Restore|snippet
IapOverlay.Restore({
    z: 0,
    fontSettings: {},
    onPurchaseComplete: function () {}
});
     */
    IapOverlay.Restore = function (settings) {
        settings.restore = true;
        return IapOverlay.Purchase(settings);
    };

    /*
     * Get the price string
     * @snippet IapOverlay.getPrice|snippet
IapOverlay.getPrice(${1:productId}, ${2:'\$0.99'})
     */
    IapOverlay.getPrice = function (productId, defaultPrice) {
        var i;
        var id = productId;
        var products = Store.getProducts();
        if (products) {
            // search
            for (i = 0; i < products.length; ++i) {
                if (products[i].productId === id) {
                    return products[i].localizedPrice;
                }
            }
        }
        // if products can't be loaded, try price from save state
        return Bento.saveState.load(id + 'Price', defaultPrice);
    };

    return IapOverlay;
});