/**
 * Shows how close the player is to a boss, and also where they are relative to their best run.
 * Also reponsible for spawning the Boss Fight / Boss Cleared overlays, and BossHealthBar  (TODO: refactor?)
 * @moduleName ProgressBar
 * @snippet ProgressBar.snippet
ProgressBar({
    position: new Vector2(0, 0)
})
 */
bento.define('gui/progressbar', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/counter',
    'bento/utils',
    'bento/tween',
    'globals',
    'entities/scoremanager',
    'gui/gametext',
    'gui/bossintrooutrotext',
    'bento/canvas',
    'bento/components/fill',
    'gui/bossoverlay',
    'gui/bosshealthbar'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Counter,
    Utils,
    Tween,
    Globals,
    ScoreManager,
    GameText,
    BossIntroOutroText,
    Canvas,
    Fill,
    BossOverlay,
    BossHealthBar
) {
    'use strict';

    return function (settings) {
        var viewport = Bento.getViewport();
        var position = settings.position;

        var targetScoreToBoss = 0;
        var targetScoreToOldBoss = 0;
        var highScore = 0;
        var currentScore = 0;
        var currentDistance = 0;
        var progress = 0.0; // 0.0 = empty bar,  1.0 = full bar
        var progressCatchupRate = 1.0; // how quickly do we approach the 'true' progress
        var tweeningProgress = false; // if progress is being tweened, we shouldn't update it manually
        var currentRunIsBest = false;
        var barWidth = 500;
        var centerOffset = 0; //-20;
        position.x -= barWidth / 2 - centerOffset;
        if (Utils.isWideScreen) {
            position.y += 80;
        }

        // allows us to slide off the screen for a boss fight, then come back afterwards.
        var startPosition = position.clone();

        var barEmpty = new Entity({
            position: new Vector2(0, 0),
            components: [
                new Sprite({
                    imageName: 'ui/progressbar-empty',
                    originRelative: new Vector2(0, 0.5),
                })
            ]
        });
        var barFullSprite = new Sprite({
            imageName: 'ui/progressbar-full',
            originRelative: new Vector2(0, 0),
        });
        barFullSprite.draw = function (data) {
            if (!this.currentAnimation || !this.visible) {
                return;
            }
            this.updateFrame();

            var width = Math.round(progress * barFull.dimension.width);
            if (!width) {
                return;
            }
            data.renderer.drawImage(
                this.spriteImage,
                this.sourceX,
                this.sourceY,
                width,
                this.frameHeight,
                (-this.origin.x) | 0,
                (-this.origin.y) | 0,
                width,
                this.frameHeight
            );
        };
        var barFull = new Entity({
            position: new Vector2(0, -17),
            components: [
                barFullSprite
            ]
        });

        var iconPlayer = new Entity({
            position: new Vector2(0, 0),
            components: [
                new Sprite({
                    imageName: 'ui/icon-player',
                    originRelative: new Vector2(0.5, 0.5),
                })
            ]
        });
        var iconBest = new Entity({
            position: new Vector2(0, 0),
            components: [
                new Sprite({
                    imageName: 'ui/icon-best',
                    originRelative: new Vector2(0.5, 0.5),
                })
            ]
        });
        var iconBoss = new Entity({
            position: new Vector2(0, 0),
            components: [
                new Sprite({
                    imageName: 'ui/icon-boss',
                    originRelative: new Vector2(0.5, 0.5),
                })
            ]
        });
        var scoreText = new Counter({
            name: 'gameText',
            value: 0,
            imageName: 'numbers',
            frameCountX: 10,
            frameCountY: 2,
            padding: 0,
            animations: {
                '0': {
                    frames: [0]
                },
                '1': {
                    frames: [1]
                },
                '2': {
                    frames: [2]
                },
                '3': {
                    frames: [3]
                },
                '4': {
                    frames: [4]
                },
                '5': {
                    frames: [5]
                },
                '6': {
                    frames: [6]
                },
                '7': {
                    frames: [7]
                },
                '8': {
                    frames: [8]
                },
                '9': {
                    frames: [9]
                },
                '-': {
                    frames: [10]
                },
                'B': {
                    frames: [11]
                },
                'O': {
                    frames: [12]
                },
                'S': {
                    frames: [13]
                },
                '!': {
                    frames: [14]
                }
            },
            align: 'center',
            spacing: new Vector2(-16, 0),
            position: new Vector2(barWidth / 2 - centerOffset, 60),
            scale: new Vector2(1.5, 1.5)
        }).extend({
            setText: function (value) {
                scoreText.setValue(value);
            },
            getText: function () {
                return scoreText.getValue();
            }
        });

        var bossHealthBar = null;

        var behavior = {
            start: function (data) {
                barEmpty.scale.x = barWidth / barEmpty.dimension.width;
                iconBoss.position.x = barWidth;
                targetScoreToBoss = ScoreManager.getTargetScoreToBoss();
                highScore = ScoreManager.getHighScore();
                if (highScore < targetScoreToBoss) {
                    iconBest.position.x = (highScore / targetScoreToBoss) * barWidth;
                } else {
                    iconBest.alpha = 0;
                }
                EventSystem.on('startBoss', onStartBoss);
                EventSystem.on('bossCleared', onBossCleared);
                EventSystem.on('targetScoreUpdated', onTargetScoreUpdated);
            },
            destroy: function (data) {
                EventSystem.off('startBoss', onStartBoss);
                EventSystem.off('bossCleared', onBossCleared);
                EventSystem.off('targetScoreUpdated', onTargetScoreUpdated);
            },
            update: function (data) {
                currentScore = ScoreManager.getVisualScore();
                currentDistance = ScoreManager.getDistancePoints();
                if (!tweeningProgress) {
                    var targetProgress = (currentDistance - targetScoreToOldBoss) / (targetScoreToBoss - targetScoreToOldBoss);
                    var progressDiff = targetProgress - progress;
                    progress += progressDiff * 0.1;
                }
                var xPos = progress * barWidth;
                iconPlayer.position.x = xPos;

                // scoreText.setText(currentScore);
                scoreText.setValue(Utils.approach(scoreText.getValue(), currentScore, 7));

                if (currentRunIsBest) {
                    iconBest.position.x = iconPlayer.position.x;
                } else if (currentScore >= highScore) {
                    currentRunIsBest = true;
                    iconBest.alpha = 1.0;
                    new Tween({
                        from: iconBest.position.y,
                        to: iconBest.position.y - 46,
                        in: 30,
                        ease: 'easeOutQuad',
                        onUpdate: function (v, t) {
                            iconBest.position.y = v;
                        }
                    });
                    new Tween({
                        from: 1,
                        to: 0.7,
                        in: 30,
                        ease: 'easeOutQuad',
                        onUpdate: function (v, t) {
                            iconBest.scale.x = iconBest.scale.y = v;
                        }
                    });
                }

            },
            draw: function (data) {}
        };

        var onStartBoss = function () {
            var player = Bento.objects.get('player');
            if (player && player.isAlive()) {
                Bento.objects.attach(new BossOverlay({
                    mode: 'boss-intro',
                    onComplete: function () {
                        bossHealthBar = new BossHealthBar({
                            position: new Vector2(viewport.width / 2, 200)
                        });
                        Bento.objects.attach(bossHealthBar);
                    }
                }));
            }
            new Tween({
                from: position.y,
                to: position.y - 400,
                in: 40,
                delay: 0,
                ease: 'easeInBack',
                onUpdate: function (v, t) {
                    position.y = v;
                }
            });
        };

        var onBossCleared = function () {
            var player = Bento.objects.get('player');
            if (player && player.isAlive()) {
                Bento.objects.attach(new BossOverlay({
                    mode: 'boss-cleared'
                }));
            }
            new Tween({
                from: position.y,
                to: startPosition.y,
                in: 40,
                delay: 200,
                ease: 'easeOutQuad',
                onUpdate: function (v, t) {
                    position.y = v;
                }
            });
            if (bossHealthBar) {
                bossHealthBar = null;
            }
        };

        // like bossCleared, except we're guaranteed that the ScoreManager will be up-to-date
        var onTargetScoreUpdated = function () {

            // wait for exp overlay to finish
            Utils.timeout(160, function () {
                targetScoreToOldBoss = targetScoreToBoss;
                targetScoreToBoss = ScoreManager.getTargetScoreToBoss();

                // tween progress to the beginning
                new Tween({
                    from: 1,
                    to: 0,
                    in: 60,
                    ease: 'easeInOutQuart',
                    onStart: function () {
                        tweeningProgress = true;
                    },
                    onUpdate: function (v, t) {
                        progress = v;
                    },
                    onComplete: function () {
                        // slowly allow progress to reach its true value again
                        tweeningProgress = false;
                        new Tween({
                            from: 0,
                            to: 1,
                            in: 600,
                            onUpdate: function (v, t) {
                                progressCatchupRate = v;
                            }
                        });
                    }
                });

                if (!currentRunIsBest && highScore > targetScoreToOldBoss && highScore < targetScoreToBoss) {
                    new Tween({
                        from: iconBest.position.x,
                        to: ((highScore - targetScoreToOldBoss) / (targetScoreToBoss - targetScoreToOldBoss)) * barWidth,
                        in: 60,
                        ease: 'easeInOutQuart',
                        onUpdate: function (v, t) {
                            iconBest.position.x = v;
                        }
                    });
                    iconBest.alpha = 1.0;
                }
            });
        };

        var entity = new Entity({
            z: Globals.Layers.GUI,
            name: 'progressBar',
            position: position,
            updateWhenPaused: 0,
            float: true,
            components: [
                barEmpty,
                barFull,
                iconPlayer,
                iconBoss,
                // iconBest,
                scoreText,
                behavior
            ]
        });

        return entity;
    };
});