/**
 * Save Manager.
 * Wrapper for Bento's SaveState.
 * New savekeys need to be defined here, paired with a default value. When using the SaveManager, it will check wether the provided key is defined, and throw a warning otherwise.
 * It will also use the paired default value, ensuring consistency between all places in the codebase where a value is loaded.
 *
 * @moduleName SaveManager
 * @snippet SaveManager.load
SaveManager.load('$1');
 * @snippet SaveManager.save
SaveManager.save('$1', ${2:value});
 * @snippet SaveManager.add
SaveManager.add('$1', ${2:value});
 */
bento.define('managers/save', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    /**
     * Save keys
     */
    var Keys = {};
    // How many times the player has opened the game.
    Keys.bootups = 0;
    // Number - Saw "hold to jump" tutorial, is a Number to count how many times it was seen
    Keys.tutorialHoldToJump = 0;
    // Number - Total coins in player's possession
    Keys.coins = 0;
    // Number - Number of coins the player collected in their current run.
    Keys.gainedCoins = 0;
    // String - Currently selected skin
    Keys.skin = 'redfish';
    // String - Currently selected world
    Keys.currentWorld = 'world-grass';
    // Array of Strings - unlocked skins
    Keys.unlockedSkins = ['redfish'];
    // Array of Strings - unlocked planet skins
    Keys.unlockedPlanetSkins = ['world-grass'];
    // String - Currently selected planet skin
    Keys.planetSkin = 'world-grass';
    // Number - Last known (server) time 
    Keys.lastKnownTime = 0;
    // Number - difference between user's time and server timne
    Keys.timeOffset = null;
    // Number - The (server) time at which the player started playing this game
    Keys.initialStartingTime = 0;
    // Number - Number of game overs
    Keys.gameOvers = 0; 
    // Number - Number of free gift received in total
    Keys.receivedGifts = 0; 
    // Number - The scheduled time when user can claim free gift
    Keys.scheduledGiftTime = 0; 
    // Number - Player should be able to use revive for ad only once, resets to 0 on screens/main
    Keys.usedRevive = 0; 
    // Number - Number of ads watched to unlock Legendary skins
    Keys.watchedAdsForSkin = 0;
    // Boolean - Allowed notifications
    Keys.notificationsAllowed = false;
    // Number - Asked for notifications
    Keys.notificationsAsked = 0;
    // Object - Saved identifiers for local notifications, can be used to cancel them
    Keys.notifications = {
        /*
        [name, defined by us]: [identifier (String), defined by Cocoon]
        */
    };
    // Boolean - Will be set to true after 1st death 
    Keys.showBanner = false;

    // Number - Used to calculate how much experience is needed for the next reward.
    Keys.expLevel = 1;
    // Number - How much experience we have gained since the last time we leveled up
    Keys.storedExp = 0;
    // Boolean - did unlock recently, only used for the badge on the XP button
    Keys.unlockedXpRecently = false;
    // Array of Strings - unlocked exp items - DEPRECATED in favour of 'expItemTimesUnlocked-<item>'
    Keys.unlockedExpItems = [];
    // Number - Similar to Keys.usedRevive, but for the 'second chance' item
    Keys.usedSecondChance = 0; 
    // Boolean - Whether the GDPR view should be shown
    Keys.hasShownGDPRView = false;
    // Boolean - Set to false when the user manually revokes consent for personalised ads
    Keys.gdprConsent = true;

    /**
     * Base keys. These allow to set up constructed SaveKeys.
     * For example, instead of saving the upgrades for each character in a big monolithic object, we can create a new save key for each character,
     * like 'upgrades-boy'
     */
    var Base = {};
    // Number - highscore for each world
    Base.highScore = 0;
    // Number - Number of bosses cleared per world
    Base.totalBossesCleared = 0; 
    // Boolean - Made purchase
    Base.purchased = false; 
    // Number - Number of deaths per-world (note: if the player revives, it doesn't count towards the death count)
    Base.totalDeaths = 0;
    // Number - How many times has a certain exp-item been unlocked?
    Base.expItemTimesUnlocked = 0;

    /**
     * Private functions
     */
    var getParsedDefault = function (def) {
        if (def !== null) {
            if (def.constructor === Array) {
                def = [].concat(def);
            } else if (typeof def === 'object') {
                def = JSON.parse(JSON.stringify(def));
            }
        }
        return def;
    };

    var checkKey = function (key) {
        var base;
        var def;
        if (Keys.hasOwnProperty(key)) {
            return true;
        } else {
            //check if this is a constructed key
            if (key.indexOf('-') !== -1) {
                base = key.split('-')[0];
                if (key === base + '-') {
                    Utils.log('only base part of key provided!');
                    return false;
                }
                if (Base.hasOwnProperty(base)) {
                    //we now know this to be a valid (constructed) key, so add it to the Keys list
                    def = getParsedDefault(Base[base]);
                    Keys[key] = def;
                    return true;
                }
            }
            Utils.log('WARNING: save key ' + key + ' is not defined!');
            return false;
        }
    };

    var save = function (key, value) {
        if (!checkKey(key)) {
            return;
        }
        Bento.saveState.save(key, value);
    };
    var load = function (key) {
        if (!checkKey(key)) {
            return undefined;
        }
        var def = getParsedDefault(Keys[key]);

        return Bento.saveState.load(key, def);
    };

    /**
     * Public functions
     */
    var manager = {
        save: save,
        load: load,
        add: function (key, value) {
            var current = load(key);
            current += value;
            save(key, current);
        }
    };

    return manager;
});