/**
 * Time manager, manages the time for energy regen and other timed events
 *
 * @moduleName TimeManager
 * @snippet TimeManager.getTime
TimeManager.getTime();
 * @snippet TimeManager.getLocalTime
TimeManager.getLocalTime();
 */
bento.define('managers/time', [
    'bento',
    'utils',
    'bento/entity',
    'managers/save',
    'bento/eventsystem'
], function (
    Bento,
    Utils,
    Entity,
    SaveManager,
    EventSystem
) {
    'use strict';

    var started = false;

    var counter = 0;
    var counterTotal = 0;

    var counterBehavior = {
        start: function () {
            console.log('globalTimer started');
        },
        destroy: function () {
            console.log('Warning: attempted to remove global timer');
            // put self back??
        },
        now: Date.now(),
        update: function (data) {
            var deltaT;
            var now;

            counter++;

            if (counter === 60) {
                // get delta time
                now = Date.now();
                deltaT = now - counterBehavior.now;

                // limit deltaT to 10seconds to be sure
                if (deltaT > 10000) {
                    deltaT = 10000;
                }

                // update now
                counterBehavior.now = now;
                manager.time += deltaT;
                counter = 0;
                counterTotal += 1;
                // save time every 60 second
                if (counterTotal % 60 === 0) {
                    manager.saveTime();
                }
            }
        }
    };
    var timerEntity = new Entity({
        z: 0,
        name: 'globalTimer',
        global: true,
        updateWhenPaused: Infinity,
        components: [counterBehavior]
    });

    var initialize = function () {
        if (!started) {
            manager.time = SaveManager.load('lastKnownTime');
            Bento.objects.attach(timerEntity);
            getServerTime(serverComplete, serverFail);


            if (Utils.isDev()) {
                window.timeManager = manager;
            }

            document.addEventListener('resume', function () {
                //Resume timemanager
                manager.resumeEvent();
            });

            started = true;
        }
    };
    var isRequestingTime = false;
    var getServerTime = function (onComplete, onFail) {
        var dateObject;
        var date;
        var todayNow = Date.now();
        var today = Math.floor(todayNow / 1000 / 60 / 60 / 24);
        var getAwsTime = function () {
            var xhr = new window.XMLHttpRequest();
            xhr.open("GET", "https://fcqd2ecsh5.execute-api.us-east-1.amazonaws.com/prod/getDate");
            xhr.setRequestHeader("Content-Type", "application/json;charset=UTF-8");
            xhr.onreadystatechange = function () {
                if (xhr.readyState === 4) {
                    if (xhr.status >= 200 && xhr.status < 300) {
                        try {
                            // get date from server
                            dateObject = JSON.parse(xhr.responseText);
                            date = dateObject.date;

                            // calculate new today
                            todayNow = date;
                            today = Math.floor(todayNow / 1000 / 60 / 60 / 24);

                            console.log("Received date from server: " + date);
                            onComplete(date);
                        } catch (e) {
                            if (onFail) {
                                Utils.log('ERROR: could not retrieve or parse date from server.' + xhr.responseText);
                                // onFail();
                                getHeaderTime();
                            }
                        }
                    } else {
                        if (onFail) {
                            Utils.log('ERROR: bad server response.');
                            // onFail();
                            getHeaderTime();
                        }
                    }
                }
            };
            // xhr.ontimeout = function (e) {
            //     if (onFail) {
            //         Utils.log('ERROR: timed out.');
            //         // onFail();
            //         getHeaderTime();
            //     }
            // };
            xhr.send();
        };
        var getHeaderTime = function () {
            var dateString;
            var xhr = new window.XMLHttpRequest();
            xhr.open('HEAD', 'https://s3.amazonaws.com/lucky-kat/beatstreet/data/date.json', true);
            xhr.onreadystatechange = function () {
                if (xhr.readyState === 4) {
                    if (xhr.status >= 200 && xhr.status < 300) {
                        try {
                            // get date from server
                            dateString = this.getResponseHeader("Date");
                            if (dateString.indexOf('GMT') === -1) {
                                dateString += ' GMT';
                            }
                            date = Date.parse(dateString);

                            // calculate new today
                            todayNow = date;
                            today = Math.floor(todayNow / 1000 / 60 / 60 / 24);

                            console.log('Date from server: ' + dateString + ' -> ' + date);

                            // headers received
                            onComplete(date);
                        } catch (e) {
                            if (onFail) {
                                Utils.log('ERROR: could not retrieve or parse date from server.');
                                onFail();
                            }
                        }
                    } else {
                        // should be 200 on success
                        if (onFail) {
                            onFail();
                        }
                    }
                }
            };
            xhr.send();
        };

        manager.hasUpdatedTime = false;
        if (!isRequestingTime) {
            isRequestingTime = true;
            console.log("Getting server time...");
            getAwsTime();
        }
    };

    var serverComplete = function (time) {
        isRequestingTime = false;
        manager.online = true;
        manager.time = time;
        manager.hasUpdatedTime = true;
        manager.saveTime();

        // we save the time difference between server and local once
        var timeOffset = SaveManager.load('timeOffset');
        if (timeOffset === null) {
            timeOffset = manager.time - Date.now();
            SaveManager.save('timeOffset', timeOffset);
        }

        // save the initial playing time (note: not accurate for softlaunch players)
        var initialStartingTime = SaveManager.load('initialStartingTime');
        if (initialStartingTime === 0) {
            initialStartingTime = time;
            SaveManager.save('initialStartingTime', initialStartingTime);
        }

        EventSystem.fire('serverTimeUpdated');
    };

    var serverFail = function () {
        var currentScreen = Bento.screens.getCurrentScreen();
        manager.online = false;
        isRequestingTime = false;

        if (!manager.hasWarned && currentScreen && currentScreen.name !== 'screens/preloader') {
            // warn only once per session
            manager.hasWarned = true;
            // EventSystem.fire('Notification', {
            //     textId: 'warningOffline'
            // });
        }

        // Date.now can be manipulated by changing the system clock, not reliable
        if (Bento.isDev()) {
            console.log('WARNING: using local time (dev only)');
            manager.time = Date.now();
        } else {
            // try again in 10 seconds
            window.setTimeout(function () {
                manager.updateServerTime();
            }, 10000);
        }
    };

    var getTime = function () {
        return manager.time;
    };

    var resumeEvent = function () {
        manager.hasWarned = false;
        // update now because there can be a big gap between the counterBehavior.now and the real time
        counterBehavior.now = Date.now();
        // update server time
        getServerTime(serverComplete, serverFail);
    };

    var manager = {
        hasWarned: false,
        online: false,
        time: SaveManager.load('lastKnownTime'),
        elapsed: 0,
        initialize: initialize,
        getTime: getTime,
        hasUpdatedTime: false,
        getLocalTime: function () {
            // retrieves server time corrected with offset
            var offset = SaveManager.load('timeOffset') || 0;
            return manager.time - offset;
        },
        resumeEvent: resumeEvent,
        saveTime: function () {
            // save last known time
            SaveManager.save('lastKnownTime', manager.time);
        },
        updateServerTime: function (onComplete) {
            // update server time
            getServerTime(function (date) {
                if (onComplete) {
                    onComplete(true);
                }
                serverComplete(date);
            }, function () {
                if (onComplete) {
                    onComplete(false);
                }
                serverFail();
            });
        }
    };

    return manager;
});