/**
 * Used by the chunkLoader to choose the next chunk to load.
 * (gives a random json asset name based on criteria such as bosses cleared, distance travelled)
 * @moduleName ChunkSelector
 */
bento.define('modules/chunkselector', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'globals',
    'entities/scoremanager',
    'managers/save'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Globals,
    ScoreManager,
    SaveManager
) {
    'use strict';

    // cooldown mechanism to prevent the same category from being selected too often
    var cooldowns = {};  // mapping of category names to remaining turns until it can be picked
    var decreaseCooldowns = function () {
        for (var k in cooldowns) {
            cooldowns[k] = Math.max(0, cooldowns[k]-1);
        }
    };
    var resetCooldowns = function () {
        for (var k in cooldowns) {
            cooldowns[k] = 0;
        }
    };
    var getCooldown = function (category) {
        return cooldowns[category.name] || 0;
    };

    var consecutiveCategory = null;
    var consecutiveCount = 0;

    // for each category, we try to never pick the same chunk twice in a row
    var previousChunks = {};  // mapping of category names to chunk names

    EventSystem.on('startLevel', function () {
        resetCooldowns();
        previousChunks = {};
    });

    /**
     * Pick a random chunk from a list of categories of chunks
     * (see Globals.levels for category definitions)
     */
    var getChunkFromCategoryList = function (categories) {
        var bossesCleared = ScoreManager.getBossesCleared();
        var distanceTraveled = ScoreManager.getDistancePoints();
        var i, cat, forcedCat;
        var totalLikeliness = 0;

        var availableCategories = categories.filter(function (c) {
            if (c.firstBoss) {
                // force this category if we didn't beat any bosses in the world yet.
                if (SaveManager.load('totalBossesCleared-'+Globals.currentWorld) === 0) {
                    forcedCat = c;
                }
                return false;
            }
            if (consecutiveCategory === c && consecutiveCount > 0) {
                forcedCat = c;
                return true;
            } 
            return distanceTraveled >= c.afterDistance
                && bossesCleared >= c.afterBosses
                && distanceTraveled < c.beforeDistance
                && bossesCleared < c.beforeBosses
                && getCooldown(c) === 0;
        });
        decreaseCooldowns();

        if (forcedCat) {
            cat = forcedCat;
        } else {
            if (availableCategories.length === 0) {
                Utils.log('no valid chunk categories available: ' + categories);
                return null;
            }

            // sum the likeliness of all categories
            for (i = 0; i < availableCategories.length; i++) {
                cat = availableCategories[i];
                totalLikeliness += cat.likeliness;
            }

            // pick a random category (using likeliness as a weighting)
            var n = Math.random() * totalLikeliness;
            
            for (i = 0; i < availableCategories.length; i++) {
                cat = availableCategories[i];
                n -= cat.likeliness;
                if (n <= 0) {
                    break;
                }
            }
        }

        // if desired by the category, we can stop it from being picked again next time.
        cooldowns[cat.name] = cat.cooldown;

        // or we can ensure it's picked multiple times in a row
        if (consecutiveCategory === cat) {
            consecutiveCount--;
        } else if (cat.consecutive) {
            consecutiveCategory = cat;
            consecutiveCount = Utils.isNumber(cat.consecutive)
                             ? cat.consecutive
                             : Utils.getRandomRange(cat.consecutive[0], cat.consecutive[1]);
        } else {
            consecutiveCategory = null;
        }

        // get the names of all Tiled json assets in this category
        var assets = Bento.assets.getAssets();
        var chunkNames = Object.keys(assets.json).filter(function (k) {
            var parts = k.split('/');
            return parts.length === 3     // has correct file depth
                && parts[0] === 'chunks'  // is in 'chunks' folder
                && parts[1] === cat.name; // has correct category
        });

        if (chunkNames.length === 0) {
            Utils.log('no chunks available in category ' + cat.name);
            return null;
        }

        // ignore the previously selected chunk from this category, if it's not the only option
        if (chunkNames.length >= 2) {
            i = chunkNames.indexOf(previousChunks[cat.name]);
            if (i !== -1) {
                chunkNames.splice(i, 1);
            }
        }

        // return a random chunk name from the chosen category
        var chunk = chunkNames[Utils.getRandom(chunkNames.length)];
        previousChunks[cat.name] = chunk;
        return chunk;
    };

    return {
        getChunk: function (level) {
            return getChunkFromCategoryList(level.chunkCategories);
        },
        getBoss: function (level) {
            return getChunkFromCategoryList(level.bossCategories);
        }
    };
});