/**
 * CMPlay SDK
 * @moduleName CMPlay
 */
bento.define('modules/cmplay', [
    'bento',
    'bento/eventsystem',
    'bento/utils'
], function (
    Bento,
    EventSystem,
    Utils
) {
    'use strict';
    
    var nothing = function () {};

    var module = {
        initPromotionSdk: nothing,
        setEnableDebug: nothing,
        getCloudData: nothing,
        pullCloudConfigData: nothing,
        getIntValue: nothing,
        getLongValue: nothing,
        getBooleanValue: nothing,
        getDoubleValue: nothing,
        getStringValue: nothing,
        onResume: nothing,
        onPause: nothing,
        onGameExit: nothing,
        reportData: nothing,
        showGDPRView: nothing,
        isAgreeGDPRPrivacyPolicy: nothing,
        isAgreeGDPRAdStayInformed: nothing,
        saveGDPRAgreedAdStayInformed: nothing,
        saveGDPRAgreedPrivacyPolicy: nothing,
        checkIsGDPREnforcedCountry: nothing,
    };

    var isExiting = false;

    var cmplay = window.CMPPromotionUtilsJs;

    var availableCodes = 'ar|de|en|es|fr|it|ja|ko|pt|ru|zh|tr|th|in';
    
    var getLanguageCode = function () {
        var language = (window.navigator.userLanguage || window.navigator.language || 'en-US').toLowerCase();
        var code = language.substr(0, 2);

        if (availableCodes.indexOf(code) === -1) {
            Utils.log('Language "'+language+'" not supported by CMPlay cloud config. Defaulting to English.');
            return 'en';
        }
        if (code === 'zh') {
            if (language === 'zh-tw' || language.indexOf('-hant') !== -1) {
                return 'zh_tw';
            } else if (language === 'zh-hk' || language.indexOf('-hans') !== -1) {
                return 'zh_hk';
            } else {
                return 'zh_cn';
            }
        }
        return code;
    };

    if (cmplay) {
        
        // NOTE: CMPlay SDK never actually calls onSuccess or onError...

        // So I manually added `callbackContext.success();` at various points on the Java end.
        // This means onSuccess will always be called, even if something went wrong.
        
        module.initPromotionSdk = function (onSuccess, onError) {
            cmplay.initPromotionSdk(onSuccess, onError);
        };
        module.setEnableDebug = function (int, onSuccess, onError) {
            cmplay.setEnableDebug(int, onSuccess, onError);
        };
        module.getCloudData = function (arg0, arg1, onSuccess, onError) {
            cmplay.getCloudData(arg0, arg1, onSuccess, onError);
        };
        module.pullCloudConfigData = function (languageCode, onSuccess, onError) {
            if (!languageCode) {
                languageCode = getLanguageCode();
            }
            cmplay.pullCloudConfigData(languageCode, onSuccess, onError);
        };
        module.getIntValue = function (funcType, section, key, defaultVal, onSuccess, onError) {
            cmplay.getIntValue(funcType, section, key, defaultVal, onSuccess, onError);
        };
        module.getLongValue = function (funcType, section, key, defaultVal, onSuccess, onError) {
            cmplay.getLongValue(funcType, section, key, defaultVal, onSuccess, onError);
        };
        module.getBooleanValue = function (funcType, section, key, defaultVal, onSuccess, onError) {
            cmplay.getBooleanValue(funcType, section, key, defaultVal, onSuccess, onError);
        };
        module.getDoubleValue = function (funcType, section, key, defaultVal, onSuccess, onError) {
            cmplay.getDoubleValue(funcType, section, key, defaultVal, onSuccess, onError);
        };
        module.getStringValue = function (funcType, section, key, defaultVal, onSuccess, onError) {
            cmplay.getStringValue(funcType, section, key, defaultVal, onSuccess, onError);
        };
        module.onResume = function (onSuccess, onError) {
            cmplay.onResume(onSuccess, onError);
        };
        module.onPause = function (onSuccess, onError) {
            // avoids calling onPause when the app is due to exit
            if (!isExiting) {
                cmplay.onPause(onSuccess, onError);
            }
        };
        module.onGameExit = function (onSuccess, onError) {
            isExiting = true;
            cmplay.onGameExit(onSuccess, onError);
        };
        module.reportData = function (arg0, arg1, onSuccess, onError) {
            cmplay.reportData(arg0, arg1, onSuccess, onError);
        };

        module.showGDPRView = function (onSuccess, onError) {
            cmplay.showGDPRView(onSuccess, onError);
        };
        module.isAgreeGDPRPrivacyPolicy = function (onSuccess, onError) {
            cmplay.isAgreeGDPRPrivacyPolicy(onSuccess, onError);
        };
        module.isAgreeGDPRAdStayInformed = function (onSuccess, onError) {
            cmplay.isAgreeGDPRAdStayInformed(onSuccess, onError);
        };
        module.saveGDPRAgreedAdStayInformed = function (isAgree, onSuccess, onError) {
            cmplay.saveGDPRAgreedAdStayInformed(isAgree, onSuccess, onError);
        };
        module.saveGDPRAgreedPrivacyPolicy = function (isAgree, onSuccess, onError) {
            cmplay.saveGDPRAgreedPrivacyPolicy(isAgree, onSuccess, onError);
        };
        module.checkIsGDPREnforcedCountry = function (onSuccess, onError) {
            cmplay.checkIsGDPREnforcedCountry(onSuccess, onError);
        };

    } else {
        Utils.log('Did not find CMPlay SDK.');

        // if the SDK did not initialize, onError will always be called
        var callErrorArgument = function (n) {
            var onError = arguments[n];
            return function () {
                if (onError) {
                    onError('CMPlay SDK not available.');
                }
            };
        };
        module.initPromotionSdk = callErrorArgument(1);
        module.setEnableDebug = callErrorArgument(2);
        module.getCloudData = callErrorArgument(3);
        module.pullCloudConfigData = callErrorArgument(2);
        module.getIntValue = callErrorArgument(5);
        module.getLongValue = callErrorArgument(5);
        module.getBooleanValue = callErrorArgument(5);
        module.getDoubleValue = callErrorArgument(5);
        module.getStringValue = callErrorArgument(5);
        module.onResume = callErrorArgument(1);
        module.onPause = callErrorArgument(1);
        module.onGameExit = callErrorArgument(1);
        module.reportData = callErrorArgument(3);
        module.showGDPRView = callErrorArgument(1);
        module.isAgreeGDPRPrivacyPolicy = callErrorArgument(1);
        module.isAgreeGDPRAdStayInformed = callErrorArgument(1);
        module.saveGDPRAgreedAdStayInformed = callErrorArgument(2);
        module.saveGDPRAgreedPrivacyPolicy = callErrorArgument(2);
        module.checkIsGDPREnforcedCountry = callErrorArgument(1);
    }

    return module;
});