/**
 * Definition of exp items
 * @moduleName ExpItems 
})
 */
bento.define('modules/expitems', [
    'bento',
    'bento/eventsystem',
    'bento/utils',
    'modules/localization',
    'managers/save',
    'globals'
], function (
    Bento,
    EventSystem,
    Utils,
    Localization,
    SaveManager,
    Globals
) {
    'use strict';

    var items = [];
    var itemsByName = {};

    // Bring over unlocks from old save files.
    // If an item exists in unlockedExpItems, that's equivalent to SaveManager.load('expItemTimesUnlocked-'+item.name) == 1 
    var applyLegacyUnlocks = function () {
        var unlockedExpItems = SaveManager.load('unlockedExpItems');
        var i, name;
        for (i=0; i<unlockedExpItems.length; i++) {
            name = unlockedExpItems[i];
            SaveManager.save('expItemTimesUnlocked-'+name, 1);
        }
        // clear the array and never use it again.
        SaveManager.save('unlockedExpItems', []);
    };

    applyLegacyUnlocks();


    var defineItem = function (name, settings) {
        var icon = settings.icon;
        var hidden = settings.hidden === true;
        var text = hidden ? '' : Localization.getText(settings.text).toUpperCase();
        var values = settings.values || [''];
        var apply = settings.apply;
        
        var item = {
            name:   name,    // for saving and lookup
            text:   text,    // localized text
            values:  values, // values displayed above text (just descriptive, but array length == how many times the item can be unlocked)
            hidden: hidden,  // excluded from the exp screen (e.g. for planet unlocks)
            icon:   icon,

            timesUnlocked: SaveManager.load('expItemTimesUnlocked-'+name),

            isUnlocked: function () {
                return item.timesUnlocked > 0;
            },
            canBeUnlocked: function () {
                return item.timesUnlocked < item.values.length;
            },

            unlock: function () {
                if (!item.canBeUnlocked()) {
                    return;
                }
                if (item.timesUnlocked === 0) {
                    SaveManager.save('unlockedXpRecently', true);
                }
                item.timesUnlocked += 1;
                SaveManager.save('expItemTimesUnlocked-'+name, item.timesUnlocked);
                if (apply) {
                    apply.call(item); // global effect
                }
                // could fire events here?
            }
        };
        items.push(item);
        itemsByName[name] = item;

        // immediately apply passive effect if we already unlocked this powerup
        if (item.isUnlocked() && apply) {
            apply.call(item);
        }
    };


    var genValues = function (from, to, steps) {
        var arr = [];
        for (var n = from; n <= to; n += (steps || 1)) {
            arr.push(n);
        }
        return arr;
    };
    var appendPercent = function (n) { return n+'%'; };
    var prependPlus   = function (n) { return '+'+n; };
    
    // The player will unlock items in the same order as they're defined here.
    // If all have been unlocked, the game tries to upgrade the one which has been unlocked the least number of times.

    defineItem('secondChance', {
        values: genValues(5, 20).map(appendPercent),   // ['5%', '6%', '7%'  ... '20%']
        text: 'xpSecondChance',
        icon: 'xp-ui/power-extralife',
        apply: function () {
            Globals.secondChanceProbability = 0.05 + 0.01 * (this.timesUnlocked-1);
        }
    });
    defineItem('trickJumpScore', {
        values: genValues(10, 25).map(appendPercent),   // ['10%', '11%', '12%'  ... '25%']
        text: 'xpTrickJumpScore',
        icon: 'xp-ui/power-trickjump',
        apply: function () {
            Globals.extraTrickJumpMultiplier = 1.10 + 0.01 * (this.timesUnlocked-1);
        }
    });
    defineItem('extraCoinsFromAds', {
        values: genValues(20, 100, 5).map(prependPlus),   // ['+20', '+25', '+30'  ... '+100']
        text: 'xpExtraCoinsFromAds',
        icon: 'xp-ui/power-adcoin',
        apply: function () {
            Globals.extraAdCoins = 20 + 5 * (this.timesUnlocked-1);
        }
    });
    defineItem('extraCoinsFreeGift', {
        values: genValues(20, 100, 5).map(prependPlus),   // ['+20', '+25', '+30'  ... '+100']
        text: 'xpExtraCoinsFreeGift',
        icon: 'xp-ui/power-giftcoin',
        apply: function () {
            Globals.extraFreeGiftCoins = 20 + 5 * (this.timesUnlocked-1);
        }
    });

    var ExpItems = {
        totalItems: function () {
            return items.length;
        },
        getItem: function (name) {
            return itemsByName[name];
        },
        forEach: function (f) {
            items.forEach(f);
        },
        getNextUnlock: function () {
            var item = null;
            var leastUnlocks = Infinity;
            for (var i=0; i<items.length; i++) {
                if (items[i].canBeUnlocked() && items[i].timesUnlocked < leastUnlocks) {
                    item = items[i];
                    leastUnlocks = item.timesUnlocked;
                }
            }
            return item;
        }
    };

    return ExpItems;
});
