/**
 * @moduleName GameAnalytics

  * @snippet GameAnalytics.initialize
GameAnalytics.initialize(
    "gameKey",
    "secretKey",
    ["coins","gems"],
    ["blaster", "cookie"]
);

  * @snippet GameAnalytics.businessEvent
GameAnalytics.businessEvent("currency", amount, "itemType", "itemId", "cartType");

  * @snippet GameAnalytics.sourceEvent
GameAnalytics.sourceEvent("currency", amount, "sourceType", "sourceId");

  * @snippet GameAnalytics.sinkEvent
GameAnalytics.sinkEvent("currency", amount, "sinkType", "sinkId");

  * @snippet GameAnalytics.startProgressionEvent
GameAnalytics.startProgressionEvent("progression01", "progression02", "progression03");

  * @snippet GameAnalytics.endProgressionEvent
GameAnalytics.endProgressionEvent(didSucceed, "progression01", "progression02", "progression03", score);

  * @snippet GameAnalytics.debugErrorEvent
GameAnalytics.debugErrorEvent("message");

  * @snippet GameAnalytics.infoErrorEvent
GameAnalytics.infoErrorEvent("message");

  * @snippet GameAnalytics.warningErrorEvent
GameAnalytics.warningErrorEvent("message");

  * @snippet GameAnalytics.warningErrorEvent
GameAnalytics.warningErrorEvent("message");

  * @snippet GameAnalytics.criticalErrorEvent
GameAnalytics.criticalErrorEvent("message");

  * @snippet GameAnalytics.customEvent
GameAnalytics.customEvent("id", value);

  * @snippet GameAnalytics.setCustomDimension
GameAnalytics.setCustomDimension(index, "value");

  * @snippet GameAnalytics.setGender
GameAnalytics.setGender("gender");

  * @snippet GameAnalytics.setBirthYear
GameAnalytics.setBirthYear(year);

  * @snippet GameAnalytics.setFacebookId
GameAnalytics.setFacebookId("facebookId");

  * @snippet GameAnalytics.registerEvents
GameAnalytics.registerEvents();

  * @snippet GameAnalytics.enable
GameAnalytics.enable();

  * @snippet GameAnalytics.disable
GameAnalytics.disable();
 */
bento.define('modules/gameanalytics', [
    'bento',
    'bento/eventsystem',
    'bento/utils'
], function (
    Bento,
    EventSystem,
    Utils
) {
    'use strict';
    // blank module
    var nothing = function () {};
    var module = {
        active: false,
        initialize: nothing,
        businessEvent: nothing,
        sourceEvent: nothing,
        sinkEvent: nothing,
        startProgressionEvent: nothing,
        endProgressionEvent: nothing,
        debugErrorEvent: nothing,
        infoErrorEvent: nothing,
        warningErrorEvent: nothing,
        errorErrorEvent: nothing,
        criticalErrorEvent: nothing,
        customEvent: nothing,
        setCustomDimension: nothing,
        setGender: nothing,
        setBirthYear: nothing,
        setFacebookId: nothing,
        registerEvents: nothing,
        enable: nothing,
        disable: nothing
    };

    // - Shorthand and declarations
    var GameAnalytics = window.GameAnalytics; // contains functionality

    var isInitialized = false;

    var checkInitialized = function () {
        if (!isInitialized) {
            Utils.log("ERROR: Trying to use analytics without the module initialized");

            // don't warn again
            checkInitialized = nothing;
        }
    };

    // - Check if the JS file exists
    if (!Utils.isDefined(GameAnalytics)) {
        // doesn't exist return blank object
        Utils.log("ERROR: GameAnalytics not found!");
        return module;
    }


    // - Configures and Initializes GameAnalytics for this game
    module.initialize = function (gameKey, secretKey, currencyNames, itemNames, custom01Names, custom02Names, custom03Names) {
        // create a string that describes this platform and version
        var buildString = "";
        if (Utils.isAndroidDevice()) {
            buildString += "android ";
        } else if (Utils.isIosDevice()) {
            buildString += "ios ";
        } else {
            buildString += "web ";
        }
        buildString += window.VERSION;

        // Define this game version by the buildString
        console.log('Configuring GameAnalytics for: "' + buildString + '"');
        GameAnalytics.configureBuild(buildString);

        // define currencies/items
        if (Utils.isDefined(currencyNames)) {
            console.log("Currencies: " + JSON.stringify(currencyNames));
            GameAnalytics.configureAvailableResourceCurrencies(currencyNames);
        }
        if (Utils.isDefined(itemNames)) {
            console.log("Items: " + JSON.stringify(itemNames));
            GameAnalytics.configureAvailableResourceItemTypes(itemNames);
        }
        if (Utils.isDefined(custom01Names)) {
            console.log("CustomResource01: " + JSON.stringify(custom01Names));
            GameAnalytics.configureAvailableCustomDimensions01(custom01Names);
        }
        if (Utils.isDefined(custom02Names)) {
            console.log("CustomResource02: " + JSON.stringify(custom02Names));
            GameAnalytics.configureAvailableCustomDimensions02(custom02Names);
        }
        if (Utils.isDefined(custom03Names)) {
            console.log("CustomResource03: " + JSON.stringify(custom03Names));
            GameAnalytics.configureAvailableCustomDimensions03(custom03Names);
        }

        // Initialize with this configuration
        GameAnalytics.initialize({
            gameKey: gameKey,
            secretKey: secretKey
        });

        console.log('GameAnalytics initialized!');
        isInitialized = true;
        module.active = true;
    };

    // - Functions to turn analytics on/off (i.e. ignoring all event calls)
    //   For example you might want to disable analytics when cheats are active
    module.enable = function () {
        module.active = true;
    };
    module.disable = function () {
        module.active = false;
    };

    // - Add business event to queue
        // currency = String currency code in ISO 4217 format. ("USD" = US Dollars) http://openexchangerates.org/currencies.json
        // amount = Integer value in cents (99 = $0.99).
        //          This is true even for currencies that don't typically use cents
        //          (e.g. JPY), so you can always multiply by 100 regardless of currency.
        // itemType = String category of the item ("GemPacks")
        // itemId = String specific Item Identifier ("1000GemsPack")
        // cartType = String location of where in the game it was bought ("shop")
    module.businessEvent = function (currency, amount, itemType, itemId, cartType, receipt, signature) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        if (Utils.isAndroidDevice()) {
            // google play
            GameAnalytics.addBusinessEvent({
                currency: currency, 
                amount: amount, 
                itemType: itemType, 
                itemId: itemId, 
                cartType: cartType, 
                receipt: receipt, 
                signature: signature
            });
        }
        if (Utils.isIosDevice()) {
            if (receipt) {
                // ios predefined receipt
                GameAnalytics.addBusinessEvent({
                    currency: currency, 
                    amount: amount, 
                    itemType: itemType, 
                    itemId: itemId, 
                    cartType: cartType, 
                    receipt: receipt
                });
            } else {
                // ios auto fetch receipt
                GameAnalytics.addBusinessEvent({
                    currency: currency, 
                    amount: amount, 
                    itemType: itemType, 
                    itemId: itemId, 
                    cartType: cartType, 
                    autoFetchReceipt: true
                });
            }
        }
    };


    // - Add resource events to queue
        // currency - String, one of the currencies defined on configuration ("gems")
        // value - Float amount of currency (100.0)
        // sourceType/sinkType = String category of the source or sink ("Upgrades")
        // sourceId/sinkId = String specific identifier of the source/sink ("characterUpradeSlot1")
    module.sourceEvent = function (currency, amount, sourceType, sourceId) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        GameAnalytics.addResourceEvent({
            flowType: GameAnalytics.EGAResourceFlowType.Source, 
            currency: currency, 
            amount: amount, 
            itemType: sourceType, 
            itemId: sourceId
        });
    };
    module.sinkEvent = function (currency, amount, sinkType, sinkId) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        GameAnalytics.addResourceEvent({
            flowType: GameAnalytics.EGAResourceFlowType.Sink, 
            currency: currency, 
            amount: amount, 
            itemType: sinkType, 
            itemId: sinkId
        });
    };


    // - Add progression events to queue
        // progression 01/02/03 - String of where the event took place, not all need to be defined
    module.startProgressionEvent = function (progression01, progression02, progression03) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        var event = {
            progressionStatus: GameAnalytics.EGAProgressionStatus.Start,
            progression01: progression01
        };

        // optional fields
        // note: We _must_ define optional fields conditionally like this
        //        otherwise the Cordova plugin will crash
        if (Utils.isDefined(progression02)) {
            event.progression02 = progression02;
        }
        if (Utils.isDefined(progression03)) {
            event.progression03 = progression03;
        }

        GameAnalytics.addProgressionEvent(event);
    };
        // didSucceed - Boolean of whether the player succeeded or not
        // progression 01/02/03 - String of where the event took place, not all need to be defined
        // value - an integer of the player score when they ended, optional
    module.endProgressionEvent = function (didSucceed, progression01, progression02, progression03, value) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        var event = {
            progressionStatus:didSucceed ? GameAnalytics.EGAProgressionStatus.Complete : GameAnalytics.EGAProgressionStatus.Fail, 
            progression01: progression01
        };

        // optional fields
        if (Utils.isDefined(progression02)) {
            event.progression02 = progression02;
        }
        if (Utils.isDefined(progression03)) {
            event.progression03 = progression03;
        }
        if (Utils.isDefined(value)) {
            event.value = value;
        }

        GameAnalytics.addProgressionEvent(event);
    };


    // - Add error events to queue
    module.debugErrorEvent = function (message) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        GameAnalytics.addErrorEvent({
            severity: GameAnalytics.EGAErrorSeverity.Debug, 
            message: message
        });
    };
    module.infoErrorEvent = function (message) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        GameAnalytics.addErrorEvent({
            severity: GameAnalytics.EGAErrorSeverity.Info, 
            message: message
        });
    };
    module.warningErrorEvent = function (message) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        GameAnalytics.addErrorEvent({
            severity: GameAnalytics.EGAErrorSeverity.Warning, 
            message: message
        });
    };
    module.errorErrorEvent = function (message) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        GameAnalytics.addErrorEvent({
            severity: GameAnalytics.EGAErrorSeverity.Error, 
            message: message
        });
    };
    module.criticalErrorEvent = function (message) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        GameAnalytics.addErrorEvent({
            severity: GameAnalytics.EGAErrorSeverity.Critical, 
            message: message
        });
    };


    // - Add design events
        // id - String id:heirarchy:is:seperated:like:this
        // value - Float value (optional)
    module.customEvent = function (id, value) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        var event = {
            eventId: id
        };
        if (Utils.isDefined(value)) {
            event.value = value;
        }

        GameAnalytics.addDesignEvent(event);
    };


    // - Set custom dimensions
        // - PERSISTS BETWEEN GAME LAUNCHES -
        // index - Integer 1, 2, or 3
        // value - Stringwith a value set in custom01Names, custom02Names, or custom02Names on initialization
    module.setCustomDimension = function (index, value) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        index = Utils.clamp(1, index, 3);
        switch (index) {
            case 1: GameAnalytics.setCustomDimension01(value); break;
            case 2: GameAnalytics.setCustomDimension02(value); break;
            case 3: GameAnalytics.setCustomDimension03(value); break;
        }
    };

    // - Demographics
        // - PERSISTS BETWEEN GAME LAUNCHES -
    module.setGender = function (gender) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        switch (gender) {
            case "male" : GameAnalytics.setGender(GameAnalytics.EGAGender.Male); break;
            case "female" : GameAnalytics.setGender(GameAnalytics.EGAGender.Female); break;
        }
    };
    module.setBirthYear = function (year) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        GameAnalytics.setBirthYear(year);
    };
    module.setFacebookId = function (facebookId) {
        checkInitialized();
        if (!module.active) {
            return;
        }
        GameAnalytics.setFacebookId(facebookId);
    };


    // Bento event handlers
    // - If desired, call this function once to set up event handlers,
    //    so that you don't need to import GameAnalytics in every module.
    // - Some handy synonyms are allowed as data fields
    //    e.g. 'sinkType' instead of 'itemType', 'id' instead of 'progression01'

    module.registerEvents = function () {

        EventSystem.on("GameAnalytics-business", function (data) {
            module.businessEvent(
                data.currency,
                data.amount,
                data.itemType,
                data.itemId,
                data.cartType,
                data.receipt,
                data.signature
            );
        });

        EventSystem.on("GameAnalytics-source", function (data) {
            module.sourceEvent(
                data.currency,
                data.amount,
                data.sourceType || data.itemType,
                data.sourceId || data.itemId
            );
        });

        EventSystem.on("GameAnalytics-sink", function (data) {
            module.sinkEvent(
                data.currency,
                data.cost || data.amount,
                data.sinkType || data.itemType,
                data.sinkId || data.itemId
            );
        });

        EventSystem.on("GameAnalytics-startProgression", function (data) {
            module.startProgressionEvent(
                data.id || data.progression01,
                data.progression02, // optional
                data.progression03  // optional
            );
        });

        EventSystem.on("GameAnalytics-endProgression", function (data) {
            module.endProgressionEvent(
                data.didSucceed,
                data.id || data.progression01,
                data.progression02,       // optional
                data.progression03,       // optional
                data.score || data.value  // optional
            );
        });

        // Custom (aka 'design') events - data can be a raw string if you don't care about the value

        EventSystem.on("GameAnalytics-custom", function (data) {
            if (typeof data === "string") {
                module.customEvent(data);
            } else {
                module.customEvent(
                    data.id,
                    data.value // optional
                );
            }
        });

        // Error events - data can be a raw string or an object with a message field

        var registerErrorEvent = function (name, errorFn) {
            EventSystem.on(name, function (data) {
                errorFn(typeof data === "string" ? data : data.message);
            });
        };
        registerErrorEvent("GameAnalytics-error", module.errorErrorEvent);
        registerErrorEvent("GameAnalytics-errorError", module.errorErrorEvent);
        registerErrorEvent("GameAnalytics-debugError", module.debugErrorEvent);
        registerErrorEvent("GameAnalytics-infoError", module.infoErrorEvent);
        registerErrorEvent("GameAnalytics-warningError", module.warningErrorEvent);
        registerErrorEvent("GameAnalytics-criticalError", module.criticalErrorEvent);
    };

    return module; 
});