/**
 * Create an array of planets from a Tiled JSON file.
 * Each object layer in the map counts as a separate planet.
 */
bento.define('modules/maploader', [
    'bento',
    'bento/utils',
    'entities/spike',
    'bento/entity',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/tiledreader',
    'bento/tiled',
    'components/planetwrap',
    'entities/shadowentity',
    'entities/planet',
    'globals',
    'bento/eventsystem',
    'bento/components/fill',
    'bento/components/eventlistener',
    'bento/tween',
    'entities/floor',
    'components/pathfollower',
    'entities/coin',
    'bento/components/sprite',
    'modules/pathmanager',
    'modules/chunkselector'
], function (
    Bento,
    Utils,
    Spike,
    Entity,
    Vector2,
    Rectangle,
    TiledReader,
    Tiled,
    PlanetWrap,
    ShadowEntity,
    Planet,
    Globals,
    EventSystem,
    Fill,
    EventListener,
    Tween,
    Floor,
    PathFollower,
    Coin,
    Sprite,
    PathManager,
    ChunkSelector
) {
    var SpawnTween = function () {
        return {
            start: function (data) {
                var entity = data.entity;
                new Tween({
                    ignoreGameSpeed: true,
                    from: 0.25,
                    to: entity.scale.x,
                    in: 30,
                    ease: 'elastic',
                    decay: 5,
                    oscillations: 2,
                    onStart: function () {
                        entity.scale.x = this.settings.from;
                        entity.scale.y = this.settings.from;
                    },
                    onUpdate: function (v, t) {
                        entity.scale.x = v;
                        entity.scale.y = v;
                    }
                });
            }
        };
    };
    var MapLoader = function (mapSettings) {
        var viewport = Bento.getViewport();
        var planetPerLayer = true;
        var level;
        var tiledMap;
        var position = new Vector2(viewport.width / 2, viewport.height / 2);
        var planets = [];
        var planetsMap = {};
        var planet, floor;
        var planetRadius = 0;
        var floorOffset; // how many pixels should the highest entity be raised above 0,0
        var tileSize = 128; // how big are the blocks on this planet
        var edgeMargin; // objects that lie beyond this margin with have shadow copies generated

        // To let us easily change the size of worlds, tileSize is a custom property,
        //  independent of the tileWidth/tileHeight used in the editor.
        // So we need a scale to correct the position of map objects.
        var gridScale;

        var pathManager;

        // clone of colliders near the edge wrap of planets
        var addShadowIfNearEdge = function (entity) {
            var shadow = null;
            if (chunkLoader.loadChunks && !chunkLoader.isBossFight) {
                return;
            }
            // remove shadow if original is removed
            entity.attach({
                name: 'shadowAttacher',
                start: function () {
                    // we can only check the position during start
                    // because then we are sure the positions are wrapped by planetWrap
                    if (entity.position.x < PlanetWrap.wrapOffset + edgeMargin) {
                        shadow = new ShadowEntity({
                            other: entity,
                            offset: new Vector2(planet.perimeter, 0)
                        });
                    } else if (entity.position.x >= PlanetWrap.wrapOffset + planet.perimeter - edgeMargin) {
                        shadow = new ShadowEntity({
                            other: entity,
                            offset: new Vector2(-planet.perimeter, 0)
                        });
                    }
                    if (shadow) {
                        shadow.attach(new PlanetWrap({
                            planet: planet
                        }));
                        Bento.objects.attach(shadow);
                    } else {
                        entity.remove(this);
                    }
                },
                destroy: function () {
                    if (shadow) {
                        shadow.removeSelf();
                    }
                }
            });
        };

        var getPlanetWidth = function (layerJson) {
            var object, objects;
            var layers;
            var k, kl;
            var width = tiledMap.width * tiledMap.tilewidth;
            var loopThroughLayer = function () {
                var i, len;
                objects = layerJson.objects || [];
                len = objects.length;
                for (i = 0; i < len; ++i) {
                    object = objects[i];
                    if (object.name && object.name === 'planetWidth') {
                        return object.x;
                    }
                }
                return null;
            };

            if (!layerJson) {
                // loop through all layers to find any planetWidth object
                layers = tiledMap.layers;
                for (k = 0, kl = layers.length; k < kl; ++k) {
                    layerJson = layers[k];
                    width = loopThroughLayer();
                    if (width !== null) {
                        break;
                    }
                }
            } else {
                width = loopThroughLayer();
            }

            return width;
        };
        var tileWidth;
        var levelWidth;
        
        // offsetX is the x tile offset at which chunks need to spawn
        // at the end of Tiled loading, the tile width of the map is added to offsetX to
        // ensure that the next map/chunk will be loaded at the end of that
        var offsetX = 0;

        // mainly used for analytics (to find out which chunk we died in)
        var bossChunkInfo = null;
        var activeChunksInfo = [
            // e.g.
            // { x: 0, assetName: 'chunks/grass-easy/foo' }
        ];

        /** 
         * Chunkloader is an object that must be attached to constantly load new chunks
         */
        var chunkLoader = {
            name: 'chunkLoader',
            objects: {},
            // constants
            tiledSize: 16,
            planetWidthX: 0,
            spawnX: 0, // tiles to spawn ahead
            deSpawnX: -2, // tiles to despawn behind
            gridScale: 1,

            // variables
            hasSpawnedX: 0, // keeps track at which x tile has attached objects
            hasDeSpawnedX: 0, // keeps track at which x tile has removed objects
            updatedX: 0, // updatedX is compared to player's x tile, to trigger spwan checks instead of every frame
            isBossFight: false, // should this be here?
            canLookAhead: false, // quick fix for boss spawning
            bossEnemies: 0,
            chunkUID: 0,  // appended to entity names to prevent conflicts (e.g. when 'btn1' exists in two different chunks)
            
            start: function () {},
            despawnAll: function (onComplete, leave) {
                var despawn = function () {
                    Utils.forEach(chunkLoader.objects, function (list) {
                        Utils.forEach(list, function (item, i, l, breakLoop) {
                            if (item.entity) {
                                // remove any deadlies and solids
                                item.entity.removeFromFamily('solid');
                                item.entity.removeFromFamily('deadly');

                                if (item.properties && item.properties.follow) {
                                    pathManager.releasePath(item.properties.follow);
                                }

                                new Tween({
                                    from: item.entity.scale.x,
                                    to: 0,
                                    in: 10,
                                    ease: 'linear',
                                    onUpdate: function (v, t) {
                                        item.entity.scale.x = v;
                                        item.entity.scale.y = v;
                                    },
                                    onComplete: function () {
                                        Bento.objects.remove(item.entity);
                                    }
                                });
                            }
                        });
                        list.length = 0;
                    });
                    chunkLoader.objects = {};

                };

                if (!leave) {
                    despawn();
                }

                // reset offsetX
                var currentX = chunkLoader.getCurrentX();
                // offsetX = currentX + chunkLoader.spawnX + 2;
                offsetX = currentX + Math.floor(levelWidth / tileWidth);
                // offsetX = currentX;

                chunkLoader.bossEnemies = 0;
                chunkLoader.canLookAhead = false;


                // Utils.timeout(10, onComplete);
                onComplete();
            },

            // make the world safe even before blocks have visibly started to despawn
            unsolidifyAll: function () {
                Utils.forEach(chunkLoader.objects, function (list) {
                    Utils.forEach(list, function (item, i, l, breakLoop) {
                        if (item.entity) {
                            item.entity.removeFromFamily('solid');
                            item.entity.removeFromFamily('deadly');
                            item.entity.getComponent('secondaryHitbox', function (secondaryHitbox) {
                                secondaryHitbox.removeFromFamily('deadly');
                            });
                        }
                    });
                });
            },

            spawnBossChunk: function (assetName) {
                // test: spawn boss fight immediately
                var currentX = chunkLoader.getCurrentX();
                offsetX = currentX;

                if (!assetName) {
                    assetName = ChunkSelector.getBoss(level);
                }

                bossChunkInfo = {
                    x: currentX,
                    assetName: assetName
                };

                chunkLoader.appendChunk(assetName, function () {
                    if (chunkLoader.isBossFight) {
                        var i, l, delay = 0;
                        for (i = 0, l = Math.floor(levelWidth / tileWidth); i < l; ++i) {
                            var toSpawn = chunkLoader.objects[currentX + i];
                            if (toSpawn && toSpawn.length) {
                                Utils.forEach(toSpawn, function (item) {
                                    if (item.entity && !item.entity.isAdded) {
                                        new Tween({
                                            // delay
                                            in: delay,
                                            ignoreGameSpeed: true,
                                            onComplete: function () {
                                                item.didSpawn = true;
                                                item.entity.attach(new SpawnTween({}));
                                                Bento.objects.attach(item.entity);
                                                addShadowIfNearEdge(item.entity);
                                            }
                                        });
                                    }
                                });
                                delay += 5;
                            }
                        }
                    }
                });

            },
            update: function () {
                // update current position
                var currentX = chunkLoader.getCurrentX();
                var toSpawnX = currentX + chunkLoader.spawnX; // the x tile at which objects will appear
                var toDeSpawnX = currentX - 4;
                var toSpawn;
                var toDeSpawn;

                if (activeChunksInfo[1] && currentX >= activeChunksInfo[1].x && !chunkLoader.isBossFight) {
                    activeChunksInfo.shift();
                }

                if (chunkLoader.updatedX !== currentX) {
                    chunkLoader.updatedX = currentX;
                    if (chunkLoader.isBossFight && currentX === Math.floor(PlanetWrap.wrapOffset / chunkLoader.gridScale / chunkLoader.tiledSize)) {
                        chunkLoader.canLookAhead = true;
                        chunkLoader.hasSpawnedX = toSpawnX;
                    }

                    // spawn
                    if (toSpawnX >= chunkLoader.hasSpawnedX) {
                        if (toSpawnX >= offsetX) {
                            if (!chunkLoader.isBossFight) {
                                // load non-boss chunks
                                chunkLoader.appendChunk(ChunkSelector.getChunk(level));
                            }
                        }
                        // spawn new entities
                        toSpawn = chunkLoader.objects[toSpawnX];
                        if (toSpawn && toSpawn.length && !chunkLoader.isBossFight) {
                            Utils.forEach(toSpawn, function (item, i, l, breakLoop) {
                                if (item.entity && !item.didSpawn && !item.entity.isAdded) {
                                    item.didSpawn = true;
                                    item.entity.attach(new SpawnTween({}));
                                    Bento.objects.attach(item.entity);
                                    addShadowIfNearEdge(item.entity);
                                }
                            });
                        }
                        // bugfix: also spawn outside of wrapping area
                        // once player passes through the end of the wrapping area, turn lookAhead true
                        /*if (chunkLoader.isBossFight && chunkLoader.canLookAhead) {
                            toSpawn = chunkLoader.objects[toSpawnX + Math.floor(levelWidth / tileWidth)];
                            if (toSpawn && toSpawn.length) {
                                Utils.forEach(toSpawn, function (item, i, l, breakLoop) {
                                    if (item.entity && !item.entity.isAdded) {
                                        item.didSpawn = true;
                                        item.entity.attach(new SpawnTween({}));
                                        Bento.objects.attach(item.entity);
                                        addShadowIfNearEdge(item.entity);
                                    }
                                });
                            }
                        }*/
                        chunkLoader.hasSpawnedX += 1;

                    }
                    // despawn
                    if (toDeSpawnX > chunkLoader.hasDeSpawnedX && !chunkLoader.isBossFight) {
                        toDeSpawn = chunkLoader.objects[toDeSpawnX];
                        if (toDeSpawn && toDeSpawn.length) {
                            Utils.forEach(toDeSpawn, function (item, i, l, breakLoop) {
                                if (item.entity && item.shouldDespawn) {
                                    new Tween({
                                        from: item.entity.scale.x,
                                        to: 0,
                                        in: 10,
                                        ease: 'linear',
                                        onUpdate: function (v, t) {
                                            item.entity.scale.x = v;
                                            item.entity.scale.y = v;
                                        },
                                        onComplete: function () {
                                            Bento.objects.remove(item.entity);
                                        }
                                    });
                                }
                            });
                        }
                        chunkLoader.hasDeSpawnedX += 1;

                        // delete column
                        delete chunkLoader.objects[toDeSpawnX];
                    }
                }
            },
            getCurrentX: function () {
                // overwrite this function
                return 0;
            },
            getCurrentChunkInfo: function () {
                return chunkLoader.isBossFight
                     ? bossChunkInfo
                     : activeChunksInfo[0];
            },
            appendChunk: function (assetName, callback) {
                // keep a copy of offsetX, because onSpawn will be called asynchronously
                var currentOffsetX = offsetX;
                var chunkUID = ++chunkLoader.chunkUID;

                new Tiled({
                    assetName: assetName,
                    drawTiles: false,
                    merge: false,
                    cacheModules: false,
                    spawnEntities: true, // require the module (asynchronously)
                    spawnBackground: false, // spawn background entities (drawn tile layers)
                    attachEntities: false, // attach after spawning
                    onObject: function (objectJSON, tilesetJSON, tileIndex) {
                        handleObject(objectJSON, chunkUID);
                    },
                    onSpawn: function (entity, object, tileData, layerIndex) {
                        onSpawn(entity, object, tileData, layerIndex, currentOffsetX);
                    },
                    onSpawnComplete: function () {
                        if (callback) {
                            callback();
                        }
                    },
                    onComplete: function () {
                        activeChunksInfo.push({
                            x: offsetX,
                            assetName: assetName
                        });
                        // update offset (synchronously)
                        offsetX += this.json.width;
                    }
                });
            },
            spawnBonusCoins: function () {
                // spawn coins in front of player
                var player = Bento.objects.get('player');
                var spawnDelay = 0;
                var spawnCoin = function (offset) {
                    var coin = new Coin({
                        position: new Vector2(player.position.x + offset.x, offset.y)
                    });
                    coin.attach(new PlanetWrap({
                        planet: planet
                    }));
                    Utils.timeout(spawnDelay, function () {
                        coin.attach({
                            name: 'despawn',
                            update: function () {
                                if (coin.position.x + tileSize * 4 < player.position.x) {
                                    coin.removeSelf();
                                }
                            }
                        });
                        Bento.objects.attach(coin);
                        new Tween({
                            from: 0,
                            to: 1,
                            in: 60,
                            ease: 'elastic',
                            decay: 4,
                            oscillations: 2,
                            onUpdate: function (v, t) {
                                coin.scale.x = v;
                                coin.scale.y = v;
                            }
                        });
                    });
                    spawnDelay += 10;
                };
                var i;

                if (player) {
                    for (i = 0; i < 10; ++i) {
                        spawnCoin(new Vector2(tileSize * 5 + tileSize * i * 2, -tileSize * (i % 2)));
                    }
                }
            }
        };
        
        // In endless mode, we modify the names in the JSON data to include the current chunkUID
        // This also should work if the data was previously mangled (because the JSON data is cached between loads of the same chunk)
        // This solution is really hacky, but less invasive than the alternatives.
        // e.g. btn1   in the first spawned chunk becomes btn1#1
        //      btn1#1 in the sixth spawned chunk becomes btn1#6
        var mangleName = function (object, key, chunkUID) {
            if (object[key]) {
                // un-mangle the name if it already has a postfix
                object[key] = object[key].split('#')[0];
                // append the new postfix
                object[key] += '#' + chunkUID;
            }
        };
        
        // parse JSON properties in the name of the object, and create hidden path indicators.
        var handleObject = function (object, chunkUID) {
            var pairs, pair, i;
            var props = object.properties;
            if (!props) {
                props = {};
                object.properties = props;
            }

            // extend object properties with those embedded in the name (a:b,c:d)
            if (object.name.indexOf(':') !== -1) {
                pairs = object.name.split(',');
                for (i = 0; i < pairs.length; i++) {
                    pair = pairs[i].split(':');
                    props[pair[0]] = pair[1];
                }
            } else {
                // perform mangling on this object's name
                if (chunkUID !== undefined) {
                    mangleName(object, 'name', chunkUID);
                }
            }

            // mangle the names that it refers to
            if (chunkUID !== undefined) {
                mangleName(props, 'toggle', chunkUID);
                mangleName(props, 'follow', chunkUID);
                mangleName(props, 'trigger', chunkUID);
            }

            if (object.polyline) {
                // this object is a path
                pathManager.createPath(object, '_'+chunkUID);
            }
        };

        /**
         * Load planet, layerJson is optional
         * If loading per layer, the props parameter can be ignored
         */
        var setupPlanet = function (layerJson, props, layerIndex) {
            if (layerJson.type !== 'objectgroup') {
                return;
            }
            var flip = false;

            // calculate the radius using the planetWidth object
            levelWidth = getPlanetWidth(layerJson);
            planetRadius = levelWidth / (Math.PI * 2) * gridScale;

            if (layerJson) {
                props = layerJson.properties || {};
                props.x = props.x || 0;
                props.y = props.y || 0;
                props.floorOffset = 8;
                floorOffset = props.floorOffset * tileSize;
            }

            flip = !!props.flip;
            if (Globals.FORCE_SPIN_DIRECTION === 'left') {
                flip = true;
            }
            if (Globals.FORCE_SPIN_DIRECTION === 'right') {
                flip = false;
            }

            planet = new Planet({
                radius: planetRadius,
                imageName: props.planetImage || 'planet-donut',
                position: new Vector2(props.x * 2, props.y * 2),
                tileSize: tileSize,
                playerSize: 128, // TODO: remove playerSize from the codebase
                flip: flip
            });
            planets.push(planet);

            if (Utils.isDefined(layerIndex)) {
                planetsMap[layerIndex] = planet;
            }
            floor = new Floor({
                planet: planet,
                yPosition: 0
            });
            Bento.objects.attach(floor);

            // setup chunk loader constants
            if (mapLoader.loadChunks) {
                chunkLoader.tiledSize = tileWidth;
                chunkLoader.planetWidthX = Math.floor(levelWidth / tileWidth);
                // spawn ahead at 3/4 of planet width
                chunkLoader.spawnX = Math.floor(chunkLoader.planetWidthX * 3 / 4);
                chunkLoader.gridScale = gridScale;

                planet.infinite = true;
            }
        };
        /**
         * Initialize data from Tiled
         */
        var load = function (settings) {
            var props;
            level = settings.level;
            tiledMap = Bento.assets.getJson(level.mapName);
            props = tiledMap.properties || {};

            mapLoader.loadChunks = Utils.isDefined(level.chunkCategories);

            position = settings.position || position;

            floorOffset = props.floorOffset * tileSize;
            edgeMargin = tileSize * 2;
            tileWidth = tiledMap.tilewidth;
            gridScale = tileSize / tileWidth;

            pathManager = new PathManager({
                gridScale: gridScale
            });

            if (!planetPerLayer) {
                setupPlanet(null, props);
            }

            if (mapLoader.loadChunks) {
                // first chunk is always the map itself, even in endless mode
                activeChunksInfo.push({ x: 0, assetName: level.mapName });
            }
        };
        /**
         * onSpawn used by Tiled, when an entity is generated, we intercept and correct for the planet wrapping etc.
         */
        var onSpawn = function (entity, object, tileData, layerIndex, currentOffsetX) {
            var shadow;
            var props = object.properties || {};
            var entityPos = new Vector2(
                (object.x + (currentOffsetX || 0) * tileWidth) * gridScale,
                object.y * gridScale - floorOffset
            );
            var tileX;
            var currentPlanet = planetsMap[layerIndex];

            if (mapLoader.loadChunks) {
                // layerIndex will be wrong for chunks, and we assume there will be only 1 planet
                currentPlanet = planet;
            }

            if (!currentPlanet) {
                Utils.log('WARNING: spawned entity has no corresponding planet. Planet wrap is skipped.');
            } else {
                if (currentPlanet.flip) {
                    entityPos.x = planet.perimeter - entityPos.x;
                }
                entity.attach(new PlanetWrap({
                    planet: currentPlanet,
                    correctBoundingBox: true
                }));

                // player spawn position
                if (entity.name === 'playerSpawn') {
                    currentPlanet.spawnPosition = entityPos;
                }
                // set planet if possible
                if (entity.setPlanet) {
                    entity.setPlanet(currentPlanet);
                }
            }
            // correct position
            entity.position = entityPos;

            // make entity follow a path if specified
            if (props.follow) {
                entity.attach(new PathFollower({
                    pathName: props.follow
                }));
                pathManager.acquirePath(props.follow);
            }

            if (object.name) {
                // overwrite name
                entity.name = object.name;
            }
            if (mapLoader.loadChunks) {
                // store all objects in a separate object, by x tile
                tileX = Math.floor(object.x / tileWidth) + (currentOffsetX || 0);
                // create a new object holder if needed
                chunkLoader.objects[tileX] = chunkLoader.objects[tileX] || [];
                chunkLoader.objects[tileX].push({
                    object: object,
                    entity: entity,
                    didSpawn: false,
                    shouldDespawn: !chunkLoader.isBossFight
                });

                // boss fight: count enemies
                if (chunkLoader.isBossFight && entity.isFamily('enemies')) {
                    chunkLoader.bossEnemies += 1;
                    entity.addToFamily('bossEnemies');  // mark enemies as bosses so we can count them when killed
                }
            }

            if (tileX > chunkLoader.spawnX || chunkLoader.isBossFight) {
                // don't spawn yet
                return;
            }

            Bento.objects.attach(entity);
            addShadowIfNearEdge(entity);
        };
        var mapLoader = {
            load: function (data) {
                var onSpawnComplete = data.onSpawnComplete;
                var onComplete = data.onComplete;
                load(data);
                new Tiled({
                    assetName: data.level.mapName,
                    drawTiles: false,
                    cacheModules: false,
                    merge: false,
                    spawnEntities: true, // require the module (asynchronously)
                    spawnBackground: false, // spawn background entities (drawn tile layers)
                    attachEntities: false, // attach after spawning
                    onLayer: function (layerJson, index) {
                        // Callback when the reader passes a layer
                        if (planetPerLayer) {
                            setupPlanet(layerJson, null, index);
                        }
                    },
                    onTile: function (tileX, tileY, tilesetJSON, tileIndex) {
                        // Callback after tile is drawn

                    },
                    onObject: function (objectJSON, tilesetJSON, tileIndex) {
                        // Callback when the reader passes a Tiled object
                        handleObject(objectJSON);
                    },
                    onComplete: function () {
                        // Synchronous callback when the reader passed all layers
                        // `this` references the tiled object (to get width and height)
                        offsetX += Math.floor(levelWidth / tileWidth);

                        // attach planets
                        Utils.forEach(planets, function (planetEntity, i, l, breakLoop) {
                            Bento.objects.attach(planetEntity);
                        });
                        if (onComplete) {
                            onComplete(planets);
                        }
                    },
                    onSpawn: onSpawn,
                    onSpawnComplete: function () {
                        // called after all modules are spawned (asynchronous)
                        EventSystem.fire('onSpawnComplete');
                        if (onSpawnComplete) {
                            onSpawnComplete(planets);
                        }
                    }
                });

                return planets;
            },
            loadChunks: false,
            chunkLoader: chunkLoader
        };
        return mapLoader;
    };

    return MapLoader;
});