/**
 * MoPub SDK integration for cheetah
 * @moduleName MoPub
 */
bento.define('modules/mopub', [
    'bento',
    'bento/eventsystem',
    'bento/utils'
], function (
    Bento,
    EventSystem,
    Utils
) {
    'use strict';
    
    var nothing = function () {};

    var canShowBanner = false;
    var canShowNative = false;
    var canShowInterstitial = false;
    var canShowRewardVideo = false;
    var isShowingBanner = false;
    var isShowingNative = false;
    var hasCompletedReward = false;

    // internal callbacks
    var onNativeLoaded = nothing;
    var onNativeLoadFailed = nothing;
    var onNativeShown = nothing;
    var onNativeHidden = nothing;
    var onNativeDestroyed = nothing;

    // user callbacks
    var onHideInterstitial = null;
    var onHideRewardVideo = null;

    var module = {
        init: nothing,
        canShowBanner:       function () { return false; },
        canShowNative:       function () { return false; },
        canShowInterstitial: function () { return false; },
        canShowRewardVideo:  function () { return false; },
        showBanner: nothing,
        hideBanner: nothing,
        showNative: nothing,
        hideNative: nothing,
        isShowingBanner:  function () { return false; },
        isShowingNative:  function () { return false; },
        showInterstitial: nothing,
        showRewardVideo:  nothing,
        enableAutoRefresh: nothing,
        disableAutoRefresh: nothing,
        // shouldShowConsentDialog: nothing,
        // showConsentDialog: nothing,
        grantConsent: nothing,
        revokeConsent: nothing,
    };

    var mp = window.cordova && window.cordova.plugins && window.cordova.plugins.MoPub;
    if (!mp) {
        Utils.log('Could not find MoPub SDK plugin.');
        return module;
    }

    // settings:
    // - adUnitId         : string  : token used for banner ads
    // - interstitialAdId : string  : token used for interstitial ads
    // - rewardVideoId    : string  : token used for reward video ads
    // - adSize           : string  : 'BANNER', 'SKYSCRAPER', 'LEADERBOARD', 'MEDIUM_RECTANGLE'
    // - bannerAtTop      : bool
    // - bannerOverlap    : bool
    // - offsetTopBar     : bool
    // - autoShow         : bool
    // additional settings
    // - debug            : bool    : catch and display all events
    module.init = function (settings) {

        var consent = (settings.consent !== false);

        var loadBanner = function () {
            mp.createBannerView(settings, function () {
                console.log('[MoPub] created banner view');
            }, function (error) {
                Utils.log('[MoPub] createBannerView error ' + error);
            });
        };
        var loadInterstitial = function () {
            mp.createInterstitialView(settings, function () {
                console.log('[MoPub] created interstitial view');
            }, function (error) {
                Utils.log('[MoPub] createInterstitialView error ' + error);
            });
        };
        var loadRewardVideo = function () {
            mp.createRewardVideo(settings, function () {
                console.log('[MoPub] created reward video');
            }, function (error) {
                Utils.log('[MoPub] createRewardVideo error ' + error);
            });
        };
        var loadNative = function () {
            mp.prepareNativeAd(settings, function () {
                console.log('[MoPub] prepared native ad');
                onNativeLoaded();
            }, function (error) {
                Utils.log('[MoPub] prepareNativeAd error ' + error);
                onNativeLoadFailed();
            });
        };

        var onInitSuccess = function () {

            createMethods();

            /**
             * Prepare all ads
             */
            loadBanner();
            loadInterstitial();
            loadRewardVideo();
            loadNative();
            
            /**
             * Create listeners for ad events (show/hide/load/error) to manage our internal state
             */

            if (settings.debug) {
                logAllEvents();
            }

            // ad completed
            // so fetch another one, fire callbacks with reward status
            document.addEventListener('interstitialDidDisappear', function () {
                loadInterstitial();
                if (onHideInterstitial) {
                    onHideInterstitial();
                }
            });
            document.addEventListener('rewardedVideoAdShouldRewardForAdUnitID', function () {
                hasCompletedReward = true;
            });
            document.addEventListener('rewardedVideoAdDidDisappearForAdUnitID', function () {
                loadRewardVideo();
                if (onHideRewardVideo) {
                    onHideRewardVideo(hasCompletedReward);
                }
                hasCompletedReward = false;
            });
            onNativeDestroyed = function () {
                loadNative();
            };

            // ad failed to load
            // so wait a while then fetch another one
            document.addEventListener('adViewDidFailToLoadAd', function () {
                window.setTimeout(loadBanner, 5000);
            });
            document.addEventListener('interstitialDidFailToLoadAd', function () {
                window.setTimeout(loadInterstitial, 5000);
            });
            document.addEventListener('rewardedVideoAdDidFailToPlayForAdUnitID', function () {
                window.setTimeout(loadRewardVideo, 3000);
            });
            document.addEventListener('rewardedVideoAdDidExpireForAdUnitID', function () {
                window.setTimeout(loadRewardVideo, 5000);
            });
            onNativeLoadFailed = function () {
                console.log('[MoPub] native ad failed to load, retrying');
                window.setTimeout(loadNative, 5000);
            };

            // ad loaded
            // so mark as able to show
            document.addEventListener('interstitialDidLoadAd', function () {
                mp.isInterstitialReady(function (ready) {
                    canShowInterstitial = ready;
                }, nothing);
            });
            document.addEventListener('rewardedVideoAdDidLoadForAdUnitID', function () {
                mp.isRewardVideoReady(function (ready) {
                    canShowRewardVideo = ready;
                }, nothing);
            });
            onNativeLoaded = function () {
                mp.isNativeAdReady(function (ready) {
                    canShowNative = ready;
                    if (ready) {
                        console.log('[MoPub] native ad loaded');
                    }
                });
            };

            // ad shown
            // so mark as no longer able to show?
            document.addEventListener('interstitialWillAppear', function () {
                canShowInterstitial = false;
            });
            document.addEventListener('rewardedVideoAdWillAppearForAdUnitID', function () {
                canShowRewardVideo = false;
            });
            onNativeShown = function () {
                canShowNative = false;
            };

        };

        var onInitError = function (error) {
            Utils.log('MoPub init failed: ' + error);
        };

        mp.initSdk(settings.adUnitId, settings.consent, onInitSuccess, onInitError);


        /**
         * Replace the dummy functions once we successfully initialized
         */
        var createMethods = function () {

            module.canShowBanner = function () {
                return navigator.onLine && canShowBanner;
            };
            module.canShowNative = function () {
                return navigator.onLine && canShowNative;
            };
            module.canShowInterstitial = function () {
                return navigator.onLine && canShowInterstitial;
            };
            module.canShowRewardVideo = function () {
                return navigator.onLine && canShowRewardVideo;
            };

            module.showBanner = function () {
                mp.showAd(function () {
                    isShowingBanner = true;
                }, function (error) {
                    console.log('[MoPub] show banner error: ' + error);
                });
            };
            module.hideBanner = function () {
                if (!isShowingBanner) {
                    console.log('[MoPub] already not showing banner');
                    return;
                }
                mp.hideAd(function () {
                    isShowingBanner = false;
                }, function (error) {
                    console.log('[MoPub] hide banner error: ' + error);
                });
            };

            module.showNative = function () {
                mp.isNativeAdReady(function (ready) {
                    if (ready) {
                        mp.showNativeAd(settings.position, function () {
                            isShowingNative = true;
                            onNativeShown();
                            console.log('[MoPub] native ad shown');
                        }, function (error) {
                            console.log('[MoPub] show native ad error: ' + error);
                        });
                    } else {
                        console.log('[MoPub] native ad not ready, could not be shown.');
                    }
                });
            };
            module.hideNative = function () {
                if (!isShowingNative) {
                    console.log('[MoPub] already not showing native ad');
                    return;
                }
                mp.hideNativeAd(function () {
                    isShowingNative = false;
                    onNativeHidden();
                    console.log('[MoPub] native ad hidden');
                }, function (error) {
                    console.log('[MoPub] hide native ad error: ' + error);
                });
            };
            module.destroyNative = function () {
                mp.destroyNativeAd(function () {
                    isShowingNative = false;
                    canShowNative = false;
                    onNativeDestroyed();
                    console.log('[MoPub] native ad destroyed');
                }, function (error) {
                    console.log('[MoPub] destroy native ad error: ' + error);
                });
            };
            module.isShowingBanner = function () {
                return isShowingBanner;
            };
            module.isShowingNative = function () {
                return isShowingNative;
            };
            module.showInterstitial = function (callback) {
                mp.isInterstitialReady(function (ready) {
                    if (ready) {
                        mp.showInterstitialAd(function () {
                            console.log('SHOW SUCCESS');
                        }, function () {
                            console.log('SHOW FAIL');
                        });
                    } else {
                        console.log('[MoPub] interstitial not ready');
                    }
                }, nothing);
                onHideInterstitial = callback;
            };
            module.showRewardVideo = function (callback) {
                mp.isRewardVideoReady(function (ready) {
                    if (ready) {
                        mp.showRewardVideo(function () {
                            console.log('SHOW SUCCESS');
                        }, function () {
                            console.log('SHOW FAIL');
                        });
                    } else {
                        console.log('[MoPub] rewarded video not ready');
                    }
                }, nothing);
                onHideRewardVideo = callback;
            };
            module.enableAutoRefresh = function () {
                mp.enableAutoRefresh();
            };
            module.disableAutoRefresh = function () {
                mp.disableAutoRefresh();
            };

            // module.shouldShowConsentDialog = function (callback) {
            //     mp.shouldShowConsentDialog(callback);
            // };
            // module.showConsentDialog = function (success, error) {
            //     mp.showConsentDialog(success, error);
            // };
            module.grantConsent = function () {
                mp.grantConsent();
            };
            module.revokeConsent = function () {
                mp.revokeConsent();
            };
        };

        /**
         * Enable debug logging
         */
        var logAllEvents = function () {
            var logEvent = function (name) {
                document.addEventListener(name, function (data) {
                    console.log('[MoPub Event] ' + name + ': ' + data.msg);
                });
            };

            logEvent('adViewDidFailToLoadAd');
            logEvent('willPresentModalViewForAd');
            logEvent('didDismissModalViewForAd');
            logEvent('willLeaveApplicationFromAd');
            logEvent('adViewDidLoadAd');

            logEvent('willPresentModalForNativeAd');
            logEvent('didDismissModalForNativeAd');
            logEvent('willLeaveApplicationFromNativeAd');

            logEvent('interstitialDidLoadAd');
            logEvent('interstitialDidFailToLoadAd');
            logEvent('interstitialWillAppear');
            logEvent('interstitialDidAppear');
            logEvent('interstitialWillDisappear');
            logEvent('interstitialDidDisappear');
            logEvent('interstitialDidReceiveTapEvent');

            logEvent('rewardedVideoAdDidLoadForAdUnitID');
            logEvent('rewardedVideoAdDidFailToPlayForAdUnitID');
            logEvent('rewardedVideoAdWillAppearForAdUnitID');
            logEvent('rewardedVideoAdDidAppearForAdUnitID');
            logEvent('rewardedVideoAdWillDisappearForAdUnitID');
            logEvent('rewardedVideoAdDidDisappearForAdUnitID');
            logEvent('rewardedVideoAdDidExpireForAdUnitID');
            logEvent('rewardedVideoAdDidReceiveTapEventForAdUnitID');
            logEvent('rewardedVideoAdShouldRewardForAdUnitID');
            logEvent('rewardedVideoAdWillLeaveApplicationForAdUnitID');
        };
        
    };

    return module;
});