/**
 * Creates paths that can be followed by entities
 * Includes an acquire/release mechanism so that paths can be removed from the game world once they are no longer needed.
 * @moduleName PathManager
 * @snippet PathManager.snippet
PathManager({
    gridScale: gridScale
})
 */
bento.define('modules/pathmanager', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    return function (settings) {
        
        var gridScale = settings.gridScale;

        // mapping of path name to number of references
        // this allows us to remove a path when nothing is following it anymore (i.e. its chunk got despawned)
        var pathReferenceCount = {};

        // Create a path indicator from tiled object (which entities can follow)
        var createPath = function (object, namePostfix) {
            var name = object.name;
            var props = object.properties || {};
            var i, p;
            var points = [];
            for (i = 0; i < object.polyline.length; i++) {
                p = object.polyline[i];
                points[i] = {
                    x: p.x * gridScale,
                    y: p.y * gridScale
                };
            }
            
            // console.log('createPath', name);
            
            Bento.objects.attach({
                name: name,
                points: points,
                timePerPoint: props.timePerPoint || null,
                ease: props.ease || null,
                toggle: props.toggle || null,
                toggleMode: props.toggleMode || null
            });
        };

        // increase a path's reference count
        var acquirePath = function (pathName) {
            if (!pathReferenceCount[pathName]) {
                pathReferenceCount[pathName] = 0;
            }
            pathReferenceCount[pathName]++;
            // console.log('acquirePath', pathName, pathReferenceCount[pathName]);
        };

        // decrease a path's reference count, and remove it from the world if nothing is using it anymore.
        var releasePath = function (pathName) {
            if (pathReferenceCount[pathName]) {
                if (--pathReferenceCount[pathName] <= 0) {
                    Bento.objects.get(pathName, Bento.objects.remove);
                    delete pathReferenceCount[pathName];
                    // console.log('path destroyed', pathName);
                }
            } else {
                Utils.log("attempted to release a path that was never in use? ("+pathName+")");
            }
            // console.log('releasePath', pathName, pathReferenceCount[pathName]);
        };

        return {
            createPath: createPath,
            acquirePath: acquirePath,
            releasePath: releasePath
        };
    };
});