/**
 * Factory for different planet graphic entities.
 * Usage:
 *   graphicEntity = PlanetSkins.create('skin name', planet);
 * @moduleName PlanetSkins
 */
bento.define('modules/planetskins', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/components/sprite',
    'globals',
    'bento/components/fill',
    'gui/backscroller',
    'bento/components/eventlistener',
    'managers/save'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Sprite,
    Globals,
    Fill,
    BackScroller,
    EventListener,
    SaveManager
) {
    'use strict';
    var skins = {
        'world-grass': {
            // Each skin creates an entity which contains the main planet image, as well as small animations and other details.
            // But some decorations would be best added to the level separately, using the Tiled map, so that each planet doesn't feel identical.
            createPlanet: function (skinName, planet) {
                var face = makeDecor('world-grass/face-grassplanet', 0, 0);

                var oldAngle;
                var behavior = {
                    start: function () {
                        oldAngle = planet.angle;
                    },
                    update: function () {
                        // make the face turn against the planet very slowly
                        // (not sure if this is worth doing, it's subtle)
                        face.angle += Utils.deltaAngle(oldAngle, planet.angle) * 0.05;
                        oldAngle = planet.angle;
                    }
                };
                return new Entity({
                    components: [
                        planet ? behavior : {},
                        new Sprite({
                            imageName: 'world-grass/planet-grass',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        face,
                        makeDecor('world-grass/bladegrass2', 1.01, 0.00),
                        makeDecor('world-grass/bladegrass1', 0.82, 0.12),
                        makeDecor('world-grass/bladegrass2', 1.01, 0.26),
                        makeDecor('world-grass/bladegrass1', 1.01, 0.51),
                        makeDecor('world-grass/bladegrass1', 1.02, 0.89),
                        makeDecor('world-grass/bladegrass2', 0.80, 0.70),
                        makeDecor('world-grass/flower1', 0.70, 0.80, 1.2),
                        makeDecor('world-grass/flower1', 0.70, 0.90, 1.4),
                        makeDecor('world-grass/flower2', 1.08, 0.20),
                        makeDecor('world-grass/flower2', 1.08, 0.80),
                        makeDecor('world-grass/grass1', 1.06, 0.34),
                        makeDecor('world-grass/grass2', 1.04, 0.60),
                        makeDecor('world-grass/grass1', 1.13, 0.75, 1.4),
                        makeDecor('world-grass/grass2', 1.04, 0.05),
                    ]
                });
            },
            createBackground: function () {
                // note: if background setup is similar in different skins, we should do reusable functions
                var viewport = Bento.getViewport();
                var cloudsTopBg = new BackScroller({
                    imageName: 'world-grass/cloud-top-bg',
                    spacing: new Vector2(-1, 0),
                    velocity: new Vector2(-2, 0),
                    wrapY: false
                });
                var cloudsTopFg = new BackScroller({
                    imageName: 'world-grass/cloud-top-fg',
                    spacing: new Vector2(-1, 0),
                    velocity: new Vector2(-3, 0),
                    wrapY: false
                });
                var cloudsBottomBg = new BackScroller({
                    imageName: 'world-grass/cloud-bottom-bg',
                    position: new Vector2(0, viewport.height),
                    spacing: new Vector2(-1, 0),
                    velocity: new Vector2(-0.8, 0),
                    parallax: 0.26,
                    wrapY: false
                });
                var cloudsBottomFg = new BackScroller({
                    imageName: 'world-grass/cloud-bottom-fg',
                    position: new Vector2(0, viewport.height),
                    spacing: new Vector2(-1, 0),
                    velocity: new Vector2(-1.2, 0),
                    parallax: 0.30,
                    wrapY: false
                });
                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: 'world-grass/background',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });
                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        // new Fill({
                        //     color: [0xad / 255, 0xdf / 255, 0xff / 255, 1]
                        // }),
                        backgroundFill,
                        // cloudsTopBg,
                        // cloudsTopFg,
                        cloudsBottomBg,
                        cloudsBottomFg
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width * 1.1;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },

        'world-forest': {
            createPlanet: function (skinName, planet) {
                var face = makeDecor('world-forest/forest-face', 0, 0);

                var oldAngle;
                var behavior = {
                    start: function () {
                        oldAngle = planet.angle;
                    },
                    update: function () {
                        face.angle += Utils.deltaAngle(oldAngle, planet.angle) * 0.05;
                        oldAngle = planet.angle;
                    }
                };
                return new Entity({
                    components: [
                        planet ? behavior : {},
                        new Sprite({
                            imageName: 'world-forest/planet-forest',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        face,
                        makeDecor('world-forest/grass1', 1.06, 0.20),
                        makeDecor('world-forest/grass1', 1.06, 0.56),
                        makeDecor('world-forest/flower1', 1.10, 0.80),
                        makeDecor('world-forest/grass2', 1.05, 0.40),
                        makeDecor('world-forest/grass2', 1.05, 0.90),
                        makeDecor('world-forest/mushroom', 0.78, 0.28),
                        makeDecor('world-forest/mushroom', 0.68, 0.35),
                        makeDecor('world-forest/mushroom', 0.76, 0.80),
                    ]
                });
            },
            createBackground: function () {
                var viewport = Bento.getViewport();
                var treeBg1 = new BackScroller({
                    imageName: 'world-forest/bg-tree1',
                    spacing: new Vector2(1080 - 203, 0),
                    position: new Vector2(0, 1201 + 300),
                    velocity: new Vector2(0, 0),
                    parallax: 0.4,
                    wrapY: false
                });
                var treeBg2 = new BackScroller({
                    imageName: 'world-forest/bg-tree2',
                    spacing: new Vector2(1080 - 349, 0),
                    position: new Vector2(400, 1201 + 300),
                    velocity: new Vector2(0, 0),
                    parallax: 0.4,
                    wrapY: false
                });
                var treeBg3 = new BackScroller({
                    imageName: 'world-forest/bg-tree3',
                    spacing: new Vector2(1080 - 197, 0),
                    position: new Vector2(780, 1201 + 300),
                    velocity: new Vector2(0, 0),
                    parallax: 0.4,
                    wrapY: false
                });
                var leafBg = new BackScroller({
                    imageName: 'world-forest/bg-leaf',
                    spacing: new Vector2(-1, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, 376),
                    parallax: 0.4,
                    wrapY: false
                });
                var leafFg = new BackScroller({
                    imageName: 'world-forest/fg-leaf',
                    spacing: new Vector2(-1, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, 343),
                    parallax: 0.46,
                    wrapY: false
                });
                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: 'world-forest/bg-sky',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });
                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        backgroundFill,
                        treeBg1,
                        treeBg2,
                        treeBg3,
                        leafBg,
                        leafFg
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },
        'world-disco': {
            createPlanet: function (skinName, planet) {
                var viewport = Bento.getViewport();
                var face = makeDecor('world-disco/disco-face', 0, 0);
                var oldAngle;
                var behavior = {
                    start: function () {
                        oldAngle = planet.angle;
                    },
                    update: function () {
                        // make the face turn against the planet very slowly
                        // (not sure if this is worth doing, it's subtle)
                        face.angle += Utils.deltaAngle(oldAngle, planet.angle) * 0.05;
                        oldAngle = planet.angle;

                        // shine behavior
                        shine.rotation = -planet.angle - shine.timer * 0.005;
                    }
                };
                var shine = new Entity({
                    name: 'shine',
                    scale: new Vector2(0.75, 0.75),
                    alpha: 0.5,
                    components: [
                        new Sprite({
                            imageName: 'world-disco/planet-shiny',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        /*new EventListener({
                            name: 'eventListener',
                            eventName: 'songBeat',
                            ignorePause: false,
                            onEvent: function (data) {
                                new Tween({
                                    from: 1,
                                    to: 0.75,
                                    in: 30,
                                    onUpdate: function (v, t) {
                                        shine.scale.x = v;
                                        shine.scale.y = v;
                                    }
                                });
                            }
                        })*/
                    ]
                });

                var spritePlanet = new Sprite({
                    imageName: 'world-disco/planet-bg',
                    originRelative: new Vector2(0.5, 0.5),
                });
                var spriteOverlay = new Sprite({
                    imageName: 'world-disco/planet-overlay',
                    originRelative: new Vector2(0.5, 0.5),
                });
                return new Entity({
                    components: [
                        planet ? behavior : {},
                        shine,
                        spritePlanet,
                        spriteOverlay,
                        face
                    ]
                });
            },
            createBackground: function () {
                var viewport = Bento.getViewport();
                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: 'world-disco/bg',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });
                var behavior = {
                    name: 'behavior',
                    update: function () {
                        arms1.position.y = viewport.height + Math.abs(32 * Math.sin(arms1.timer / 10));
                        arms2.position.y = viewport.height + Math.abs(32 * Math.sin(arms2.timer / 9));
                    }
                };
                var background = new Entity({
                    name: 'background',
                    components: [
                        backgroundFill
                    ]
                });
                var arms1 = new Entity({
                    name: 'arms1',
                    position: new Vector2(viewport.width / 2 + 64, viewport.height),
                    components: [
                        new Sprite({
                            imageName: 'world-disco/arms-bg',
                            originRelative: new Vector2(0.5, 1),
                            frameCountX: 1,
                            frameCountY: 1,
                            animations: {
                                default: {
                                    speed: 0,
                                    frames: [0]
                                }
                            }
                        })
                    ]
                });
                var arms2 = new Entity({
                    name: 'arms2',
                    position: new Vector2(viewport.width / 2 - 32, viewport.height),
                    components: [
                        new Sprite({
                            imageName: 'world-disco/arms-fg',
                            originRelative: new Vector2(0.5, 1),
                            frameCountX: 1,
                            frameCountY: 1,
                            animations: {
                                default: {
                                    speed: 0,
                                    frames: [0]
                                }
                            }
                        })
                    ]
                });
                var lightsTop = new BackScroller({
                    imageName: 'world-disco/lights',
                    position: new Vector2(0, viewport.height / 3),
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(-2, 0),
                    wrapY: false
                });
                lightsTop.alpha = 0.05;

                var lightsBottom = new BackScroller({
                    imageName: 'world-disco/lights',
                    position: new Vector2(0, viewport.height / 6 * 5),
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(2, 0),
                    wrapY: false
                });
                lightsBottom.alpha = 0.5;
                var container = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        behavior,
                        background,
                        lightsTop,
                        lightsBottom,
                        arms1,
                        arms2
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(container);
            }
        },
        'world-agrume': {
            // Each skin creates an entity which contains the main planet image, as well as small animations and other details.
            // But some decorations would be best added to the level separately, using the Tiled map, so that each planet doesn't feel identical.
            createPlanet: function (skinName, planet) {
                var face = new Entity({
                    name: 'face',
                    components: [
                        new Sprite({
                            imageName: skinName + '/face',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                return new Entity({
                    components: [
                        new Sprite({
                            imageName: skinName + '/planet',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        face
                    ]
                });
            },
            createBackground: function (skinName) {
                // note: if background setup is similar in different skins, we should do reusable functions
                var viewport = Bento.getViewport();
                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: skinName + '/gradient',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });
                var bg1 = new BackScroller({
                    imageName: skinName + '/bg-2',
                    spacing: new Vector2(viewport.width / 4, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(viewport.width / 2 - 432 / 2, viewport.height),
                    parallax: 0.4,
                    wrapY: false
                });
                var bg2 = new BackScroller({
                    imageName: skinName + '/bg-1',
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, viewport.height),
                    parallax: 0.3,
                    wrapY: false
                });
                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        backgroundFill,
                        bg1,
                        bg2
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width * 1.1;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },
        'world-ice': {
            // Each skin creates an entity which contains the main planet image, as well as small animations and other details.
            // But some decorations would be best added to the level separately, using the Tiled map, so that each planet doesn't feel identical.
            createPlanet: function (skinName, planet) {
                return new Entity({
                    components: [
                        new Sprite({
                            imageName: skinName + '/planet',
                            originRelative: new Vector2(0.5, 0.5),
                        })
                    ]
                });
            },
            createBackground: function (skinName) {
                // note: if background setup is similar in different skins, we should do reusable functions
                var viewport = Bento.getViewport();
                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: skinName + '/gradient',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });
                var bg1 = new BackScroller({
                    imageName: skinName + '/bg-1',
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, 536 + 128),
                    wrapY: false
                });
                var bg2 = new BackScroller({
                    imageName: skinName + '/bg-2',
                    spacing: new Vector2(-1, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(-128, viewport.height),
                    parallax: 0.2,
                    wrapY: false
                });
                var bg3 = new BackScroller({
                    imageName: skinName + '/bg-3',
                    spacing: new Vector2(-1, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, viewport.height),
                    parallax: 0.1,
                    wrapY: false
                });
                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        backgroundFill,
                        bg3,
                        bg2,
                        bg1
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width * 1.1;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },
        'world-maki': {
            // Each skin creates an entity which contains the main planet image, as well as small animations and other details.
            // But some decorations would be best added to the level separately, using the Tiled map, so that each planet doesn't feel identical.
            createPlanet: function (skinName, planet) {
                var face = new Entity({
                    name: 'face',
                    components: [
                        new Sprite({
                            imageName: skinName + '/face',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                return new Entity({
                    components: [
                        new Sprite({
                            imageName: skinName + '/planet',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        face
                    ]
                });
            },
            createBackground: function (skinName) {
                // note: if background setup is similar in different skins, we should do reusable functions
                var viewport = Bento.getViewport();
                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: skinName + '/gradient',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });
                var bg1 = new BackScroller({
                    imageName: skinName + '/bg',
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, viewport.height),
                    parallax: 0.4,
                    wrapY: false
                });
                bg1.alpha = 0.5;
                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        backgroundFill,
                        bg1
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width * 1.1;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },
        'world-pizza': {
            // Each skin creates an entity which contains the main planet image, as well as small animations and other details.
            // But some decorations would be best added to the level separately, using the Tiled map, so that each planet doesn't feel identical.
            createPlanet: function (skinName, planet) {
                var face = new Entity({
                    name: 'face',
                    components: [
                        new Sprite({
                            imageName: skinName + '/face',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                return new Entity({
                    components: [
                        new Sprite({
                            imageName: skinName + '/planet',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        face
                    ]
                });
            },
            createBackground: function (skinName) {
                // note: if background setup is similar in different skins, we should do reusable functions
                var viewport = Bento.getViewport();
                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: skinName + '/gradient',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });
                var bg1 = new BackScroller({
                    imageName: skinName + '/bg-1',
                    spacing: new Vector2(720, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(540, viewport.height),
                    parallax: 0.4,
                    wrapY: false
                });
                var bg2 = new BackScroller({
                    imageName: skinName + '/bg-2',
                    spacing: new Vector2(720, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(540 + 200, viewport.height),
                    parallax: 0.4,
                    wrapY: false
                });
                var bg3 = new BackScroller({
                    imageName: skinName + '/bg-3',
                    spacing: new Vector2(540, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, viewport.height),
                    parallax: 0.3,
                    wrapY: false
                });
                var bg4 = new BackScroller({
                    imageName: skinName + '/bg-4',
                    spacing: new Vector2(720, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(360, viewport.height),
                    parallax: 0.2,
                    wrapY: false
                });
                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        backgroundFill,
                        bg4, 
                        bg3, 
                        bg2,
                        bg1
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width * 1.1;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },
        'world-space': {
            // Each skin creates an entity which contains the main planet image, as well as small animations and other details.
            // But some decorations would be best added to the level separately, using the Tiled map, so that each planet doesn't feel identical.
            createPlanet: function (skinName, planet) {
                var face = new Entity({
                    name: 'face',
                    components: [
                        new Sprite({
                            imageName: skinName + '/face',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                return new Entity({
                    components: [
                        new Sprite({
                            imageName: skinName + '/planet',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        face
                    ]
                });
            },
            createBackground: function (skinName) {
                // note: if background setup is similar in different skins, we should do reusable functions
                var viewport = Bento.getViewport();
                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: skinName + '/gradient',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });
                var bg1 = new BackScroller({
                    imageName: skinName + '/bg-1',
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(-2, 0),
                    position: new Vector2(0, viewport.height / 2 + 128),
                    wrapY: false
                });
                var bg2 = new BackScroller({
                    imageName: skinName + '/bg-2',
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, viewport.height),
                    wrapY: false
                });
                var bg3 = new BackScroller({
                    imageName: skinName + '/bg-3',
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, 606),
                    wrapY: false
                });
                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        backgroundFill,
                        bg3,
                        bg2,
                        bg1
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width * 1.1;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },
        'world-vinyl': {
            // Each skin creates an entity which contains the main planet image, as well as small animations and other details.
            // But some decorations would be best added to the level separately, using the Tiled map, so that each planet doesn't feel identical.
            createPlanet: function (skinName, planet) {
                var face = new Entity({
                    name: 'faceVinyl', // not called face so it's not scaled in WorldSelectButton
                    components: [
                        new Sprite({
                            imageName: skinName + '/face',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                return new Entity({
                    components: [
                        new Sprite({
                            imageName: skinName + '/planet',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        face
                    ]
                });
            },
            createBackground: function (skinName) {
                // note: if background setup is similar in different skins, we should do reusable functions
                var viewport = Bento.getViewport();
                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: skinName + '/gradient',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });
                var bg1 = new BackScroller({
                    imageName: skinName + '/bg1',
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, viewport.height),
                    parallax: 0.4,
                    wrapY: false
                });
                var bg2 = new BackScroller({
                    imageName: skinName + '/bg2',
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, viewport.height),
                    parallax: 0.3,
                    wrapY: false
                });
                var bg3 = new BackScroller({
                    imageName: skinName + '/bg3',
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, viewport.height),
                    parallax: 0.2,
                    wrapY: false
                });
                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        backgroundFill,
                        bg3,
                        bg2,
                        bg1
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width * 1.1;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },

        
        'world-cog': {
            createPlanet: function (skinName, planet) {
                var face = new Entity({
                    name: 'cogFace', // not called face so it's not scaled in WorldSelectButton
                    components: [
                        new Sprite({
                            imageName: skinName + '/cog-face',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                return new Entity({
                    components: [
                        new Sprite({
                            imageName: skinName + '/cog-planet',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        face
                    ]
                });
            },
            createBackground: function (skinName, planet) {
                var viewport = Bento.getViewport();

                var oldAngle;
                var behavior = {
                    start: function () {
                        oldAngle = planet.angle;
                    },
                    update: function () {
                        cog1.rotation += Utils.deltaAngle(oldAngle, planet.angle) * 0.2;
                        cog2.rotation -= Utils.deltaAngle(oldAngle, planet.angle) * 0.2;
                        cog3.rotation += Utils.deltaAngle(oldAngle, planet.angle) * 0.2;
                        oldAngle = planet.angle;
                    }
                };

                var cog1 = new Entity({
                    name: 'cog',
                    position: new Vector2(viewport.width*0.19, viewport.height*0.67),
                    scale: new Vector2(1.1, 1.1),
                    alpha: 0.3,
                    components: [
                        new Sprite({
                            imageName: skinName + '/cog-1',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                var cog2 = new Entity({
                    name: 'cog2',
                    position: new Vector2(viewport.width*0.54, viewport.height*0.80),
                    scale: new Vector2(1.3, 1.3),
                    alpha: 0.3,
                    components: [
                        new Sprite({
                            imageName: skinName + '/cog-2',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                var cog3 = new Entity({
                    name: 'cog3',
                    position: new Vector2(viewport.width*0.82, viewport.height*0.63),
                    scale: new Vector2(1.1, 1.1),
                    alpha: 0.3,
                    components: [
                        new Sprite({
                            imageName: skinName + '/cog-3',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                var steamBg = new BackScroller({
                    imageName: skinName+'/steam-bg',
                    position: new Vector2(0, 439),
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(-2, 0),
                    alpha: 0.25,
                    wrapY: false
                });
                var steamFg = new BackScroller({
                    imageName: skinName+'/steam-fg',
                    position: new Vector2(0, 400),
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(-3, 0),
                    alpha: 0.4,
                    wrapY: false
                });
                var pipe1 = new BackScroller({
                    imageName: skinName+'/pipe-1',
                    position: new Vector2(0, viewport.height),
                    spacing: new Vector2(-1, 0),
                    velocity: new Vector2(-0.8, 0),
                    parallax: 0.26,
                    wrapY: false
                });
                var pipe2 = new BackScroller({
                    imageName: skinName+'/pipe-2',
                    position: new Vector2(0, viewport.height*0.3),
                    alpha: 0.4,
                    spacing: new Vector2(-1, 0),
                    velocity: new Vector2(-1.2, 0),
                    parallax: 0.30,
                    wrapY: false
                });
                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: skinName+'/cog-bg',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });
                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        backgroundFill,
                        cog1,
                        cog2,
                        cog3,
                        steamBg,
                        steamFg,
                        pipe1,
                        pipe2,
                        planet ? behavior : {},
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width * 1.1;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },

        'world-coin': {
            spikeStyle: 'light',
            createPlanet: function (skinName, planet) {
                var face = new Entity({
                    name: 'face',
                    components: [
                        new Sprite({
                            imageName: 'world-coin/coin-face',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                return new Entity({
                    components: [
                        new Sprite({
                            imageName: 'world-coin/coin-planet',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        face,
                    ]
                });
            },
            createBackground: function (skinName, planet) {
                // note: if background setup is similar in different skins, we should do reusable functions
                var viewport = Bento.getViewport();

                var oldAngle;
                var t = 0;

                var behavior = {
                    start: function () {
                        oldAngle = planet.angle;
                    },
                    update: function (data) {
                        t += data.speed * 0.007 - Utils.deltaAngle(oldAngle, planet.angle) * 0.5;
                        flower1.rotation = 0.3*Math.sin(t);
                        flower2.rotation = 0.3*Math.sin(t+1);
                        flower3.rotation = 0.3*Math.sin(t+2);
                        oldAngle = planet.angle;
                    }
                };

                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: skinName + '/coin-bg',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });

                var flower1 = new Entity({
                    name: 'flower',
                    position: new Vector2(viewport.width*0.80, viewport.height*0.75),
                    scale: new Vector2(1.1, 1.1),
                    alpha: 1.0,
                    components: [
                        new Sprite({
                            imageName: skinName + '/flower-bg-1',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                var flower2 = new Entity({
                    name: 'flower2',
                    position: new Vector2(viewport.width*0.23, viewport.height*0.5),
                    scale: new Vector2(1.3, 1.3),
                    alpha: 1.0,
                    components: [
                        new Sprite({
                            imageName: skinName + '/flower-bg-2',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                var flower3 = new Entity({
                    name: 'flower3',
                    position: new Vector2(viewport.width*0.70, viewport.height*0.26),
                    scale: new Vector2(1.1, 1.1),
                    alpha: 1.0,
                    components: [
                        new Sprite({
                            imageName: skinName + '/flower-bg-3',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });

                var treasure = new BackScroller({
                    imageName: skinName + '/treasure',
                    spacing: new Vector2(0, 0),
                    velocity: new Vector2(0, 0),
                    position: new Vector2(0, viewport.height+200),
                    parallax: 0.1,
                    wrapY: false,
                });

                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        backgroundFill,
                        flower1,
                        flower2,
                        flower3,
                        treasure,
                        planet ? behavior : {},
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width * 1.1;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },
        'world-watermelon': {
            spikeStyle: 'light',
            createPlanet: function (skinName, planet) {
                var face = new Entity({
                    name: 'face',
                    components: [
                        new Sprite({
                            imageName: skinName + '/watermelon-face',
                            originRelative: new Vector2(0.5, 0.5)
                        })
                    ]
                });
                return new Entity({
                    components: [
                        new Sprite({
                            imageName: skinName + '/watermelon-planet',
                            originRelative: new Vector2(0.5, 0.5),
                        }),
                        face
                    ]
                });
            },
            createBackground: function (skinName, planet) {
                var viewport = Bento.getViewport();

                var oldAngle;
                var t = 0;

                var behavior = {
                    start: function () {
                        oldAngle = planet.angle;
                    },
                    update: function (data) {
                        t += data.speed * 0.03 - Utils.deltaAngle(oldAngle, planet.angle) * 0.5;
                        leaf1.rotation = 0.1*Math.sin(t);
                        leaf2.rotation = 0.1*Math.sin(t+1);
                        oldAngle = planet.angle;
                    }
                };

                var backgroundFill = new Entity({
                    scale: new Vector2(64, 0),
                    components: [
                        new Sprite({
                            imageName: skinName + '/watermelon-bg',
                            originRelative: new Vector2(0, 0),
                            frameCountX: 3,
                            animations: {
                                default: {
                                    frames: [1] // take the middle frame for edge bleeding
                                }
                            }
                        })
                    ]
                });

                var leaf1 = new Entity({
                    name: 'leaf',
                    position: new Vector2(viewport.width*0.14, viewport.height*0.68),
                    scale: new Vector2(1.0, 1.0),
                    alpha: 1.0,
                    components: [
                        new Sprite({
                            imageName: skinName + '/leaf-1',
                            originRelative: new Vector2(0.65, 0.1)
                        })
                    ]
                });
                var leaf2 = new Entity({
                    name: 'leaf2',
                    position: new Vector2(viewport.width*0.83, viewport.height*0.65),
                    scale: new Vector2(1.2, 1.2),
                    alpha: 1.0,
                    components: [
                        new Sprite({
                            imageName: skinName + '/leaf-2',
                            originRelative: new Vector2(0.4, 0.1)
                        })
                    ]
                });

                var vines1 = new BackScroller({
                    imageName: skinName+'/liana-1',
                    position: new Vector2(0, viewport.height*0.9),
                    spacing: new Vector2(0, 0),
                    // velocity: new Vector2(-0.8, 0),
                    velocity: new Vector2(0, 0),
                    parallax: 0.26,
                    wrapY: false
                });
                var vines2 = new BackScroller({
                    imageName: skinName+'/liana-2',
                    position: new Vector2(0, viewport.height*0.98),
                    alpha: 1.0,
                    spacing: new Vector2(0, 0),
                    // velocity: new Vector2(-1.2, 0),
                    velocity: new Vector2(0, 0),
                    parallax: 0.30,
                    wrapY: false
                });
                var vinesTop = new BackScroller({
                    imageName: skinName+'/liana-top',
                    position: new Vector2(0, 440),
                    alpha: 1.0,
                    spacing: new Vector2(0, 0),
                    // velocity: new Vector2(-1.2, 0),
                    velocity: new Vector2(0, 0),
                    parallax: 0.30,
                    wrapY: false
                });

                var background = new Entity({
                    z: Globals.Layers.BACKGROUND,
                    name: 'background',
                    float: true,
                    components: [
                        backgroundFill,
                        leaf1,
                        leaf2,
                        vines1,
                        vines2,
                        vinesTop,
                        planet ? behavior : {},
                    ]
                });
                backgroundFill.scale.x = viewport.width / backgroundFill.dimension.width * 1.1;
                backgroundFill.scale.y = viewport.height / backgroundFill.dimension.height;
                Bento.objects.attach(background);
            },
        },

        // ADD MORE SKINS HERE
        'my skin': {
            createPlanet: function (skinName, planet) {
                return new Entity({
                    components: [
                        new Sprite({
                            imageName: 'my planet image',
                            originRelative: new Vector2(0.5, 0.5),
                        })
                    ]
                });
            },
            createBackground: function (skinName) {}
        }
    };

    // shortcut to create a sprite with 4 frames, relative to centre of planet
    var makeDecor = function (imageName, radius, angleMul, scale) {
        var sprite = new Sprite({
            imageName: imageName,
            originRelative: new Vector2(0.5, 0.5),
            frameCountX: 2,
            frameCountY: 2,
            animations: {
                default: {
                    speed: 0.2,
                    frames: [0, 1, 2, 3]
                }
            }
        });
        sprite.angle = angleMul * 2 * Math.PI;
        sprite.scale = scale || 1;
        sprite.origin.y += radius * 256;

        sprite.draw = function (data) {
            var renderer = data.renderer;
            if (!this.currentAnimation || !this.visible) {
                return;
            }
            this.updateFrame();
            renderer.rotate(-this.angle);
            renderer.drawImage(
                this.spriteImage,
                this.sourceX,
                this.sourceY,
                this.frameWidth,
                this.frameHeight,
                (-this.origin.x) | 0,
                (-this.origin.y) | 0,
                this.frameWidth * this.scale,
                this.frameHeight * this.scale
            );
            renderer.rotate(this.angle);
        };

        return sprite;
    };

    return {
        /**
         * Create a new planet graphical entity from a given skin name.
         * The `planet` parameter can be omitted if you are creating this skin outside of a real level.
         * @snippet PlanetSkins.create|snippet
            PlanetSkins.create('${1:skin}', ${2:planet})
         */
        create: function (skinName, planet) {
            var skin = skins[skinName];
            if (!skin) {
                Utils.log('No such skin "' + skinName + '"');
                skin = skins['world-grass'];
            }
            var graphic = skin.createPlanet(skinName, planet);
            graphic.name = 'planetGraphic';
            if (planet && planet.radius) {
                graphic.scale.scalarMultiplyWith(planet.radius / 256);
            }
            return graphic;
        },
        createBackground: function (skinName, planet) {
            var skin = skins[skinName] || skins['world-grass'];
            skin.createBackground(skinName, planet);
        },

        getCurrentSkin: function () {
            return skins[SaveManager.load('planetSkin')] || skins['world-grass'];
        }
    };
});