/**
 * Utility to simplify GET and POST requests
 * @moduleName XHR
 * @snippet XHR|get
XHR({
    get: '${1:url}',
    headers: {}, // extra headers
    params: {},  // GET parameters to be appended to URL
    onSuccess: function (res) {
        // res = {response: 'responseText'} or parsed JSON
    },
    onFail: function (res) {
        // res = {response: 'responseText'} or parsed JSON
    }
})
 * @snippet XHR|post
XHR({
    get: '${1:url}',
    headers: {}, // extra headers
    params: {},  // parameters to be encoded in POST body
    json: {},    // or, JSON data to go in POST body
    onSuccess: function (res) {
        // res = {response: 'responseText'} or parsed JSON
    },
    onFail: function (res) {
        // res = {response: 'responseText'} or parsed JSON
    }
})
 */
bento.define('xhr', [], function () {
    'use strict';
    return function (settings) {
        var xhr = new window.XMLHttpRequest();

        var getPostBody = function () {
            // turns the params string into a url encoded form
            var param;
            var params = settings.params;
            var first = true;
            var url = '';
            for (param in params) {
                if (!params.hasOwnProperty(param)) {
                    continue;
                }

                if (!first) {
                    url += '&';
                }
                url += param + '=' + params[param];
                first = false;
            }
            return url;
        };
        var addHeaders = function () {
            var key;
            var headers = settings.headers;
            for (key in headers) {
                if (!headers.hasOwnProperty(key)) {
                    continue;
                }
                xhr.setRequestHeader(key, headers[key]);
            }
        };

        if (settings.get) {
            if (settings.params) {
                settings.get += '?'+getPostBody();
            }
            xhr.open('GET', settings.get, true);
        } else if (settings.post) {
            xhr.open('POST', settings.post, true);
            if (settings.json) {
                xhr.setRequestHeader('Content-Type', 'text/plain');
                xhr.setRequestHeader('X-Requested-With', 'XMLHttpRequest');
            } else {
                xhr.setRequestHeader('Content-type', 'application/x-www-form-urlencoded');
            }
        }

        if (settings.headers) {
            addHeaders();
        }

        xhr.onreadystatechange = function () {
            if (xhr.readyState === 4) {
                if (xhr.status >= 200 && xhr.status < 300) {
                    if (!settings.onSuccess) {
                        return;
                    }
                    try {
                        settings.onSuccess(JSON.parse(xhr.responseText));
                    } catch (e) {
                        settings.onSuccess({
                            response: xhr.responseText
                        });
                    }
                } else {
                    if (!settings.onFail) {
                        return;
                    }
                    try {
                        settings.onFail(JSON.parse(xhr.responseText));
                    } catch (e) {
                        settings.onFail({
                            response: xhr.responseText
                        });
                    }
                }
            }
        };

        if (settings.json) {
            xhr.send(JSON.stringify(settings.params));
        } else if (settings.post) {
            xhr.send(getPostBody());
        } else {
            xhr.send();
        }

        return xhr;
    };
});