/**
 * Main screen
 */
bento.define('screens/main', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/screen',
    'bento/tween',
    'globals',
    'entities/planet',
    'entities/spike',
    'components/planetwrap',
    'bento/components/fill',
    'entities/player',
    'modules/maploader',
    'entities/background',
    'entities/musicmanager',
    'gui/taptostart',
    'entities/playercarrier',
    'entities/zoom',
    'bento/components/eventlistener',
    'entities/camera',
    'gui/deathoverlay',
    'entities/speedmanager',
    'entities/scoremanager',
    'gui/progressbar',
    'entities/tutorials/holdtojump',
    'managers/save',
    'gui/coincounter',
    'gui/overlays/reviveforad',
    'gui/buttons/pause',
    'gui/overlays/secondchance',
    'modules/mopub'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Screen,
    Tween,
    Globals,
    Planet,
    Spike,
    PlanetWrap,
    Fill,
    Player,
    MapLoader,
    Background,
    MusicManager,
    TapToStart,
    PlayerCarrier,
    Zoom,
    EventListener,
    Camera,
    DeathOverlay,
    SpeedManager,
    ScoreManager,
    ProgressBar,
    TutorialHoldToJump,
    SaveManager,
    CoinCounter,
    ReviveForAd,
    PauseButton,
    SecondChance,
    MoPub
) {
    'use strict';

    var player;
    var planets;
    var currentPlanet;

    // this is used to tilt the world slightly ahead of where the player is looking
    var desiredAngle;

    // multiplier to specify how quickly the planet should reach the desired angle
    // 0.0 = don't rotate planet at all
    // 1.0 = instantly snap to desired angle
    // tween this from 0 to prevent jitters when the player teleports.
    var desiredAngleSpeed = 1.0;


    var onShow = function (settings) {
        var viewport = Bento.getViewport();
        var level = findLevelByName(Globals.currentWorld);
        var speedManager;
        var mapLoader = new MapLoader({});
        var chunkLoader = mapLoader.chunkLoader;
        var scoreManager = new ScoreManager({
            worldName: Globals.currentWorld
        });
        var zoom = new Zoom({
            zoom: 0.8
        });
        var buttonPause = new PauseButton({});

        SaveManager.save('gainedCoins', 0);
        SaveManager.save('usedRevive', 0);
        SaveManager.save('usedSecondChance', 0);
        desiredAngleSpeed = 1.0;

        PlanetWrap.wrapOffset = 0; // reset planet wrap position

        /**
         * Experiment: follow camera with fish
         */
        var camera = new Camera({
            speed: new Vector2(0.1, 0.1)
        }).attach({
            update: function () {
                var groundPosition = new Vector2(player.position.x, 32);
                var position = Utils.wrapPoint(groundPosition, currentPlanet.radius, currentPlanet.angle);
                camera.ignoreEdge();
                position.addTo(currentPlanet.position).addTo(groundPosition);
                camera.centerTo(position);

                if (player.hasStarted()) {
                    camera.setSpeed(new Vector2(0.1, 0.1));
                }
            }
        });

        /**
         * Experiment: move player along the planet
         */
        var playerMover = {
            update: function () {
                if (player.hasStarted() && player.getPlanet() === currentPlanet) {
                    if (player.getDirection() === 'left') {
                        desiredAngle -= 0.005;
                    } else {
                        desiredAngle += 0.005;
                    }
                }
            }
        };

        /**
         * Experiment: zoom out planet as fish walks along surface
         */
        var zoomer = {
            zoom: 0.8,
            zoomTarget: 0.8,
            update: function () {
                // // interpolate to zoom target
                // zoomer.zoom += (zoomer.zoomTarget - zoomer.zoom) * 0.01;
                zoomer.zoom = 0.7 + 0.1 * Math.cos(desiredAngle * 2);
                zoom.zoom(zoomer.zoom);

                // move planet down
                viewport.y = 128 - 256 * Math.cos(desiredAngle);
            }
        };
        var tapToStart = new TapToStart({
            position: Vector2(viewport.width * 0.5, viewport.height * 0.3)
        });

        var updateDesiredAngle = function (data) {
            if (data.animate) {
                new Tween({
                    from: desiredAngle,
                    to: data.direction === 'left' ? Globals.VIEW_TILT : -Globals.VIEW_TILT,
                    in: 160,
                    delay: 0,
                    applyOnDelay: 0,
                    ease: 'easeInOutQuad',
                    onUpdate: function (v, t) {
                        desiredAngle = v;
                    }
                });
            } else {
                desiredAngle = data.direction === 'left' ? Globals.VIEW_TILT : -Globals.VIEW_TILT;
            }
        };

        /**
         * Aligns planet with player position
         */
        var planetMover = {
            update: function () {
                if (Globals.DISABLE_PLANET_ROTATION) {
                    return;
                }
                if (player.getPlanet() !== currentPlanet) {
                    // move the player to the new planet
                    player.position.x = currentPlanet.spawnPosition.x;
                    player.position.y = currentPlanet.spawnPosition.y;
                    player.setPlanet(currentPlanet);
                } else if (player.isAlive()) {
                    player.getComponent('planetWrap', function (playerWrap) {
                        var delta = Utils.deltaAngle(playerWrap.angle, desiredAngle);
                        currentPlanet.angle += delta * desiredAngleSpeed;

                        // useful for parallax effects etc.
                        EventSystem.fire('planetAngleChanged', {
                            planet: currentPlanet,
                            angle: currentPlanet.angle,
                            delta: delta
                        });
                    });
                }
            }
        };

        Bento.objects.attach(scoreManager);
        Bento.objects.attach(zoom);

        this.dimension.width = Infinity;
        this.dimension.height = Infinity;
        loadPlanets(mapLoader, function () {
            desiredAngle = currentPlanet.flip ? Globals.VIEW_TILT : -Globals.VIEW_TILT;
            player = new Player({
                planet: currentPlanet,
                position: currentPlanet.spawnPosition.clone()
            });
            player.setSize(currentPlanet.playerSize);
            speedManager = new SpeedManager({
                player: player,
                planet: currentPlanet,
                chunkLoader: chunkLoader
            });

            Bento.objects.attach(player);

            // tween in angle speed
            new Tween({
                from: 0,
                to: desiredAngleSpeed,
                in: 60,
                onUpdate: function (v, t) {
                    desiredAngleSpeed = v;
                }
            });
            Bento.objects.attach(planetMover);

            // set up chunkloader (depedency on the player module)
            if (mapLoader.loadChunks) {
                chunkLoader.getCurrentX = function () {
                    return Math.floor(player.position.x / chunkLoader.gridScale / chunkLoader.tiledSize);
                };
                Bento.objects.attach(chunkLoader);
                Bento.objects.attach(speedManager);

                // TEMP: for testing purposes
                window.startBoss = function () {
                    EventSystem.fire('startBoss', {
                        bossNumber: 0
                    });
                };

                // HUD entities
                Bento.objects.attach(new ProgressBar({
                    position: new Vector2(viewport.width * 0.5, 80)
                }));
                Bento.objects.attach(new CoinCounter({}));
                Bento.objects.attach(buttonPause);
            }
            // reset game speed
            SpeedManager.reset();

            // experiments
            // Bento.objects.attach(camera);
            // Bento.objects.attach(zoomer);
            // Bento.objects.attach(playerMover);
            Bento.objects.attach(new EventListener({
                eventName: 'setDirection',
                onEvent: updateDesiredAngle
            }));
            Bento.objects.attach(tapToStart);

            if (Globals.SKIP_BUTTON && Globals.levels[Globals.currentLevel + 1]) {
                Bento.objects.attach(ClickButton({
                    z: Globals.Layers.GUI,
                    imageName: 'ui/button-play',
                    frameCountX: 1,
                    frameCountY: 1,
                    animations: {
                        up:       { frames: [0] },
                        down:     { frames: [0] },
                        inactive: { frames: [0] },
                    },
                    position: new Vector2(viewport.width * 0.91, viewport.height * 0.05),
                    scale: Vector2(0.3, 0.3),
                    alpha: 0.4,
                    float: true,
                    onClick: function () {
                        Globals.currentLevel++;
                        Bento.screens.show('screens/main');
                    }
                }));
            }
        });

        EventSystem.on('startLevel', onStartLevel);
        EventSystem.on('playerDead', onPlayerDead);
        EventSystem.on('nextPlanet', onNextPlanet);
        EventSystem.on('startBoss', onStartBoss);

        // TODO: move this tutorial
        if (SaveManager.load('tutorialHoldToJump') < 2) {
            Bento.objects.attach(new TutorialHoldToJump({}));
        }

        // prevent ad-refresh from eating inputs during gameplay
        MoPub.disableAutoRefresh();
    };

    var onHide = function () {
        EventSystem.off('startLevel', onStartLevel);
        EventSystem.off('playerDead', onPlayerDead);
        EventSystem.off('nextPlanet', onNextPlanet);
        EventSystem.off('startBoss', onStartBoss);

        MoPub.enableAutoRefresh();
    };


    var loadPlanets = function (mapLoader, onComplete) {
        var viewport = Bento.getViewport();
        var level = findLevelByName(Globals.currentWorld);
        mapLoader.load({
            level: level,
            onComplete: function (planetsSpawned) {
                planets = planetsSpawned;
                currentPlanet = planets[0];
                viewport.x = currentPlanet.position.x - viewport.width / 2;
                viewport.y = currentPlanet.position.y - viewport.height * (Utils.isWideScreen ? 0.65 : 0.75);
                EventSystem.fire('setActivePlanet', currentPlanet);
            },
            onSpawnComplete: function (planetsSpawned) {
                if (onComplete) {
                    onComplete();
                }
            }
        });
    };

    var findLevelByName = function (name) {
        var currentLevel = null;
        Utils.forEach(Globals.levels, function (level, i, l, breakLoop) {
            if (level.mapName === name) {
                currentLevel = level;
                breakLoop();
            }
        });
        if (!currentLevel) {
            Utils.log("WARNING: map " + name + " was not found!");
        }
        return currentLevel;
    };

    // called when we catch a rocket
    var onNextPlanet = function (data) {
        var viewport = Bento.getViewport();

        var sourceEntity = data.source;
        var destEntity = data.destination;
        var nextPlanet = destEntity.planet;
        if (!nextPlanet) {
            console.log(destEntity);
            Utils.log('destination does not have a planet property! (rocket must travel to a named position indicator)');
        }
        desiredAngleSpeed = 0;

        Bento.objects.attach(new PlayerCarrier({
            sourcePlanet: currentPlanet,
            destPlanet: nextPlanet,
            posInSourcePlanet: sourceEntity.position,
            posInDestPlanet: destEntity.position,
            onComplete: function () {
                currentPlanet = nextPlanet;
                player.position.x = destEntity.position.x;
                player.position.y = destEntity.position.y;
                EventSystem.fire('setActivePlanet', currentPlanet);
                EventSystem.fire('setDirection', {
                    direction: currentPlanet.flip ? 'left' : 'right',
                    animate: false
                });
                player.setPlanet(currentPlanet);
                Tween({
                    from: 0,
                    to: 1.0,
                    in: 60,
                    ease: 'linear',
                    onUpdate: function (v, t) {
                        desiredAngleSpeed = v;
                    }
                });
            }
        }));

        Tween({
            from: viewport.x,
            to: nextPlanet.position.x - viewport.width * 0.5,
            in: Globals.ROCKET_FLIGHT_DURATION,
            delay: Globals.ROCKET_FLIGHT_DELAY,
            ease: 'easeInOutQuad',
            onUpdate: function (v, t) {
                viewport.x = v;
            }
        });
        Tween({
            from: viewport.y,
            to: nextPlanet.position.y - viewport.height * 0.7,
            in: Globals.ROCKET_FLIGHT_DURATION,
            delay: Globals.ROCKET_FLIGHT_DELAY,
            ease: 'easeInOutQuad',
            onUpdate: function (v, t) {
                viewport.y = v;
            }
        });
    };

    var onStartLevel = function () {
        EventSystem.fire('GameAnalytics-custom', {
            id: Globals.currentWorld+':start'
        });
        // TODO: we may want to find a proper usage for progression events
        // EventSystem.fire('GameAnalytics-startProgression', {
        //     id: Globals.currentWorld
        // });
    };

    var onPlayerDead = function () {
        Utils.timeout(30, function () {
            if (SaveManager.load('usedSecondChance') === 0 && Math.random() < Globals.secondChanceProbability) {
                Bento.objects.attach(new SecondChance({}));
            } else if (SaveManager.load('usedRevive') === 0 && ScoreManager.getScore() >= 300 && MoPub.canShowRewardVideo()) {
                Bento.objects.attach(new ReviveForAd({}));
            } else {
                Bento.objects.attach(new DeathOverlay({}));
            }
        });

        // analytics for which chunk we died in
        Bento.objects.get('chunkLoader', function (chunkLoader) {
            var culprit = chunkLoader.getCurrentChunkInfo();
            // console.log('died in chunk', culprit);

            if (culprit) {
                var nameParts = culprit.assetName.split('/');

                // trim the chunks directory itself
                if (nameParts[0] === 'chunks') {
                    nameParts.shift();
                }
                EventSystem.fire('GameAnalytics-custom', {
                    id: 'chunkDeath:'+nameParts.join(':'),
                    value: ScoreManager.getScore()
                });
            }
        });
    };

    var onStartBoss = function (data) {
        var chunkLoader = Bento.objects.get('chunkLoader');
        var speedManager = Bento.objects.get('speedManager');
        var currentBoss = data.bossNumber;
        var enemiesKilled = 0;

        if (!chunkLoader) {
            Utils.log('Attempted to start boss without chunkLoader');
        } else if (!speedManager) {
            Utils.log('Attempted to start boss without speedManager');
        } else {
            chunkLoader.despawnAll(function () {
                // test
                chunkLoader.isBossFight = true;

                // delay boss spawn a little (until slowdown is complete)
                new Tween({
                    ignoreGameSpeed: true,
                    in: 30,
                    onComplete: function () {
                        // chunkLoader.spawnBossChunk(4);
                        // currentBoss = SaveManager.load('bossesCleared-' + Globals.currentWorld);
                        // chunkLoader.spawnBossChunk(currentBoss % chunkLoader.totalChunksBoss);
                        chunkLoader.spawnBossChunk();
                    }
                });
            }, false);

            // chunkloader depends on the position of the player, we can't wrap at 0 anymore
            // so we wrap the planet at the current position of player
            currentPlanet.infinite = false;
            // align with the tilegrid?
            // PlanetWrap.wrapOffset = player.position.x;
            PlanetWrap.wrapOffset = Math.floor(player.position.x / chunkLoader.gridScale / chunkLoader.tiledSize) * chunkLoader.tiledSize * chunkLoader.gridScale;

            // speedManager.tweenBackToNormal(); // play against bosses on normal speed?

            var bossListener = new EventListener({
                name: 'endBossFight',
                eventName: 'killedEnemy',
                ignorePause: false,
                onEvent: function (enemy) {

                    // don't count non-boss enemies
                    if (!enemy.isFamily('bossEnemies')) {
                        return;
                    }

                    var timer = 0;
                    var bossCleared = function () {
                        Bento.objects.remove(bossListener);
                        Bento.audio.playSound('sfx_applause');

                        // restore chunkloader
                        chunkLoader.despawnAll(function () {
                            // test
                            chunkLoader.isBossFight = false;
                            currentPlanet.infinite = true;
                        });

                        chunkLoader.spawnBonusCoins();

                        EventSystem.fire('bossCleared');

                        // keep track of bosses cleared
                        SaveManager.add('totalBossesCleared-' + Globals.currentWorld, 1);
                    };
                    // last of the enemies?
                    enemiesKilled += 1;
                    if (enemiesKilled >= chunkLoader.bossEnemies) {
                        // using a timeout with the tween is risky regarding reviving the player
                        // we add a timer to the bossListener object instead
                        bossListener.update = function () {
                            timer += 1;
                            if (timer > 60) {
                                bossCleared();
                                bossListener.update = null;
                            }
                        };
                        chunkLoader.unsolidifyAll();  // make the world safe immediately.
                    }

                    // announce that we killed a boss enemy (useful for boss health bar)
                    EventSystem.fire('killedBossEnemy', {
                        enemiesKilled: enemiesKilled
                    });
                }
            });
            Bento.objects.attach(bossListener);
        }
    };

    return new Screen({
        onShow: onShow,
        onHide: onHide
    });
});