/**
 * Extends utils with more useful functions.
 * See bento/utils
 */
bento.define('utils', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    Utils,
    Tween
) {
    'use strict';
    var EPSILON = Number.EPSILON || Math.pow(2, -52);
    var utils = {
        EPSILON: EPSILON,
        isWideScreen: false, // iPhoneX and Android with 1:2 ratio
        /**
         * List of in app purchase ids, these ids must exist in the App Store or Google Play
         * @snippet Utils.IapToId|snippet
            Utils.IapToId['$1']
         */
        IAP: [
            'com.luckykat.fishhop.noads',
            'com.luckykat.fishhop.epicpack', // legacy IAP name. Should result in epic1 and epic2 being unlocked.
            'com.luckykat.fishhop.epic1',
            'com.luckykat.fishhop.epic2',
            'com.luckykat.fishhop.worlds1',
        ],
        IapToId: { 
            // map generic iap name to the actual and vice versa (note: never rename these)
            'com.luckykat.fishhop.noads': 'noads',
            'com.luckykat.fishhop.epicpack': 'epicpack',
            'com.luckykat.fishhop.epic1': 'epic1',
            'com.luckykat.fishhop.epic2': 'epic2',
            'com.luckykat.fishhop.worlds1': 'epic1',
            'noads': 'com.luckykat.fishhop.noads',
            'epicpack': 'com.luckykat.fishhop.epicpack',
            'epic1': 'com.luckykat.fishhop.epic1',
            'epic2': 'com.luckykat.fishhop.epic2',
            'worlds1': 'com.luckykat.fishhop.worlds1',
        },
        isFB: function () {
            return false;
        },
        /**
         * Wrapper for tweens to simulate a timeout
         */
        timeout: function (time, callback) {
            var tween = new Tween({
                from: 0,
                to: 1,
                in: time,
                ease: 'linear',
                onComplete: callback
            });
            return tween;
        },
        //http://javascript.about.com/od/problemsolving/a/modulobug.htm
        modulo: function (b, n) {
            return ((b % n) + n) % n;
        },
        /**
         * Adds elements from arr2 into arr1, if they don't already exist.
         * (this modifies arr1 in-place and returns it)
         */
        concatUnique: function (arr1, arr2) {
            var i, item, len = arr2.length;
            for (i = 0; i < len; i++) {
                item = arr2[i];
                if (arr1.indexOf(item) === -1) {
                    arr1.push(item);
                }
            }
            return arr1;
        },

        /**
         * Gets the difference between angles, but take a shortcut so that we don't travel more than 180 degrees.
         */
        deltaAngle: function (a, b) {
            while (b > a + Math.PI) {
                b -= 2 * Math.PI;
            }
            while (b < a - Math.PI) {
                b += 2 * Math.PI;
            }
            return b - a;
        },

        wrapPoint: function (point, radius, additionalAngle) {
            var angle = point.x / radius + (additionalAngle || 0);
            var pointRadius = radius - point.y;
            var sin = Math.sin(angle);
            var cos = Math.cos(angle);
            var result = new Vector2(
                sin * pointRadius - point.x,
                (1 - cos) * pointRadius - point.y - pointRadius
            );
            result.sin = sin;
            result.cos = cos;
            return result;
        },

        samePlanet: function (e1, e2) {
            var w1 = e1.getComponent('planetWrap');
            var w2 = e2.getComponent('planetWrap');
            return !w1 || !w2 || w1.planet === w2.planet;
        },

        oppositeDirection: function (dir) {
            return dir === 'left' ? 'right' : 'left';
        },

        getHashParameter: function (key, fallback, url) {
            var match;
            var result;
            url = url || window.location.hash;

            match = RegExp('[#&]' + key + '=([^&]*)').exec(url);
            result = match && decodeURIComponent(match[1].replace(/\+/g, ' '));

            if (result !== null) {
                return result;
            } else if (Utils.isDefined(fallback)) {
                return fallback;
            } else {
                return null;
            }
        },
        /**
         * Get tile property safely, even if it's undefined in Tiled
         * @snippet Utils.getTileProperty|snippet
            Utils.getTileProperty(settings, '${1:prop}', ${2:0})
         */
        getTileProperty: function (settings, propName, defaultValue) {
            return ((settings.tiled || {}).tileProperties || {})[propName] || defaultValue;
        },
        /**
         * Get object property safely, even if it's undefined in Tiled
         * @snippet Utils.getObjectProperty|snippet
            Utils.getObjectProperty(settings, '${1:prop}', ${2:0})
         */
        getObjectProperty: function (settings, propName, defaultValue) {
            return ((settings.tiled || {}).objectProperties || {})[propName] || defaultValue;
        },
        /**
         * Get object safely, even if it's undefined in Tiled
         * @snippet Utils.getTiledObject|snippet
            Utils.getTiledObject(settings)
         */
        getTiledObject: function (settings) {
            return ((settings.tiled || {}).object || {});
        },
        isFiniteRectangle: function (r) {
            return isFinite(r.x) && isFinite(r.y) && isFinite(r.width) && isFinite(r.height);
        },
        /**
         * Get time string in hh:mm:ss format from milliseconds
         * @snippet Utils.getTimeString|String
                Utils.getTimeString(${1:time});
         */
        getTimeString: function (milliSeconds) {
            var currentTime = milliSeconds;
            var s = currentTime / 1000;
            var m = s / 60;
            var h = m / 60;
            var addDigit = function (value) {
                return ('00' + value).slice(-2);
            };

            if (currentTime > 0) {
                return addDigit(Math.floor(h)) + ':' + addDigit(Math.floor(m) % 60) + ':' + addDigit(Math.floor(s) % 60);
            } else {
                return '00:00:00';
            }
        },

        isIPhoneX: function () {
            var device = window.device;
            return device && ['iPhone10,3', 'iPhone10,6'].indexOf(device.model) !== -1;
        },

        // schedule a bunch of tweens to happen one after the other
        // (each one borrows settings from the last)
        sequenceTweens: function (settingsArr) {
            var i, settings, prevSettings = {};
            for (i=0; i<settingsArr.length; i++) {
                settings = settingsArr[i];
                settings.delay = (settings.delay || 0) + (prevSettings.delay || 0) + (prevSettings.in || 0);
                settings.applyOnDelay = false;
                if (!settings.onUpdate) {
                    settings.onUpdate = prevSettings.onUpdate;
                }
                if (settings.from === undefined) {
                    settings.from = prevSettings.to;
                }
                new Tween(settings);
                prevSettings = settings;
            }
        },

        currentScreenName: function () {
            var screen = Bento.screens.getCurrentScreen();
            if (!screen) {
                return 'noscreen';
            }
            var parts = screen.name.split('/');
            return parts[1] || parts[0];
        },

        openUrl: function (url) {
            if (window.Cocoon) {
                window.Cocoon.App.openURL(url);
            } else if (window.Cordova && window.Cordova.InAppBrowser) {
                window.Cordova.InAppBrowser.open(url, '_system');
            } else {
                window.open(url, '_blank');
            }
        }
    };
    Utils.extend(Utils, utils);
    return Utils;
});